@@ -20,7 +20,7 @@
 #include "libcman.h"
 #include "cman_tool.h"
 
-#define OPTION_STRING		("m:n:v:e:2p:c:r:i:N:t:o:k:Vwfqah?d::")
+#define OPTION_STRING		("m:n:v:e:2p:c:r:i:N:t:o:k:F:Vwfqah?d::")
 #define OP_JOIN			1
 #define OP_LEAVE		2
 #define OP_EXPECTED		3
@@ -31,13 +31,14 @@
 #define OP_STATUS		8
 #define OP_NODES		9
 #define OP_SERVICES		10
+#define OP_DEBUG		11
 
 
 static void print_usage(int subcmd)
 {
 	printf("Usage:\n");
 	printf("\n");
-	printf("%s <join|leave|kill|expected|votes|version|wait|status|nodes|services> [options]\n",
+	printf("%s <join|leave|kill|expected|votes|version|wait|status|nodes|services|debug> [options]\n",
 	       prog_name);
 	printf("\n");
 	printf("Options:\n");
@@ -56,6 +57,7 @@ static void print_usage(int subcmd)
 		printf("  -e <votes>       Number of expected votes for the cluster\n");
 		printf("  -p <port>        UDP port number for cman communications\n");
 		printf("  -n <nodename>    The name of this node (defaults to hostname)\n");
+		printf("  -c <clustername> The name of the cluster to join\n");
 		printf("  -N <id>          Node id\n");
 		printf("  -w               Wait until node has joined a cluster\n");
 		printf("  -q               Wait until the cluster is quorate\n");
@@ -108,6 +110,8 @@ static void print_usage(int subcmd)
 		printf("nodes              Show local record of cluster nodes\n");
 		printf("  -f                 Also show when node was last fenced\n");
 		printf("  -a                 Also show node address(es)\n");
+		printf("  -n <nodename>      Only show information for specific node\n");
+		printf("  -F <format>        Specify output format (see man page)\n");
 		printf("\n");
 	}
 
@@ -197,7 +201,8 @@ static void show_status(void)
 	char tmpbuf[1024];
 	cman_extra_info_t *einfo = (cman_extra_info_t *)info_buf;
 	int quorate;
-	int i,j;
+	int i;
+	int j;
 	int portnum;
 	char *addrptr;
 
@@ -232,6 +237,10 @@ static void show_status(void)
 		printf(" Shutdown");
 	if (einfo->ei_flags & CMAN_EXTRA_FLAG_ERROR)
 		printf(" Error");
+	if (einfo->ei_flags & CMAN_EXTRA_FLAG_DISALLOWED)
+		printf(" DisallowedNodes");
+	if (einfo->ei_flags & CMAN_EXTRA_FLAG_DIRTY)
+		printf(" Dirty");
 	printf(" \n");
 
 	printf("Ports Bound: ");
@@ -268,6 +277,23 @@ static void show_status(void)
 	}
 	printf("\n");
 
+	if (einfo->ei_flags & CMAN_EXTRA_FLAG_DISALLOWED) {
+		int count;
+		int numnodes;
+		cman_node_t *nodes;
+
+		count = cman_get_node_count(h);
+		nodes = malloc(sizeof(cman_node_t) * count);
+
+		if (cman_get_disallowed_nodes(h, count, &numnodes, nodes) == 0) {
+			printf("Disallowed nodes: ");
+			for (i=0; i<numnodes; i++) {
+				printf("%s ", nodes[i].cn_name);
+			}
+			printf("\n");
+		}
+	}
+	cman_finish(h);
 }
 
 static int node_compare(const void *va, const void *vb)
@@ -277,12 +303,47 @@ static int node_compare(const void *va, const void *vb)
 	return a->cn_nodeid - b->cn_nodeid;
 }
 
+static int node_filter(commandline_t *comline, const char *node)
+{
+	int i;
+
+	for (i = 0; i < comline->num_nodenames; i++) {
+		if (strcmp(comline->nodenames[i], node) == 0) {
+			return TRUE;
+		}
+	}
+
+	return FALSE;
+}
+
+static int get_format_opt(const char *opt)
+{
+	if (!opt)
+		return FMT_NONE;
+
+	if (!strcmp(opt, "id"))
+		return FMT_ID;
+	if (!strcmp(opt, "name"))
+		return FMT_NAME;
+	if (!strcmp(opt, "type"))
+		return FMT_TYPE;
+	if (!strcmp(opt, "addr"))
+		return FMT_ADDR;
+
+	return FMT_NONE;
+}
+
 static void show_nodes(commandline_t *comline)
 {
 	cman_handle_t h;
 	int count;
 	int i;
+	int j;
+	int k;
 	int numnodes;
+	int dis_count;
+	int format[MAX_FORMAT_OPTS];
+	cman_node_t *dis_nodes;
 	cman_node_t *nodes;
 	struct tm *jtime;
 	struct tm *ftime;
@@ -300,14 +361,67 @@ static void show_nodes(commandline_t *comline)
 	if (!nodes)
 		die("cannot allocate memory for nodes list\n");
 
+	if (comline->format_opts != NULL) {
+		char *format_str = comline->format_opts;
+		char *format_tmp;
+		for (i = 0; i < MAX_FORMAT_OPTS; i++) {
+			format_tmp = strtok(format_str, ",");
+			format_str = NULL;
+			format[i] = get_format_opt(format_tmp);
+		}
+	}
+
 	if (cman_get_nodes(h, count, &numnodes, nodes) < 0)
 		die("cman_get_nodes failed: %s", cman_error(errno));
 
+
+	/* Get Disallowed nodes, so we can show them as such */
+	dis_nodes = malloc(sizeof(cman_node_t) * count);
+
+	if (cman_get_disallowed_nodes(h, count, &dis_count, dis_nodes) == 0) {
+		for (i = 0; i < numnodes; i++) {
+			for (j = 0; j < dis_count; j++) {
+				if (dis_nodes[j].cn_nodeid == nodes[i].cn_nodeid)
+					nodes[i].cn_member = 2;
+			}
+		}
+	}
+
 	/* Sort by nodeid to be friendly */
 	qsort(nodes, numnodes, sizeof(cman_node_t), node_compare);
 
-	printf("Node  Sts   Inc   Joined               Name\n");
-	for (i=0; i<numnodes; i++) {
+	if (dis_count) {
+		printf("NOTE: There are %d disallowed nodes,\n", dis_count);
+		printf("      members list may seem inconsistent across the cluster\n");
+	}
+
+	if (!comline->format_opts) {
+		printf("Node  Sts   Inc   Joined               Name\n");
+	}
+
+	for (i = 0; i < numnodes; i++) {
+		char member_type;
+
+		if (comline->num_nodenames > 0) {
+			if (node_filter(comline, nodes[i].cn_name) == 0) {
+				continue;
+			}
+		}
+
+		switch (nodes[i].cn_member) {
+		case 0:
+			member_type = 'X';
+			break;
+		case 1:
+			member_type = 'M';
+			break;
+		case 2:
+			member_type = 'd';
+			break;
+		default:
+			member_type = '?';
+			break;
+		}
 
 		jtime = localtime(&nodes[i].cn_jointime.tv_sec);
 		if (nodes[i].cn_jointime.tv_sec && nodes[i].cn_member)
@@ -315,11 +429,13 @@ static void show_nodes(commandline_t *comline)
 		else
 			strcpy(jstring, "                   ");
 
-		printf("%4d   %c  %5d   %s  %s\n",
-		       nodes[i].cn_nodeid, nodes[i].cn_member?'M':'X',
-		       nodes[i].cn_incarnation, jstring, nodes[i].cn_name);
+		if (!comline->format_opts) {
+			printf("%4d   %c  %5d   %s  %s\n",
+			       nodes[i].cn_nodeid, member_type,
+			       nodes[i].cn_incarnation, jstring, nodes[i].cn_name);
+		}
 
-		if (comline->fence_opt) {
+		if (comline->fence_opt && !comline->format_opts) {
 			char agent[255];
 			uint64_t fence_time;
 			int fenced;
@@ -336,31 +452,65 @@ static void show_nodes(commandline_t *comline)
 				}
 			}
 		}
+		
+		int numaddrs;
+		struct cman_node_address addrs[MAX_INTERFACES];
 
-		if (comline->addresses_opt)
-		{
-			int numaddrs;
-			struct cman_node_address addrs[MAX_INTERFACES];
+		if (comline->addresses_opt || comline->format_opts) {
 			if (!cman_get_node_addrs(h, nodes[i].cn_nodeid, MAX_INTERFACES, &numaddrs, addrs) &&
 				numaddrs)
 			{
-				int i;
-				printf("       Addresses: ");
-				for (i=0; i<numaddrs; i++)
-				{
-					print_address(addrs[i].cna_address);
-					printf(" ");
+				if (!comline->format_opts) {
+					printf("       Addresses: ");
+					for (i = 0; i < numaddrs; i++)
+					{
+						print_address(addrs[i].cna_address);
+						printf(" ");
+					}
+					printf("\n");
 				}
-				printf("\n");
 			}
 		}
+
+		if (comline->format_opts) {
+			for (j = 0; j < MAX_FORMAT_OPTS; j++) {
+				switch (format[j]) {
+				case FMT_NONE:
+					break;
+				case FMT_ID:
+					printf("%d ", nodes[i].cn_nodeid);
+					break;
+				case FMT_NAME:
+					printf("%s ", nodes[i].cn_name);
+					break;
+				case FMT_TYPE:
+					printf("%c ", member_type);
+					break;
+				case FMT_ADDR:
+					for (k = 0; k < numaddrs; k++) {
+						print_address(addrs[k].cna_address);
+						if (k != (numaddrs - 1)) {
+							printf(",");
+						}
+					}
+					printf(" ");
+					break;
+				default:
+					break;
+				}
+			}
+			printf("\n");
+		}
 	}
+
 	free(nodes);
+	free(dis_nodes);
+	cman_finish(h);
 }
 
 static void show_services(void)
 {
-	system("group_tool");
+	system("group_tool ls");
 }
 
 
@@ -530,6 +680,18 @@ static void kill_node(commandline_t *comline)
 	cman_finish(h);
 }
 
+static void set_debuglog(commandline_t *comline)
+{
+	cman_handle_t h;
+
+	h = open_cman_handle(1);
+
+	if (cman_set_debuglog(h, comline->verbose))
+		perror("setting debuglog failed");
+
+	cman_finish(h);
+}
+
 
 static int get_int_arg(char argopt, char *arg)
 {
@@ -627,9 +789,13 @@ static void decode_arguments(int argc, char *argv[], commandline_t *comline)
 			comline->clustername_opt = TRUE;
 			break;
 
+		case 'F':
+			comline->format_opts = strdup(optarg);
+			break;
+
 		case 'V':
 			printf("cman_tool %s (built %s %s)\n",
-				CMAN_RELEASE_NAME, __DATE__, __TIME__);
+				RELEASE_VERSION, __DATE__, __TIME__);
 			printf("%s\n", REDHAT_COPYRIGHT);
 			exit(EXIT_SUCCESS);
 			break;
@@ -714,6 +880,10 @@ static void decode_arguments(int argc, char *argv[], commandline_t *comline)
 			if (comline->operation)
 				die("can't specify two operations");
 			comline->operation = OP_SERVICES;
+		} else if (strcmp(argv[optind], "debug") == 0) {
+			if (comline->operation)
+				die("can't specify two operations");
+			comline->operation = OP_DEBUG;
 		} else if (strcmp(argv[optind], "remove") == 0) {
 			comline->remove = TRUE;
 		} else if (strcmp(argv[optind], "force") == 0) {
@@ -828,6 +998,10 @@ int main(int argc, char *argv[])
 	case OP_SERVICES:
 		show_services();
 		break;
+
+	case OP_DEBUG:
+		set_debuglog(&comline);
+		break;
 	}
 
 	exit(EXIT_SUCCESS);
