@@ -28,6 +28,7 @@
 #include <netdb.h>
 
 /* openais headers */
+#include <openais/service/swab.h>
 #include <openais/totem/totemip.h>
 #include <openais/totem/totempg.h>
 #include <openais/totem/aispoll.h>
@@ -35,7 +36,7 @@
 #include <openais/service/config.h>
 #include <openais/lcr/lcr_comp.h>
 #include <openais/service/swab.h>
-#include <openais/service/print.h>
+#include <openais/service/logsys.h>
 
 #include "cnxman-socket.h"
 #include "commands.h"
@@ -57,6 +58,7 @@ struct totem_ip_address mcast_addr[MAX_INTERFACES];
 struct totem_ip_address ifaddrs[MAX_INTERFACES];
 int num_interfaces;
 uint64_t incarnation;
+int num_ais_nodes;
 
 static int config_run;
 static char errorstring[512];
@@ -133,7 +135,7 @@ static inline void objdb_get_int(struct objdb_iface_ver0 *objdb, unsigned int ob
  * Exports the interface for the service
  */
 static struct openais_service_handler cman_service_handler = {
-	.name		    		= (unsigned char *)"openais CMAN membership service 2.01",
+	.name		    		= (char *)"openais CMAN membership service 2.01",
 	.id			        = CMAN_SERVICE,
 	.flow_control			= OPENAIS_FLOW_CONTROL_NOT_REQUIRED,
 	.lib_exit_fn		       	= cman_exit_fn,
@@ -202,9 +204,7 @@ static int cman_readconfig(struct objdb_iface_ver0 *objdb, char **error_string)
 
 	/* Initialise early logging */
 	if (getenv("CMAN_DEBUGLOG"))
-	{
 		debug_mask = atoi(getenv("CMAN_DEBUGLOG"));
-	}
 
 	if (getenv("CMAN_PIPE"))
 		startup_pipe = atoi(getenv("CMAN_PIPE"));
@@ -240,7 +240,7 @@ static int cman_exec_init_fn(struct objdb_iface_ver0 *objdb)
 {
 	unsigned int object_handle;
 
-	/* We can only work if our config inerface was run first */
+	/* We can only work if our config interface was run first */
 	if (!config_run)
 		return 0;
 
@@ -267,6 +267,8 @@ static int cman_exec_init_fn(struct objdb_iface_ver0 *objdb)
 	close(startup_pipe);
 	startup_pipe = 0;
 
+	cman_flush_debuglog();
+
 	/* Start totem */
 	totempg_groups_initialize(&group_handle, cman_deliver_fn, cman_confchg_fn);
 	totempg_groups_join(group_handle, cman_group, 1);
@@ -285,6 +287,7 @@ int cman_exit_fn(void *conn_info)
 
 int ais_add_ifaddr(char *mcast, char *ifaddr, int portnum)
 {
+	struct totem_ip_address localhost;
 	unsigned int totem_object_handle;
 	unsigned int interface_object_handle;
 	char tmp[132];
@@ -321,10 +324,19 @@ int ais_add_ifaddr(char *mcast, char *ifaddr, int portnum)
 			if (!ret)
 				ret = totemip_parse(&ifaddrs[num_interfaces], ifaddr,
 						    mcast_addr[num_interfaces].family);
-			if (!ret)
-				num_interfaces++;
-			else
+			if (ret) {
+				errno = EPROTOTYPE;
 				return ret;
+			}
+
+			/* Check it's not bound to localhost, sigh */
+			totemip_localhost(mcast_addr[num_interfaces].family, &localhost);
+			if (totemip_equal(&localhost, &ifaddrs[num_interfaces])) {
+				errno = EADDRINUSE;
+				return -1;
+			}
+
+			num_interfaces++;
 		}
 	}
 
@@ -401,12 +413,15 @@ static void cman_confchg_fn(enum totem_configuration_type configuration_type,
 	P_AIS("confchg_fn called type = %d, seq=%lld\n", configuration_type, ring_id->seq);
 
 	incarnation = ring_id->seq;
+	num_ais_nodes = member_list_entries;
 
 	/* Tell the cman membership layer */
 	for (i=0; i<left_list_entries; i++)
 		del_ais_node(left_list[i]);
-	for (i=0; i<joined_list_entries; i++)
-		add_ais_node(joined_list[i], incarnation, member_list_entries);
+
+	/* Joining nodes are only added after a valid TRANSITION message
+	 * is received.
+	 */
 
 	/* Save the left list for later so we can do a consolidated confchg message */
 	if (configuration_type == TOTEM_CONFIGURATION_TRANSITIONAL) {
@@ -456,6 +471,8 @@ static int comms_init_ais(struct objdb_iface_ver0 *objdb)
 			       "totem", strlen("totem"),
 			       &object_handle) == 0)
 	{
+		char *value;
+
 		objdb->object_key_create(object_handle, "version", strlen("version"),
 					 "2", 2);
 
@@ -466,6 +483,29 @@ static int comms_init_ais(struct objdb_iface_ver0 *objdb)
 		objdb->object_key_create(object_handle, "vsftype", strlen("vsftype"),
 					 "none", strlen("none")+1);
 
+		/* Set the token timeout is 10 seconds, but don't overrride anything that
+		   might be in cluster.conf */
+		if (objdb_get_string(objdb, object_handle, "token", &value)) {
+			global_objdb->object_key_create(object_handle, "token", strlen("token"),
+							"10000", strlen("10000")+1);
+		}
+		if (objdb_get_string(objdb, object_handle, "token_retransmits_before_loss_const", &value)) {
+			global_objdb->object_key_create(object_handle, "token_retransmits_before_loss_const",
+							strlen("token_retransmits_before_loss_const"),
+							"20", strlen("20")+1);
+		}
+
+		/* Extend consensus & join timeouts per bz#214290 */
+		if (objdb_get_string(objdb, object_handle, "join", &value)) {
+			global_objdb->object_key_create(object_handle, "join", strlen("join"),
+							"60", strlen("60")+1);
+		}
+		if (objdb_get_string(objdb, object_handle, "consensus", &value)) {
+			global_objdb->object_key_create(object_handle, "consensus", strlen("consensus"),
+							"4800", strlen("4800")+1);
+		}
+
+
 		/* Set RRP mode appropriately */
 		if (num_interfaces > 1) {
 			global_objdb->object_key_create(object_handle, "rrp_mode", strlen("rrp_mode"),
@@ -476,9 +516,11 @@ static int comms_init_ais(struct objdb_iface_ver0 *objdb)
 							"none", strlen("none")+1);
 		}
 
-		sprintf(tmp, "%d", 1);
-		objdb->object_key_create(object_handle, "secauth", strlen("secauth"),
-					 tmp, strlen(tmp)+1);
+		if (objdb_get_string(objdb, object_handle, "secauth", &value)) {
+			sprintf(tmp, "%d", 1);
+			objdb->object_key_create(object_handle, "secauth", strlen("secauth"),
+						 tmp, strlen(tmp)+1);
+		}
 
 		if (key_filename)
 		{
@@ -486,7 +528,7 @@ static int comms_init_ais(struct objdb_iface_ver0 *objdb)
 						 key_filename, strlen(key_filename)+1);
 		}
 		else /* Use the cluster name as key,
-		      * This isn't a good isolation strategey but it does make sure that
+		      * This isn't a good isolation strategy but it does make sure that
 		      * clusters on the same port/multicast by mistake don't actually interfere
 		      * and that we have some form of encryption going.
 		      */
@@ -513,7 +555,7 @@ static int comms_init_ais(struct objdb_iface_ver0 *objdb)
 		char *logstr;
 
 		/* Default logging facility is "local4" unless overridden by the user */
-		if (!objdb_get_string(objdb, object_handle, "syslog_facility", &logstr)) {
+		if (objdb_get_string(objdb, object_handle, "syslog_facility", &logstr)) {
 			objdb->object_key_create(object_handle, "syslog_facility", strlen("syslog_facility"),
 						 "local4", strlen("local4")+1);
 		}
