@@ -2,7 +2,7 @@
 *******************************************************************************
 **
 **  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
-**  Copyright (C) 2004-2006 Red Hat, Inc.  All rights reserved.
+**  Copyright (C) 2004-2007 Red Hat, Inc.  All rights reserved.
 **
 **  This copyrighted material is made available to anyone wishing to use,
 **  modify, copy, or redistribute it subject to the terms and conditions
@@ -30,9 +30,12 @@
 #include <netinet/in.h>
 #include <sys/errno.h>
 
+#include <openais/service/swab.h>
 #include <openais/totem/totemip.h>
 #include <openais/totem/totempg.h>
 #include <openais/service/swab.h>
+#include <openais/service/logsys.h>
+#include <openais/service/timer.h>
 #include <openais/totem/aispoll.h>
 #include "list.h"
 #include "cnxman-socket.h"
@@ -60,7 +63,7 @@ static LIST_INIT(cluster_members_list);
        unsigned int config_version;
 static struct cluster_node *us;
 static int quorum;
-static int two_node;
+extern int two_node;
        unsigned int quorumdev_poll=10000;
        unsigned int shutdown_timeout=5000;
        unsigned int ccsd_poll_interval=1000;
@@ -68,18 +71,18 @@ static int cluster_is_quorate;
        char cluster_name[MAX_CLUSTER_NAME_LEN+1];
 static char nodename[MAX_CLUSTER_MEMBER_NAME_LEN+1];
 static int wanted_nodeid;
-extern poll_handle ais_poll_handle;
 static struct cluster_node *quorum_device;
 static uint16_t cluster_id;
 static int ais_running;
-static poll_timer_handle quorum_device_timer;
+static time_t join_time;
+static openais_timer_handle quorum_device_timer;
 
 /* If CCS gets out of sync, we poll it until it isn't */
-static poll_timer_handle ccsd_timer;
+static openais_timer_handle ccsd_timer;
 static unsigned int wanted_config_version;
 static int config_error;
 
-static poll_timer_handle shutdown_timer;
+static openais_timer_handle shutdown_timer;
 static struct connection *shutdown_con;
 static uint32_t shutdown_flags;
 static int shutdown_yes;
@@ -93,7 +96,9 @@ static int get_highest_nodeid(void);
 static int send_port_open_msg(unsigned char port);
 static int send_port_enquire(int nodeid);
 static void process_internal_message(char *data, int len, int nodeid, int byteswap);
-static void recalculate_quorum(int allow_decrease);
+static void recalculate_quorum(int allow_decrease, int by_current_nodes);
+static void send_kill(int nodeid, uint16_t reason);
+static char *killmsg_reason(int reason);
 
 static void set_port_bit(struct cluster_node *node, uint8_t port)
 {
@@ -128,6 +133,18 @@ static int get_port_bit(struct cluster_node *node, uint8_t port)
 	return ((node->port_bits[byte] & (1<<bit)) != 0);
 }
 
+static int have_disallowed(void)
+{
+	struct cluster_node *node;
+
+	list_iterate_items(node, &cluster_members_list) {
+		if (node->state == NODESTATE_AISONLY)
+			return 1;
+	}
+
+	return 0;
+}
+
 /* If "cluster_is_quorate" is 0 then all activity apart from protected ports is
  * blocked. */
 static void set_quorate(int total_votes)
@@ -142,9 +159,21 @@ static void set_quorate(int total_votes)
 	}
 
 	if (cluster_is_quorate && !quorate)
-		log_msg(LOG_INFO, "quorum lost, blocking activity\n");
+		log_printf(LOG_INFO, "quorum lost, blocking activity\n");
 	if (!cluster_is_quorate && quorate)
-		log_msg(LOG_INFO, "quorum regained, resuming activity\n");
+		log_printf(LOG_INFO, "quorum regained, resuming activity\n");
+
+	/* If we are newly quorate, then kill any AISONLY nodes */
+	if (!cluster_is_quorate && quorate) {
+		struct cluster_node *node = NULL;
+		struct list *tmp;
+
+		list_iterate(tmp, &cluster_members_list) {
+			node = list_item(tmp, struct cluster_node);
+			if (node->state == NODESTATE_AISONLY)
+				send_kill(node->node_id, CLUSTER_KILL_REJOIN);
+		}
+	}
 
 	cluster_is_quorate = quorate;
 
@@ -298,11 +327,11 @@ static int calculate_quorum(int allow_decrease, int max_expected, unsigned int *
 }
 
 /* Recalculate cluster quorum, set quorate and notify changes */
-static void recalculate_quorum(int allow_decrease)
+static void recalculate_quorum(int allow_decrease, int by_current_nodes)
 {
 	unsigned int total_votes;
 
-	quorum = calculate_quorum(allow_decrease, 0, &total_votes);
+	quorum = calculate_quorum(allow_decrease, by_current_nodes?cluster_members:0, &total_votes);
 	set_quorate(total_votes);
 	notify_listeners(NULL, EVENT_REASON_STATECHANGE, 0);
 }
@@ -375,7 +404,7 @@ int cman_join_cluster(char *name, unsigned short cl_id, int two_node_flag, int e
 	ais_running = 1;
 
 	if (read_ccs_nodes(&config_version, 1)) {
-		log_msg(LOG_ERR, "Can't initialise list of nodes from CCS\n");
+		log_printf(LOG_ERR, "Can't initialise list of nodes from CCS\n");
 		return -EINVAL;
 	}
 
@@ -386,6 +415,7 @@ int cman_join_cluster(char *name, unsigned short cl_id, int two_node_flag, int e
 		strcpy(nodename, un.nodename);
 	}
 
+	time(&join_time);
 	us = add_new_node(nodename, wanted_nodeid, -1, expected_votes,
 			  NODESTATE_MEMBER);
 	set_port_bit(us, 0);
@@ -424,6 +454,7 @@ static int do_cmd_get_extrainfo(char *cmdbuf, char **retbuf, int retsize, int *r
 	int total_votes = 0;
 	int max_expected = 0;
 	int addrlen;
+	int uncounted = 0;
 	struct cluster_node *node;
 	struct sockaddr_storage *ss;
 	char *ptr;
@@ -437,6 +468,8 @@ static int do_cmd_get_extrainfo(char *cmdbuf, char **retbuf, int retsize, int *r
 			total_votes += node->votes;
 			max_expected = max(max_expected, node->expected_votes);
 		}
+		if (node->state == NODESTATE_AISONLY)
+			uncounted = 1;
 	}
 	if (quorum_device && quorum_device->state == NODESTATE_MEMBER)
 		total_votes += quorum_device->votes;
@@ -467,6 +500,10 @@ static int do_cmd_get_extrainfo(char *cmdbuf, char **retbuf, int retsize, int *r
 		einfo->flags |= CMAN_EXTRA_FLAG_ERROR;
 	if (shutdown_con)
 		einfo->flags |= CMAN_EXTRA_FLAG_SHUTDOWN;
+	if (uncounted)
+		einfo->flags |= CMAN_EXTRA_FLAG_UNCOUNTED;
+	if (us->flags & NODE_FLAGS_DIRTY)
+		einfo->flags |= CMAN_EXTRA_FLAG_DIRTY;
 
 	ptr = einfo->addresses;
 	for (i=0; i<num_interfaces; i++) {
@@ -585,9 +622,22 @@ static int do_cmd_set_expected(char *cmdbuf, int *retlen)
 	unsigned int total_votes;
 	unsigned int newquorum;
 	unsigned int newexp;
+	struct cluster_node *node = NULL;
+	struct list *tmp;
 
 	if (!we_are_a_cluster_member)
 		return -ENOENT;
+
+	/* If there are any AISONLY nodes then we can't allow
+	   the user to set expected votes as it may destroy data */
+	list_iterate(tmp, &cluster_members_list) {
+		node = list_item(tmp, struct cluster_node);
+		if (node->state == NODESTATE_AISONLY) {
+			log_printf(LOG_NOTICE, "Attempt to set expected votes when cluster has AISONLY nodes in it.");
+			return -EINVAL;
+		}
+	}
+
 	memcpy(&newexp, cmdbuf, sizeof(int));
 	newquorum = calculate_quorum(1, newexp, &total_votes);
 
@@ -647,7 +697,7 @@ static int do_cmd_kill_node(char *cmdbuf, int *retlen)
 	if ((node = find_node_by_nodeid(nodeid)) == NULL)
 		return -EINVAL;
 
-	if (node->state != NODESTATE_MEMBER)
+	if (node->state != NODESTATE_MEMBER && node->state != NODESTATE_AISONLY)
 		return -EINVAL;
 
 	node->leave_reason = CLUSTER_LEAVEFLAG_KILLED;
@@ -741,7 +791,7 @@ static int do_cmd_set_votes(char *cmdbuf, int *retlen)
 		return -EINVAL;
 	}
 
-	recalculate_quorum(1);
+	recalculate_quorum(1, 0);
 
 	send_reconfigure(arg.nodeid, RECONFIG_PARAM_NODE_VOTES, arg.newvotes);
 
@@ -811,7 +861,7 @@ static void check_shutdown_status()
 	/* All replies safely gathered in ? */
 	if (shutdown_yes + shutdown_no >= shutdown_expected) {
 
-		poll_timer_delete(ais_poll_handle, shutdown_timer);
+		openais_timer_delete(shutdown_timer);
 
 		if (shutdown_yes >= shutdown_expected ||
 		    shutdown_flags & SHUTDOWN_ANYWAY) {
@@ -823,13 +873,13 @@ static void check_shutdown_status()
 		}
 		else {
 			reply = -EBUSY;
+
+			/* Tell originator that shutdown was cancelled */
+			send_status_return(shutdown_con, CMAN_CMD_TRY_SHUTDOWN, reply);
+			shutdown_con = NULL;
 		}
 
 		P_MEMB("shutdown decision is: %d (yes=%d, no=%d) flags=%x\n", reply, shutdown_yes, shutdown_no, shutdown_flags);
-
-		/* Tell originator what we decided */
-		send_status_return(shutdown_con, CMAN_CMD_TRY_SHUTDOWN, reply);
-		shutdown_con = NULL;
 	}
 }
 
@@ -896,8 +946,8 @@ static int do_cmd_try_shutdown(struct connection *con, char *cmdbuf)
 
 		/* Start the timer. If we don't get a full set of replies before this goes
 		   off we'll cancel the shutdown */
-		poll_timer_add(ais_poll_handle, shutdown_timeout, NULL,
-			       shutdown_timer_fn, &shutdown_timer);
+		openais_timer_add_duration((unsigned long long)shutdown_timeout*1000000, NULL,
+					   shutdown_timer_fn, &shutdown_timer);
 
 		notify_listeners(NULL, EVENT_REASON_TRY_SHUTDOWN, flags);
 
@@ -947,7 +997,7 @@ static int do_cmd_register_quorum_device(char *cmdbuf, int *retlen)
         /* Keep this list valid so it doesn't confuse other code */
         list_init(&quorum_device->addr_list);
 
-	log_msg(LOG_INFO, "quorum device registered\n");
+	log_printf(LOG_INFO, "quorum device registered\n");
         return 0;
 }
 
@@ -964,7 +1014,7 @@ static int do_cmd_unregister_quorum_device(char *cmdbuf, int *retlen)
 
         quorum_device = NULL;
 
-	log_msg(LOG_INFO, "quorum device unregistered\n");
+	log_printf(LOG_INFO, "quorum device unregistered\n");
         return 0;
 }
 
@@ -972,19 +1022,19 @@ static void ccsd_timer_fn(void *arg)
 {
 	int ccs_err;
 
-	log_msg(LOG_DEBUG, "Polling ccsd for updated information\n");
+	log_printf(LOG_DEBUG, "Polling ccsd for updated information\n");
 	ccs_err = read_ccs_nodes(&config_version, 0);
 	if (ccs_err || config_version < wanted_config_version) {
-		log_msg(LOG_ERR, "Can't read CCS to get updated config version %d. Activity suspended on this node\n",
+		log_printf(LOG_ERR, "Can't read CCS to get updated config version %d. Activity suspended on this node\n",
 				wanted_config_version);
 
-		poll_timer_add(ais_poll_handle, ccsd_poll_interval, NULL,
-			       ccsd_timer_fn, &ccsd_timer);
+		openais_timer_add_duration((unsigned long long)ccsd_poll_interval*1000000, NULL,
+					   ccsd_timer_fn, &ccsd_timer);
 	}
 	else {
-		log_msg(LOG_ERR, "Now got CCS information version %d, continuing\n", config_version);
+		log_printf(LOG_ERR, "Now got CCS information version %d, continuing\n", config_version);
 		config_error = 0;
-		recalculate_quorum(0);
+		recalculate_quorum(0, 0);
 	}
 }
 
@@ -996,14 +1046,14 @@ static void quorum_device_timer_fn(void *arg)
 		return;
 
 	gettimeofday(&now, NULL);
-	if (quorum_device->last_hello.tv_sec + quorumdev_poll < now.tv_sec) {
+	if (quorum_device->last_hello.tv_sec + quorumdev_poll/1000 < now.tv_sec) {
 		quorum_device->state = NODESTATE_DEAD;
-		log_msg(LOG_INFO, "lost contact with quorum device\n");
-		recalculate_quorum(0);
+		log_printf(LOG_INFO, "lost contact with quorum device\n");
+		recalculate_quorum(0, 0);
 	}
 	else {
-		poll_timer_add(ais_poll_handle, quorumdev_poll, quorum_device,
-			       quorum_device_timer_fn, &quorum_device_timer);
+		openais_timer_add_duration((unsigned long long)quorumdev_poll*1000000, quorum_device,
+					   quorum_device_timer_fn, &quorum_device_timer);
 	}
 }
 
@@ -1020,17 +1070,17 @@ static int do_cmd_poll_quorum_device(char *cmdbuf, int *retlen)
 		gettimeofday(&quorum_device->last_hello, NULL);
                 if (quorum_device->state == NODESTATE_DEAD) {
                         quorum_device->state = NODESTATE_MEMBER;
-                        recalculate_quorum(0);
+                        recalculate_quorum(0, 0);
 
-			poll_timer_add(ais_poll_handle, quorumdev_poll, quorum_device,
-				       quorum_device_timer_fn, &quorum_device_timer);
+			openais_timer_add_duration((unsigned long long)quorumdev_poll*1000000, quorum_device,
+						   quorum_device_timer_fn, &quorum_device_timer);
                 }
         }
         else {
                 if (quorum_device->state == NODESTATE_MEMBER) {
                         quorum_device->state = NODESTATE_DEAD;
-                        recalculate_quorum(0);
-			poll_timer_delete(ais_poll_handle, quorum_device_timer);
+                        recalculate_quorum(0, 0);
+			openais_timer_delete(quorum_device_timer);
                 }
         }
 
@@ -1042,7 +1092,7 @@ static int do_cmd_update_fence_info(char *cmdbuf)
 {
 	struct cl_fence_info *f = (struct cl_fence_info *)cmdbuf;
 	struct cluster_node *node;
-	char msg[sizeof(struct cl_fencemsg)+strlen(f->fence_agent)];
+	char msg[sizeof(struct cl_fencemsg)+strlen(f->fence_agent)+1];
 	struct cl_fencemsg *fence_msg = (struct cl_fencemsg *)msg;
 
 	node = find_node_by_nodeid(f->nodeid);
@@ -1139,6 +1189,7 @@ int process_command(struct connection *con, int cmd, char *cmdbuf,
 	int err = -EINVAL;
 	struct cl_version cnxman_version;
 	char *outbuf = *retbuf;
+	int value;
 
 	P_MEMB("command to process is %x\n", cmd);
 
@@ -1159,6 +1210,17 @@ int process_command(struct connection *con, int cmd, char *cmdbuf,
 		err = 0;
 		break;
 
+	case CMAN_CMD_SET_DIRTY:
+		us->flags |= NODE_FLAGS_DIRTY;
+		err = 0;
+		break;
+
+	case CMAN_CMD_SET_DEBUGLOG:
+		memcpy(&value, cmdbuf, sizeof(int));
+		set_debuglog(value);
+		err = 0;
+		break;
+
 	case CMAN_CMD_START_CONFCHG:
 		con->confchg = 1;
 		err = 0;
@@ -1382,7 +1444,7 @@ void unbind_con(struct connection *con)
 	   then we take its closedown as a "Yes" to the "can we shutdown"
 	   question. If it previously answered "No", we need to change its vote */
 	if (shutdown_con && con->events) {
-		if (!con->shutdown_reply) {
+		if (con->shutdown_reply) {
 			if (con->shutdown_reply == SHUTDOWN_REPLY_YES)
 				shutdown_yes--;
 			if (con->shutdown_reply == SHUTDOWN_REPLY_NO)
@@ -1416,20 +1478,20 @@ int our_nodeid()
 static int valid_transition_msg(int nodeid, struct cl_transmsg *msg)
 {
 	if (strcmp(msg->clustername, cluster_name) != 0) {
-		log_msg(LOG_ERR, "Node %d conflict, remote cluster name='%s', local='%s'\n",
+		log_printf(LOG_ERR, "Node %d conflict, remote cluster name='%s', local='%s'\n",
 			nodeid, msg->clustername, cluster_name);
 		return -1;
 	}
 
 	if (msg->cluster_id != cluster_id) {
-		log_msg(LOG_ERR, "Node %d conflict, remote cluster id=%d, local=%d\n",
+		log_printf(LOG_ERR, "Node %d conflict, remote cluster id=%d, local=%d\n",
 			nodeid, msg->cluster_id, cluster_id);
 		return -1;
 	}
 
 	if (msg->major_version != CNXMAN_MAJOR_VERSION) {
 
-		log_msg(LOG_ERR, "Node %d conflict, remote version id=%d, local=%d\n",
+		log_printf(LOG_ERR, "Node %d conflict, remote version id=%d, local=%d\n",
 			nodeid, msg->major_version, CNXMAN_MAJOR_VERSION);
 		return -1;
 	}
@@ -1441,22 +1503,22 @@ static int valid_transition_msg(int nodeid, struct cl_transmsg *msg)
 		ccs_err = read_ccs_nodes(&config_version, 0);
 		if (ccs_err || config_version < msg->config_version) {
 			config_error = 1;
-			log_msg(LOG_ERR, "Can't read CCS to get updated config version %d. Activity suspended on this node\n",
+			log_printf(LOG_ERR, "Can't read CCS to get updated config version %d. Activity suspended on this node\n",
 				msg->config_version);
 
 			wanted_config_version = msg->config_version;
-			poll_timer_add(ais_poll_handle, ccsd_poll_interval, NULL,
-				       ccsd_timer_fn, &ccsd_timer);
+			openais_timer_add_duration((unsigned long long)ccsd_poll_interval*1000000, NULL,
+						   ccsd_timer_fn, &ccsd_timer);
 		}
 		if (config_version > msg->config_version) {
 			// TODO tell everyone else to update...
 		}
-		recalculate_quorum(0);
+		recalculate_quorum(0, 0);
 	}
 
 
 	if (msg->config_version != config_version) {
-		log_msg(LOG_ERR, "Node %d conflict, remote config version id=%d, local=%d\n",
+		log_printf(LOG_ERR, "Node %d conflict, remote config version id=%d, local=%d\n",
 			nodeid, msg->config_version, config_version);
 		return -1;
 	}
@@ -1485,6 +1547,7 @@ void send_transition_msg(int last_memb_count, int first_trans)
 	msg->config_version = config_version;
 	msg->flags = us->flags;
 	msg->fence_time = us->fence_time;
+	msg->join_time = join_time;
 	strcpy(msg->clustername, cluster_name);
 	if (us->fence_agent)
 	{
@@ -1497,6 +1560,9 @@ void send_transition_msg(int last_memb_count, int first_trans)
 		len += 1;
 	}
 
+	if (have_disallowed())
+		msg->flags |= NODE_FLAGS_SEESDISALLOWED;
+
 	comms_send_message(msg, len,
 			   0,0,
 			   0,  /* multicast */
@@ -1584,25 +1650,25 @@ static void do_reconfigure_msg(void *data)
 				node->expected_votes = msg->value;
 			}
 		}
-		recalculate_quorum(1);  /* Allow decrease */
+		recalculate_quorum(1, 0);  /* Allow decrease */
 		break;
 
 	case RECONFIG_PARAM_NODE_VOTES:
 		node->votes = msg->value;
-		recalculate_quorum(1);  /* Allow decrease */
+		recalculate_quorum(1, 0);  /* Allow decrease */
 		break;
 
 	case RECONFIG_PARAM_CONFIG_VERSION:
 		if (read_ccs_nodes(&config_version, 0)) {
-			log_msg(LOG_ERR, "Can't read CCS to get updated config version %d. Activity suspended on this node\n",
+			log_printf(LOG_ERR, "Can't read CCS to get updated config version %d. Activity suspended on this node\n",
 				msg->value);
 
 			config_error = 1;
-			recalculate_quorum(0);
+			recalculate_quorum(0, 0);
 
 			wanted_config_version = config_version;
-			poll_timer_add(ais_poll_handle, ccsd_poll_interval, NULL,
-				       ccsd_timer_fn, &ccsd_timer);
+			openais_timer_add_duration((unsigned long long)ccsd_poll_interval*1000000, NULL,
+						   ccsd_timer_fn, &ccsd_timer);
 		}
 		break;
 	}
@@ -1641,13 +1707,75 @@ static void do_process_transition(int nodeid, char *data, int len)
 		P_MEMB("Transition message from %d does not match current config - should quit ?\n", nodeid);
 		return; // PJC ???
 	}
+
+	/* If the remote node can see AISONLY nodes then we can't join as we don't
+	   know the full state */
+	if (msg->flags & NODE_FLAGS_SEESDISALLOWED && !have_disallowed()) {
+		/* Must use syslog directly here or the message will never arrive */
+		syslog(LOG_CRIT, "CMAN: Joined a cluster with disallowed nodes. must die");
+		exit(2);
+	}
+	msg->flags &= ~NODE_FLAGS_SEESDISALLOWED;
+
 	node = find_node_by_nodeid(nodeid);
 	assert(node);
 
-	if (node->flags & NODE_FLAGS_GOTTRANSITION) {
-
+	/* Newer nodes 6.1.0 onwards, set the DIRTY flag if they have state. If the new node has been down
+	   and has state then we mark it disallowed because we cannot merge stateful nodes */
+	if (msg->flags & NODE_FLAGS_DIRTY && node->flags & NODE_FLAGS_BEENDOWN) {
+		/* Don't duplicate messages */
+		if (node->state != NODESTATE_AISONLY) {
+			if (cluster_is_quorate) {
+				P_MEMB("Killing node %s because it has rejoined the cluster with existing state", node->name);
+				log_printf(LOG_CRIT, "Killing node %s because it has rejoined the cluster with existing state", node->name);
+				node->state = NODESTATE_AISONLY;
+				send_kill(nodeid, CLUSTER_KILL_REJOIN);
+			}
+			else {
+				P_MEMB("Node %s not joined to cman because it has existing state", node->name);
+				log_printf(LOG_CRIT, "Node %s not joined to cman because it has existing state", node->name);
+				node->state = NODESTATE_AISONLY;
+			}
+		}
+		return;
 	}
-	node->flags = msg->flags;
+
+        /* This is for older nodes. If the join_time of the node matches that already stored AND
+	   the node has been down, then we kill it as this must be a rejoin */
+	if (msg->join_time == node->cman_join_time && node->flags & NODE_FLAGS_BEENDOWN) {
+		/* Don't duplicate messages */
+		if (node->state != NODESTATE_AISONLY) {
+			if (cluster_is_quorate) {
+				P_MEMB("Killing node %s because it has rejoined the cluster without cman_tool join", node->name);
+				log_printf(LOG_CRIT, "Killing node %s because it has rejoined the cluster without cman_tool join", node->name);
+				node->state = NODESTATE_AISONLY;
+				send_kill(nodeid, CLUSTER_KILL_REJOIN);
+			}
+			else {
+				P_MEMB("Node %s not joined to cman because it has rejoined an inquorate cluster", node->name);
+				log_printf(LOG_CRIT, "Node %s not joined to cman because it has rejoined an inquorate cluster", node->name);
+				node->state = NODESTATE_AISONLY;
+			}
+		}
+		return;
+	}
+	else {
+		node->cman_join_time = msg->join_time;
+		add_ais_node(nodeid, incarnation, num_ais_nodes);
+	}
+
+	/* If the cluster already has some AISONLY nodes then we can't make
+	   sense of the membership. So the new node has to also be AISONLY
+	   until we are consistent again */
+	if (have_disallowed() && !node->us)
+		node->state = NODESTATE_AISONLY;
+
+	node->flags = msg->flags; /* This will clear the BEENDOWN flag of course */
+
+	/* Take into account any new expected_votes value that the new node has */
+	node->expected_votes = msg->expected_votes;
+	recalculate_quorum(0, 0);
+
 	if (node->fence_agent && msg->fence_agent[0] && strcmp(node->fence_agent, msg->fence_agent))
 	{
 		free(node->fence_agent);
@@ -1662,7 +1790,7 @@ static void do_process_transition(int nodeid, char *data, int len)
 	if (node->fence_time && !msg->fence_time &&
 	    node->fence_agent && !msg->fence_agent[0])
 	{
-		char msg[sizeof(struct cl_fencemsg)+strlen(node->fence_agent)];
+		char msg[sizeof(struct cl_fencemsg)+strlen(node->fence_agent)+1];
 		struct cl_fencemsg *fence_msg = (struct cl_fencemsg *)msg;
 
 		fence_msg->cmd = CLUSTER_MSG_FENCESTATUS;
@@ -1716,8 +1844,9 @@ static void process_internal_message(char *data, int len, int nodeid, int need_b
 
 	case CLUSTER_MSG_PORTSTATUS:
 		if (nodeid != us->node_id) {
-			P_MEMB("got PORTRESULT from %d, low bytes = %x %x\n", data[1], data[2]);
-			memcpy(node->port_bits, data+1, PORT_BITS_SIZE);
+			P_MEMB("got PORTRESULT from %d, low bytes = %x %x\n", nodeid, data[1], data[2]);
+			if (node)
+				memcpy(node->port_bits, data+1, PORT_BITS_SIZE);
 		}
 		break;
 
@@ -1730,7 +1859,8 @@ static void process_internal_message(char *data, int len, int nodeid, int need_b
 		killmsg = (struct cl_killmsg *)data;
 		P_MEMB("got KILL for node %d\n", killmsg->nodeid);
 		if (killmsg->nodeid == wanted_nodeid) {
-			log_msg(LOG_INFO, "cman killed by node %d for reason %d\n", nodeid, killmsg->reason);
+			log_printf(LOG_INFO, "cman killed by node %d because %s\n", nodeid,
+				killmsg_reason(killmsg->reason));
 			exit(1);
 		}
 		break;
@@ -1740,15 +1870,19 @@ static void process_internal_message(char *data, int len, int nodeid, int need_b
 		P_MEMB("got LEAVE from node %d, reason = %d\n", nodeid, leavemsg->reason);
 
 		/* We got our own leave message back. now quit */
-		if (node && node->node_id == us->node_id)
+		if (node && node->node_id == us->node_id) {
+			/* Tell whomever asked us to leave that we are now going down */
+			if (shutdown_con)
+				send_status_return(shutdown_con, CMAN_CMD_TRY_SHUTDOWN, 0);
 			exit(0);
+		}
 
 		/* Someone else, make a note of the reason for leaving */
 		if (node)
 			node->leave_reason = leavemsg->reason;
 
 		/* Mark it as leaving, and remove it when we get an AIS node down event for it */
-		if (node && node->state == NODESTATE_MEMBER)
+		if (node && (node->state == NODESTATE_MEMBER || node->state == NODESTATE_AISONLY))
 			node->state = NODESTATE_LEAVING;
 		break;
 
@@ -1767,7 +1901,7 @@ static void process_internal_message(char *data, int len, int nodeid, int need_b
 		break;
 
 	default:
-		log_msg(LOG_WARNING, "Unknown protocol message %d received\n", msg->cmd);
+		log_printf(LOG_WARNING, "Unknown protocol message %d received\n", msg->cmd);
 		break;
 
 	}
@@ -1797,11 +1931,11 @@ void add_ccs_node(char *nodename, int nodeid, int votes, int expected_votes)
 	if (totemip_parse(&ipaddr, nodename, 0))
 	{
 		if (!nodeid) {
-			log_msg(LOG_ERR, "Error, can't find IP address and no nodeid for node %s - ignoring it\n", nodename);
+			log_printf(LOG_ERR, "Error, can't find IP address and no nodeid for node %s - ignoring it\n", nodename);
 			return;
 		}
 		else {
-			log_msg(LOG_WARNING, "Warning, can't resolve IP address for node %s\n", nodename);
+			log_printf(LOG_WARNING, "Warning, can't resolve IP address for node %s\n", nodename);
 			memset(&ipaddr, 0, sizeof(ipaddr));
 		}
 	}
@@ -1814,7 +1948,7 @@ void add_ais_node(int nodeid, uint64_t incarnation, int total_members)
 {
 	struct cluster_node *node;
 
-	P_MEMB("add_ais_node ID=%d, incarnation = %d\n",nodeid, incarnation);
+	P_MEMB("add_ais_node ID=%d, incarnation = %lld\n",nodeid, incarnation);
 
 	node = find_node_by_nodeid(nodeid);
 	if (!node && total_members == 1) {
@@ -1827,11 +1961,11 @@ void add_ais_node(int nodeid, uint64_t incarnation, int total_members)
 		char tempname[256];
 		node = malloc(sizeof(struct cluster_node));
 		if (!node) {
-			log_msg(LOG_ERR, "error allocating node struct for id %d, but CCS doesn't know about it anyway\n",
+			log_printf(LOG_ERR, "error allocating node struct for id %d, but CCS doesn't know about it anyway\n",
 				nodeid);
 			return;
 		}
-		log_msg(LOG_ERR, "Got node from AIS id %d with no CCS entry\n", nodeid);
+		log_printf(LOG_ERR, "Got node from AIS id %d with no CCS entry\n", nodeid);
 
 		memset(node, 0, sizeof(struct cluster_node));
 		node_add_ordered(node);
@@ -1843,14 +1977,12 @@ void add_ais_node(int nodeid, uint64_t incarnation, int total_members)
 		node->name = strdup(tempname);
 	}
 
-	node->incarnation = incarnation;
-
-	gettimeofday(&node->join_time, NULL);
-
 	if (node->state == NODESTATE_DEAD) {
+		gettimeofday(&node->join_time, NULL);
+		node->incarnation = incarnation;
 		node->state = NODESTATE_MEMBER;
 		cluster_members++;
-		recalculate_quorum(0);
+		recalculate_quorum(0, 0);
 	}
 }
 
@@ -1874,21 +2006,32 @@ void del_ais_node(int nodeid)
 		node->flags &= ~NODE_FLAGS_FENCED;
 
 	node->flags &= ~NODE_FLAGS_FENCEDWHILEUP;
+	node->flags |= NODE_FLAGS_BEENDOWN;
 
-	if (node->state == NODESTATE_MEMBER) {
+	switch (node->state) {
+	case NODESTATE_MEMBER:
 		node->state = NODESTATE_DEAD;
 		cluster_members--;
-		recalculate_quorum(0);
-		return;
-	}
-	if (node->state == NODESTATE_LEAVING) {
+		recalculate_quorum(0, 0);
+		break;
+
+	case NODESTATE_AISONLY:
+		node->state = NODESTATE_DEAD;
+		break;
+
+	case NODESTATE_LEAVING:
 		node->state = NODESTATE_DEAD;
 		cluster_members--;
 
-		if ((node->leave_reason & 0xF) == CLUSTER_LEAVEFLAG_REMOVED)
-			recalculate_quorum(1);
-		else
-			recalculate_quorum(0);
+		if ((node->leave_reason & 0xF) & CLUSTER_LEAVEFLAG_REMOVED) 
+			recalculate_quorum(1, 1);
+		else 
+			recalculate_quorum(0, 0);
+		break;
+
+	case NODESTATE_JOINING:
+	case NODESTATE_DEAD:
+		break;
 	}
 }
 
@@ -1940,3 +2083,23 @@ static struct cluster_node *find_node_by_name(char *name)
 	return NULL;
 }
 
+static char *killmsg_reason(int reason)
+{
+	static char msg[1024];
+
+	switch (reason)
+	{
+	case CLUSTER_KILL_REJECTED:
+		return "our membership application was rejected";
+
+	case CLUSTER_KILL_CMANTOOL:
+		return "we were killed by cman_tool or other application";
+
+	case CLUSTER_KILL_REJOIN:
+		return "we rejoined the cluster without a full restart";
+
+	default:
+		sprintf(msg, "we got kill message number %d", reason);
+		return msg;
+	}
+}
