@@ -29,6 +29,7 @@
 #include <netinet/in.h>
 #include <sys/errno.h>
 
+#include <openais/service/swab.h>
 #include <openais/totem/aispoll.h>
 #include <openais/totem/totemip.h>
 
@@ -157,6 +158,8 @@ static int process_client(poll_handle handle, int fd, int revent, void *data)
 		int len;
 		int totallen = 0;
 
+		memset(buf, 0, (MAX_CLUSTER_MESSAGE + sizeof(struct sock_header)));
+
 		len = read(fd, buf, sizeof(struct sock_header));
 
 		P_DAEMON("read %d bytes from fd %d\n", len, fd);
@@ -185,6 +188,11 @@ static int process_client(poll_handle handle, int fd, int revent, void *data)
 			send_status_return(con, msg->command, -EINVAL);
 			return 0;
 		}
+		if ((msg->length-len) > MAX_CLUSTER_MESSAGE) {
+			P_DAEMON("message on socket is too big\n");
+			send_status_return(con, msg->command, -EINVAL);
+			return 0;
+		}
 
 		totallen = len;
 
@@ -329,7 +337,7 @@ static int open_local_sock(const char *name, int name_len, mode_t mode, poll_han
 		unlink(name);
 	local_socket = socket(PF_UNIX, SOCK_STREAM, 0);
 	if (local_socket < 0) {
-		log_msg(LOG_ERR, "Can't create local socket %s: %s\n", name, strerror(errno));
+		log_printf(LOG_ERR, "Can't create local socket %s: %s\n", name, strerror(errno));
 		write_cman_pipe("Can't create local cman socket");
 		return -1;
 	}
@@ -341,13 +349,13 @@ static int open_local_sock(const char *name, int name_len, mode_t mode, poll_han
 	memcpy(sockaddr.sun_path, name, name_len);
 	sockaddr.sun_family = AF_UNIX;
 	if (bind(local_socket, (struct sockaddr *) &sockaddr, sizeof(sockaddr))) {
-		log_msg(LOG_ERR, "can't bind local socket to %s: %s\n", name, strerror(errno));
+		log_printf(LOG_ERR, "can't bind local socket to %s: %s\n", name, strerror(errno));
 		write_cman_pipe("Can't bind to local cman socket");
 		close(local_socket);
 		return -1;
 	}
 	if (listen(local_socket, 1) != 0) {
-		log_msg(LOG_ERR, "listen on %s failed: %s\n", name, strerror(errno));
+		log_printf(LOG_ERR, "listen on %s failed: %s\n", name, strerror(errno));
 		write_cman_pipe("listen failed on local cman socket");
 		close(local_socket);
 		return -1;
@@ -358,7 +366,7 @@ static int open_local_sock(const char *name, int name_len, mode_t mode, poll_han
 
 	con = malloc(sizeof(struct connection));
 	if (!con) {
-		log_msg(LOG_ERR, "Can't allocate space for local connection: %s\n", strerror(errno));
+		log_printf(LOG_ERR, "Can't allocate space for local connection: %s\n", strerror(errno));
 		write_cman_pipe("malloc failed for connection info");
 		close(local_socket);
 		return -1;
@@ -473,8 +481,8 @@ int cman_init()
 	ais_poll_handle = aisexec_poll_handle;
 	barrier_init();
 
-	log_msg(LOG_INFO, "CMAN %s (built %s %s) started\n",
-		CMAN_RELEASE_NAME, __DATE__, __TIME__);
+	log_printf(LOG_INFO, "CMAN %s (built %s %s) started\n",
+		   RELEASE_VERSION, __DATE__, __TIME__);
 
 	fd = open_local_sock(CLIENT_SOCKNAME, sizeof(CLIENT_SOCKNAME), 0660, ais_poll_handle, CON_CLIENT);
 	if (fd < 0)
