@@ -73,7 +73,7 @@ int main(int argc, char *argv[])
 
 		case 'V':
 			printf("%s %s (built %s %s)\n", prog_name,
-				FENCE_RELEASE_NAME, __DATE__, __TIME__);
+				RELEASE_VERSION, __DATE__, __TIME__);
 			printf("%s\n", REDHAT_COPYRIGHT);
 			exit(EXIT_SUCCESS);
 			break;
