@@ -15,7 +15,7 @@
 #include "ccs.h"
 #include "copyright.cf"
 
-#define OPTION_STRING			("cj:f:Dn:O:hVSw")
+#define OPTION_STRING			("cj:f:Dn:O:hVS")
 #define LOCKFILE_NAME			"/var/run/fenced.pid"
 
 struct client {
@@ -607,17 +607,12 @@ static void decode_arguments(int argc, char **argv, commandline_t *comline)
 			break;
 
 		case 'V':
-			printf("fenced %s (built %s %s)\n", FENCE_RELEASE_NAME,
+			printf("fenced %s (built %s %s)\n", RELEASE_VERSION,
 				 __DATE__, __TIME__);
 			printf("%s\n", REDHAT_COPYRIGHT);
 			exit(EXIT_SUCCESS);
 			break;
 
-		case 'w':
-			/* do nothing, this is a fence_tool option that
-			   we ignore when fence_tool starts us */
-			break;
-
 		case ':':
 		case '?':
 			fprintf(stderr, "Please use '-h' for usage.\n");
