@@ -24,6 +24,7 @@
 #include <linux/blkdev.h>
 #include <linux/mm.h>
 #include <linux/aio.h>
+#include <linux/writeback.h>
 #include <asm/uaccess.h>
 
 #include "gfs_ioctl.h"
@@ -1528,6 +1529,7 @@ gfs_fsync(struct file *file, struct dentry *dentry, int datasync)
 {
 	struct gfs_inode *ip = get_v2ip(dentry->d_inode);
 	struct gfs_holder i_gh;
+	struct inode *inode = dentry->d_inode;
 	int error;
 
 	atomic_inc(&ip->i_sbd->sd_ops_file);
@@ -1538,8 +1540,15 @@ gfs_fsync(struct file *file, struct dentry *dentry, int datasync)
 
 	if (gfs_is_jdata(ip))
 		gfs_log_flush_glock(ip->i_gl);
-	else
-		i_gh.gh_flags |= GL_SYNC;
+	else {
+		if ((!datasync) || (inode->i_state & I_DIRTY_DATASYNC)) {
+			struct writeback_control wbc = {
+				.sync_mode = WB_SYNC_ALL,
+				.nr_to_write = 0,
+			};
+			error = sync_inode(inode, &wbc);
+		}
+	}
 
 	gfs_glock_dq_uninit(&i_gh);
 
@@ -1573,13 +1582,7 @@ gfs_lock(struct file *file, int cmd, struct file_lock *fl)
 
 	if (sdp->sd_args.ar_localflocks) {
 		if (IS_GETLK(cmd)) {
-			int conflict;
-			struct file_lock tmp;
-
-			conflict = posix_test_lock(file, fl, &tmp);
-			fl->fl_type = F_UNLCK;
-			if (conflict)
-				memcpy(fl, &tmp, sizeof(struct file_lock));
+			posix_test_lock(file, fl);
 			return 0;
 		} else {
 			return posix_lock_file_wait(file, fl);
@@ -1594,6 +1597,7 @@ gfs_lock(struct file *file, int cmd, struct file_lock *fl)
 		return gfs_lm_plock(sdp, &name, file, cmd, fl);
 }
 
+#if 0
 /**
  * gfs_sendfile - Send bytes to a file or socket
  * @in_file: The file to read from
@@ -1633,6 +1637,7 @@ gfs_sendfile(struct file *in_file, loff_t *offset, size_t count, read_actor_t ac
 
 	return retval;
 }
+#endif
 
 /**
  * do_flock - Acquire a flock on a file
@@ -1758,7 +1763,7 @@ struct file_operations gfs_file_fops = {
 	.release = gfs_close,
 	.fsync = gfs_fsync,
 	.lock = gfs_lock,
-	.sendfile = gfs_sendfile,
+	/* .sendfile = gfs_sendfile, */
 	.flock = gfs_flock,
 };
 
