@@ -2,7 +2,7 @@
 *******************************************************************************
 **
 **  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
-**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
+**  Copyright (C) 2004-2006 Red Hat, Inc.  All rights reserved.
 **
 **  This copyrighted material is made available to anyone wishing to use,
 **  modify, copy, or redistribute it subject to the terms and conditions
@@ -296,7 +296,7 @@ depend_sync_old(struct gfs_rgrpd *rgd)
 	struct gfs_sbd *sdp = rgd->rd_sbd;
 	struct gfs_depend *gd;
 
-	for (;;) {
+	while (!list_empty(&rgd->rd_depend)) {
 		/* Oldest entries are in prev direction */
 		gd = list_entry(rgd->rd_depend.prev,
 				struct gfs_depend,
@@ -666,6 +666,7 @@ compute_bitstructs(struct gfs_rgrpd *rgd)
 	rgd->rd_bh = kmalloc(length * sizeof(struct buffer_head *), GFP_KERNEL);
 	if (!rgd->rd_bh) {
 		kfree(rgd->rd_bits);
+		rgd->rd_bits = NULL;
 		return -ENOMEM;
 	}
 	memset(rgd->rd_bh, 0, length * sizeof(struct buffer_head *));
@@ -1121,6 +1122,7 @@ recent_rgrp_add(struct gfs_rgrpd *new_rgd)
 		if (++count >= max)
 			goto out;
 	}
+	new_rgd->rd_try_counter = 0;
 	list_add_tail(&new_rgd->rd_recent, &sdp->sd_rg_recent);
 
  out:
@@ -1198,24 +1200,31 @@ get_local_rgrp(struct gfs_inode *ip)
 	int skipped = 0;
 	int loops = 0;
 	int error;
+	int try_flag;
+	unsigned int try_threshold = gfs_tune_get(sdp, gt_rgrp_try_threshold);
 
 	/* Try recently successful rgrps */
 
 	rgd = recent_rgrp_first(sdp, ip->i_last_rg_alloc);
 
 	while (rgd) {
+		try_flag = (rgd->rd_try_counter >= try_threshold) ?
+			0: LM_FLAG_TRY;
 		error = gfs_glock_nq_init(rgd->rd_gl,
-					  LM_ST_EXCLUSIVE, LM_FLAG_TRY,
+					  LM_ST_EXCLUSIVE, try_flag,
 					  &al->al_rgd_gh);
 		switch (error) {
 		case 0:
-			if (try_rgrp_fit(rgd, al))
+			if (try_rgrp_fit(rgd, al)) {
+				rgd->rd_try_counter = 0;
 				goto out;
+			}
 			gfs_glock_dq_uninit(&al->al_rgd_gh);
 			rgd = recent_rgrp_next(rgd, TRUE);
 			break;
 
 		case GLR_TRYFAILED:
+			rgd->rd_try_counter++;
 			rgd = recent_rgrp_next(rgd, FALSE);
 			break;
 
@@ -1492,7 +1501,7 @@ blkfree_internal(struct gfs_sbd *sdp, uint64_t bstart, uint32_t blen,
 	rgd = gfs_blk2rgrpd(sdp, bstart);
 	if (!rgd) {
 		if (gfs_consist(sdp))
-			printk("GFS: fsid=%s: block = %"PRIu64"\n",
+			printk("GFS: fsid=%s: block = %llu\n",
 			       sdp->sd_fsname, bstart);
 		return NULL;
 	}
@@ -1649,6 +1658,9 @@ gfs_blkalloc(struct gfs_inode *ip, uint64_t *block)
 	al->al_alloced_data++;
 
 	gfs_trans_add_quota(sdp, +1, ip->i_di.di_uid, ip->i_di.di_gid);
+
+	/* total=0, free=-1, dinodes=0 */
+	gfs_statfs_modify(sdp, 0, -1, 0);
 }
 
 /**
@@ -1703,6 +1715,9 @@ gfs_metaalloc(struct gfs_inode *ip, uint64_t *block)
 
 	gfs_trans_add_quota(sdp, +1, ip->i_di.di_uid, ip->i_di.di_gid);
 
+	/* total=0, free=-1, dinode=0 */
+	gfs_statfs_modify(sdp, 0, -1, 0);
+
 	return 0;
 }
 
@@ -1718,6 +1733,7 @@ gfs_metaalloc(struct gfs_inode *ip, uint64_t *block)
 int
 gfs_dialloc(struct gfs_inode *dip, uint64_t *block)
 {
+	struct gfs_sbd *sdp = dip->i_sbd;
 	struct gfs_alloc *al = dip->i_alloc;
 	struct gfs_rgrpd *rgd = al->al_rgd;
 	uint32_t goal, blk;
@@ -1757,6 +1773,9 @@ gfs_dialloc(struct gfs_inode *dip, uint64_t *block)
 	al->al_alloced_di++;
 	al->al_alloced_meta++;
 
+	/* total=0, free=-1, dinodes=1 */
+	gfs_statfs_modify(sdp, 0, -1, +1);
+
 	return error;
 }
 
@@ -1789,6 +1808,9 @@ gfs_blkfree(struct gfs_inode *ip, uint64_t bstart, uint32_t blen)
 	gfs_trans_add_quota(sdp, -(int64_t)blen,
 			    ip->i_di.di_uid,
 			    ip->i_di.di_gid);
+
+	/* total=0, free=+blen, dinodes=0 */
+	gfs_statfs_modify(sdp, 0, blen, 0);
 }
 
 /**
@@ -1823,6 +1845,9 @@ gfs_metafree(struct gfs_inode *ip, uint64_t bstart, uint32_t blen)
 	gfs_trans_add_bh(rgd->rd_gl, rgd->rd_bh[0]);
 	gfs_rgrp_out(&rgd->rd_rg, rgd->rd_bh[0]->b_data);
 
+	/* total=0, free=blen, dinode=0 */
+	gfs_statfs_modify(sdp, 0, blen, 0);
+
 	gfs_trans_add_quota(sdp, -(int64_t)blen,
 			    ip->i_di.di_uid,
 			    ip->i_di.di_gid);
@@ -1857,6 +1882,9 @@ gfs_difree_uninit(struct gfs_rgrpd *rgd, uint64_t addr)
 
 	gfs_trans_add_bh(rgd->rd_gl, rgd->rd_bh[0]);
 	gfs_rgrp_out(&rgd->rd_rg, rgd->rd_bh[0]->b_data);
+
+	/* total=0, free=1, dinodes=-1 */
+	gfs_statfs_modify(rgd->rd_sbd, 0, +1, -1);
 }
 
 /**
@@ -1905,7 +1933,7 @@ gfs_rlist_add(struct gfs_sbd *sdp, struct gfs_rgrp_list *rlist, uint64_t block)
 	rgd = gfs_blk2rgrpd(sdp, block);
 	if (!rgd) {
 		if (gfs_consist(sdp))
-			printk("GFS: fsid=%s: block = %"PRIu64"\n",
+			printk("GFS: fsid=%s: block = %llu\n",
 			       sdp->sd_fsname, block);
 		return;
 	}
