@@ -668,7 +668,7 @@ make_rgrp(uint64_t offset, uint64_t size)
 	rgl->rg.rg_header.mh_format = GFS_FORMAT_RG;
 	rgl->rg.rg_free = rgl->ri.ri_data;
 
-	osi_list_add(&rgl->list, &rglist_new);
+	osi_list_add_prev(&rgl->list, &rglist_new);
 	return offset + size;
 }
 
@@ -851,7 +851,7 @@ main(int argc, char *argv[])
 			break;
 		case 'V':
 			printf("%s %s (built %s %s)\n", argv[0],
-			       GFS_RELEASE_NAME, __DATE__, __TIME__);
+			       RELEASE_VERSION, __DATE__, __TIME__);
 			printf("%s\n", REDHAT_COPYRIGHT);
 			exit(0);
 		case 'h':
