@@ -22,6 +22,7 @@
 #include <sys/ioctl.h>
 #include <assert.h>
 #include <time.h>
+#include <mntent.h>
 
 #include "global.h"
 #include "gfs_ondisk.h"
@@ -150,7 +151,7 @@ static void decode_arguments(int argc, char *argv[], commandline_t *comline)
 
 
     case 'V':
-      printf("gfs_mkfs %s (built %s %s)\n", GFS_RELEASE_NAME, __DATE__, __TIME__);
+      printf("gfs_mkfs %s (built %s %s)\n", RELEASE_VERSION, __DATE__, __TIME__);
       printf("%s\n", REDHAT_COPYRIGHT);
       exit(EXIT_SUCCESS);
       break;
@@ -242,6 +243,36 @@ void are_you_sure(commandline_t *comline)
 }
 
 
+/**
+ * check_mount -
+ * @
+ *
+ */
+
+void check_mount(char *device)
+{
+	struct mntent *mnt;
+	FILE *fp;
+
+	if ((fp = setmntent("/proc/mounts", "r")) == NULL) {
+		die("error opening /proc/mounts");
+	}
+
+	while ((mnt = getmntent(fp)) != NULL) {
+		if (strcmp(device, mnt->mnt_fsname) == 0) {
+			printf("cannot create filesystem: ");
+			printf("%s appears to be mounted\n", device);
+			endmntent(fp);
+			exit(EXIT_FAILURE);
+		}
+	}
+
+	endmntent(fp);
+
+	return;
+}
+
+
 /**
  * print_results - print out summary information
  * @comline: the command line
@@ -307,6 +338,8 @@ int main(int argc, char *argv[])
 
 	decode_arguments(argc, argv, &comline);
 
+	check_mount(comline.device);
+
 	if (!comline.expert) {
 		char buf[256];
 		if (test_locking(comline.lockproto, comline.locktable, buf, 256))
