@@ -42,9 +42,13 @@
 
 #include <syslog.h>
 
+#define RGLIST_DUMMY_BLOCK -2
+
 int display(int identify_only);
 extern void eol(int col);
-extern void do_indirect_extended(char *buf);
+extern int do_indirect_extended(char *buf, struct iinfo *ii);
+extern void savemeta(const char *in_fn, const char *out_fn, int slow);
+extern void restoremeta(const char *in_fn, const char *out_device);
 
 /* ------------------------------------------------------------------------ */
 /* UpdateSize - screen size changed, so update it                           */
@@ -322,6 +326,25 @@ void print_usage(void)
 	Erase();
 }
 
+
+
+/* ------------------------------------------------------------------------ */
+/* get_block_type                                                           */
+/* returns: metatype if block is a GFS2 structure block type                */
+/*          0 if block is not a GFS2 structure                              */
+/* ------------------------------------------------------------------------ */
+int get_block_type(const char *lpBuffer)
+{
+	int ret_type = 0;
+
+	if (*(lpBuffer+0)==0x01 && *(lpBuffer+1)==0x16 &&
+	    *(lpBuffer+2)==0x19 && *(lpBuffer+3)==0x70 &&
+	    *(lpBuffer+4)==0x00 && *(lpBuffer+5)==0x00 &&
+	    *(lpBuffer+6)==0x00) /* If magic number appears at the start */
+		ret_type = *(lpBuffer+7);
+	return ret_type;
+}
+
 /* ------------------------------------------------------------------------ */
 /* display_block_type                                                       */
 /* returns: metatype if block is a GFS2 structure block type                */
@@ -341,7 +364,10 @@ int display_block_type(const char *lpBuffer)
 		if (edit_row[dmode] == -1)
 			COLORS_HIGHLIGHT;
 	}
-	print_gfs2("%lld    (0x%"PRIx64")", block, block);
+	if (block == RGLIST_DUMMY_BLOCK)
+		print_gfs2("RG List       ");
+	else
+		print_gfs2("%lld    (0x%"PRIx64")", block, block);
 	if (termlines) {
 		if (edit_row[dmode] == -1)
 			COLORS_NORMAL;
@@ -355,10 +381,11 @@ int display_block_type(const char *lpBuffer)
 	else
 		printf(" ");
 
-	if (*(lpBuffer+0)==0x01 && *(lpBuffer+1)==0x16 && *(lpBuffer+2)==0x19 &&
-		*(lpBuffer+3)==0x70 && *(lpBuffer+4)==0x00 && *(lpBuffer+5)==0x00 &&
-		*(lpBuffer+6)==0x00) { /* If magic number appears at the start */
-		ret_type = *(lpBuffer+7);
+	if (block == RGLIST_DUMMY_BLOCK) {
+		ret_type = GFS2_METATYPE_RG;
+		struct_len = sizeof(struct gfs2_rgrp);
+	}
+	else if ((ret_type = get_block_type(lpBuffer))) {
 		switch (*(lpBuffer+7)) {
 		case GFS2_METATYPE_SB:   /* 1 */
 			print_gfs2("(superblock)");
@@ -431,6 +458,8 @@ int display_block_type(const char *lpBuffer)
 		print_gfs2("-------------------- Root directory ------------------");
 	else if (!gfs1 && block == sbd.sd_sb.sb_master_dir.no_addr)
 		print_gfs2("------------------- Master directory -----------------");
+	else if (!gfs1 && block == RGLIST_DUMMY_BLOCK)
+		print_gfs2("----------------------- RG List ----------------------");
 	else {
 		if (gfs1) {
 			if (block == sbd1->sb_rindex_di.no_addr)
@@ -569,13 +598,14 @@ uint64_t masterblock(const char *fn)
 }
 
 /* ------------------------------------------------------------------------ */
-/* print_rindex - print the rgindex file.                                   */
+/* parse_rindex - print the rgindex file.                                   */
 /* ------------------------------------------------------------------------ */
-int print_rindex(struct gfs2_inode *di)
+int parse_rindex(struct gfs2_inode *di, int print_rindex)
 {
 	int error, start_line;
 	struct gfs2_rindex ri;
 	char buf[sizeof(struct gfs2_rindex)];
+	char highlighted_addr[32];
 
 	start_line = line;
 	error = 0;
@@ -583,10 +613,11 @@ int print_rindex(struct gfs2_inode *di)
 			   di->i_di.di_size / sizeof(struct gfs2_rindex));
 	eol(0);
 	lines_per_row[dmode] = 6;
+	memset(highlighted_addr, 0, sizeof(highlighted_addr));
 	for (print_entry_ndx=0; ; print_entry_ndx++) {
 		error = gfs2_readi(di, (void *)&buf,
-						   print_entry_ndx * sizeof(struct gfs2_rindex),
-						   sizeof(struct gfs2_rindex));
+				   print_entry_ndx * sizeof(struct gfs2_rindex),
+				   sizeof(struct gfs2_rindex));
 		gfs2_rindex_in(&ri, buf);
 		if (!error) /* end of file */
 			break;
@@ -596,16 +627,30 @@ int print_rindex(struct gfs2_inode *di)
 			 termlines - start_line - 2)) {
 			if (edit_row[dmode] == print_entry_ndx) {
 				COLORS_HIGHLIGHT;
-				sprintf(estring, "%" PRIx64, ri.ri_addr);
+				sprintf(highlighted_addr, "%llx", (unsigned long long)ri.ri_addr);
 			}
 			print_gfs2("RG #%d", print_entry_ndx);
+			if (!print_rindex)
+				print_gfs2(" located at: %llu (0x%llx)",
+					   ri.ri_addr, ri.ri_addr);
 			eol(0);
 			if (edit_row[dmode] == print_entry_ndx)
 				COLORS_NORMAL;
-			gfs2_rindex_print(&ri);
+			if(print_rindex)
+				gfs2_rindex_print(&ri);
+			else {
+				struct gfs2_rgrp rg;
+				struct gfs2_buffer_head *tmp_bh;
+
+				tmp_bh = bread(&sbd, ri.ri_addr);
+				gfs2_rgrp_in(&rg, tmp_bh->b_data);
+				gfs2_rgrp_print(&rg);
+				brelse(tmp_bh, not_updated);
+			}
 			last_entry_onscreen[dmode] = print_entry_ndx;
 		}
 	}
+	strcpy(estring, highlighted_addr);
 	end_row[dmode] = print_entry_ndx;
 	return error;
 }
@@ -792,49 +837,51 @@ void print_inode_type(__be16 de_type)
 /* ------------------------------------------------------------------------ */
 /* display_indirect                                                         */
 /* ------------------------------------------------------------------------ */
-int display_indirect(void)
+int display_indirect(struct iinfo *ind, int indblocks, int level, uint64_t startoff)
 {
 	int start_line, total_dirents, indir_blocks;
-	int i, cur_height = -1;
-	uint64_t factor[5];
+	int i, cur_height = -1, pndx;
+	uint64_t factor[5]={0,0,0,0,0};
 	int offsets[5];
 
 	last_entry_onscreen[dmode] = 0;
-	eol(0);
+	if (!level)
+		eol(0);
 	start_line = line;
 	if (!has_indirect_blocks())
 		return -1;
 
-	indir_blocks = indirect_blocks;
-	if (!indirect_blocks) {
-		if (gfs2_struct_type == GFS2_METATYPE_SB)
-			print_gfs2("The superblock has 2 directories");
-		else
-			print_gfs2("This directory contains %d directory entries.",
-					   indirect[0].dirents);
-		indir_blocks = 1; /* not really an indirect block, but treat it as one */
-	}
-	else {
-		if (gfs2_struct_type == GFS2_METATYPE_DI) {
-			if (S_ISDIR(di.di_mode))
-				print_gfs2("This directory contains %d indirect blocks",
-						   indirect_blocks);
+	indir_blocks = indblocks;
+	if (!level) {
+		if (!indblocks) {
+			if (gfs2_struct_type == GFS2_METATYPE_SB)
+				print_gfs2("The superblock has 2 directories");
 			else
-				print_gfs2("This inode contains %d indirect blocks",
-						   indirect_blocks);
+				print_gfs2("This directory contains %d directory entries.",
+					   ind->ii[0].dirents);
+			indir_blocks = 1; /* not really an indirect block, but treat it as one */
+		}
+		else {
+			if (gfs2_struct_type == GFS2_METATYPE_DI) {
+				if (S_ISDIR(di.di_mode))
+					print_gfs2("This directory contains %d indirect blocks",
+						   indblocks);
+				else
+					print_gfs2("This inode contains %d indirect blocks",
+						   indblocks);
+			}
+			else
+				print_gfs2("This indirect block contains %d indirect blocks",
+					   indblocks);
 		}
-		else
-			print_gfs2("This indirect block contains %d indirect blocks",
-					   indirect_blocks);
 	}
 	total_dirents = 0;
 	/* Figure out multiplication factors for indirect pointers. */
-	if ((indir_blocks == indirect_blocks) && !S_ISDIR(di.di_mode)) {
+	if ((indir_blocks == indblocks) && !S_ISDIR(di.di_mode)) {
 		memset(&offsets, 0, sizeof(offsets));
 		/* See if we are on an inode or have one in history. */
-		cur_height = 0;
-		if (gfs2_struct_type != GFS2_METATYPE_DI) {
-			cur_height = 0;
+		cur_height = level;
+		if (!level && gfs2_struct_type != GFS2_METATYPE_DI) {
 			for (i = 0; i <= blockhist && i < 5; i++) {
 				offsets[i] = blockstack[(blockhist - i) % BLOCK_STACK_SIZE].edit_row[dmode];
 				if (blockstack[(blockhist - i) % BLOCK_STACK_SIZE].gfs2_struct_type == GFS2_METATYPE_DI)
@@ -843,24 +890,36 @@ int display_indirect(void)
 			}
 		}
 		if (cur_height >= 0) {
+			int diptrs, inptrs;
+
+			if (gfs1) {
+				diptrs = 483;
+				inptrs = 501;
+			} else {
+				diptrs = (bufsize - sizeof(sizeof(struct gfs2_dinode))) / sizeof(uint64_t);
+				inptrs = (bufsize - sizeof(sizeof(struct gfs2_meta_header))) /
+					sizeof(uint64_t);
+			}
 			/* Multiply out the max factor based on inode height.*/
 			/* This is how much data is represented by each      */
 			/* indirect pointer at each height.                  */
 			factor[0] = 1ull;
 			for (i = 0; i < di.di_height; i++)
-				factor[i + 1] = factor[i] * (gfs1 ? 501 : 509);
+				factor[i + 1] = factor[i] * inptrs;
 		}
-		print_gfs2("  (at height=%d)", cur_height);
-	}
-	if (indirect_blocks) {
+		if (!level)
+			print_gfs2("  (at height=%d)", cur_height);
 		eol(0);
-		print_gfs2("Indirect blocks:");
 	}
-	eol(0);
-	for (print_entry_ndx = start_row[dmode];
-		 (!termlines || print_entry_ndx < termlines - start_line - 2
-		  + start_row[dmode]) && print_entry_ndx < indir_blocks;
-		 print_entry_ndx++) {
+	if (!level && indblocks) {
+		print_gfs2("Indirect blocks:");
+		eol(0);
+	}
+	for (pndx = start_row[dmode];
+		 (!termlines || pndx < termlines - start_line - 2
+		  + start_row[dmode]) && pndx < indir_blocks;
+		 pndx++) {
+		print_entry_ndx = pndx;
 		if (termlines) {
 			if (edit_row[dmode] >= 0 &&
 				line - start_line - 2 == edit_row[dmode] -
@@ -868,18 +927,24 @@ int display_indirect(void)
 				COLORS_HIGHLIGHT;
 			move(line, 1);
 		}
-		if (indir_blocks == indirect_blocks) {
-			print_gfs2("%d => ", print_entry_ndx);
+		if (indir_blocks == indblocks) {
+			if (!termlines) {
+				int h;
+
+				for (h = 0; h < level; h++)
+					print_gfs2("   ");
+			}
+			print_gfs2("%d => ", pndx);
 			if (termlines)
 				move(line,9);
-			print_gfs2("0x%llx / %lld", indirect[print_entry_ndx].block,
-					   indirect[print_entry_ndx].block);
+			print_gfs2("0x%llx / %lld", ind->ii[pndx].block,
+					   ind->ii[pndx].block);
 			if (termlines) {
 				if (edit_row[dmode] >= 0 &&
 					line - start_line - 2 == edit_row[dmode] -
 					start_row[dmode]) { 
 					sprintf(estring, "%"PRIx64,
-							indirect[print_entry_ndx].block);
+							ind->ii[print_entry_ndx].block);
 					strcpy(edit_fmt, "%"PRIx64);
 					edit_size[dmode] = strlen(estring);
 					COLORS_NORMAL;
@@ -887,14 +952,14 @@ int display_indirect(void)
 			}
 			if (!S_ISDIR(di.di_mode)) {
 				int hgt;
-				uint64_t file_offset = 0ull;
+				uint64_t file_offset = startoff;
 				float human_off;
 				char h;
 				
 				/* Now divide by how deep we are at the moment.      */
 				/* This is how much data is represented by each      */
 				/* indirect pointer for each height we've traversed. */
-				offsets[0] = print_entry_ndx;
+				offsets[0] = pndx;
 				for (hgt = cur_height; hgt >= 0; hgt--)
 					file_offset += offsets[cur_height - hgt] *
 						factor[di.di_height - hgt - 1] * bufsize;
@@ -909,18 +974,40 @@ int display_indirect(void)
 				print_gfs2("(data offset 0x%llx / %lld / %6.2f%c)",
 						   file_offset, file_offset, human_off, h);
 				print_gfs2("   ");
+				if (!termlines && level + 1 < di.di_height) {
+					struct iinfo *more_indir;
+					int more_ind;
+					char *tmpbuf;
+
+					more_indir = malloc(sizeof(struct iinfo));
+					tmpbuf = malloc(bufsize);
+					if (tmpbuf) {
+						do_lseek(fd,
+							 ind->ii[pndx].block * bufsize);
+						do_read(fd, tmpbuf,
+							bufsize); /* read in the desired block */
+						memset(more_indir, 0, sizeof(struct iinfo));
+						more_ind = do_indirect_extended(tmpbuf,
+										more_indir);
+						display_indirect(more_indir,
+								 more_ind, level + 1, file_offset);
+						free(tmpbuf);
+					}
+					free(more_indir);
+				}
+				print_entry_ndx = pndx; /* restore after recursion */
 			}
 		}
-		if (indirect[print_entry_ndx].is_dir) {
+		if (ind->ii[pndx].is_dir) {
 			int d;
 			
-			if (indirect[print_entry_ndx].dirents > 1 &&
-				indir_blocks == indirect_blocks)
+			if (ind->ii[pndx].dirents > 1 &&
+				indir_blocks == indblocks)
 				print_gfs2("(directory leaf with %d entries)",
-						   indirect[print_entry_ndx].dirents);
-			for (d = 0; d < indirect[print_entry_ndx].dirents; d++) {
+						   ind->ii[pndx].dirents);
+			for (d = 0; d < ind->ii[pndx].dirents; d++) {
 				total_dirents++;
-				if (indirect[print_entry_ndx].dirents > 1) {
+				if (ind->ii[pndx].dirents > 1) {
 					eol(5);
 					if (termlines) {
 						if (edit_row[dmode] >=0 &&
@@ -929,19 +1016,19 @@ int display_indirect(void)
 							start_row[dmode]) {
 							COLORS_HIGHLIGHT;
 							sprintf(estring, "%"PRIx64,
-									indirect[print_entry_ndx].dirent[d].block);
+									ind->ii[pndx].dirent[d].block);
 							strcpy(edit_fmt, "%"PRIx64);
 						}
 					}
 					print_gfs2("%d. (%d). %lld (0x%llx) / %lld (0x%llx): ",
 							   total_dirents, d + 1,
-							   indirect[print_entry_ndx].dirent[d].dirent.de_inum.no_formal_ino,
-							   indirect[print_entry_ndx].dirent[d].dirent.de_inum.no_formal_ino,
-							   indirect[print_entry_ndx].dirent[d].block,
-							   indirect[print_entry_ndx].dirent[d].block);
+							   ind->ii[pndx].dirent[d].dirent.de_inum.no_formal_ino,
+							   ind->ii[pndx].dirent[d].dirent.de_inum.no_formal_ino,
+							   ind->ii[pndx].dirent[d].block,
+							   ind->ii[pndx].dirent[d].block);
 				}
-				print_inode_type(indirect[print_entry_ndx].dirent[d].dirent.de_type);
-				print_gfs2(" %s", indirect[print_entry_ndx].dirent[d].filename);
+				print_inode_type(ind->ii[pndx].dirent[d].dirent.de_type);
+				print_gfs2(" %s", ind->ii[pndx].dirent[d].filename);
 				if (termlines) {
 					if (edit_row[dmode] >= 0 &&
 						line - start_line - 2 == edit_row[dmode] -
@@ -955,7 +1042,7 @@ int display_indirect(void)
 	if (line >= 7) /* 7 because it was bumped at the end */
 		last_entry_onscreen[dmode] = line - 7;
 	eol(0);
-	end_row[dmode] = (indirect_blocks ? indirect_blocks:indirect[0].dirents);
+	end_row[dmode] = (indblocks ? indblocks:ind->ii[0].dirents);
 	if (end_row[dmode] < last_entry_onscreen[dmode])
 		end_row[dmode] = last_entry_onscreen[dmode];
 	lines_per_row[dmode] = 1;
@@ -973,6 +1060,18 @@ int block_is_rindex(void)
 	return FALSE;
 }
 
+/* ------------------------------------------------------------------------ */
+/* block_is_rglist - there's no such block as the rglist.  This is a        */
+/*                   special case meant to parse the rindex and follow the  */
+/*                   blocks to the real rgs.                                */
+/* ------------------------------------------------------------------------ */
+int block_is_rglist(void)
+{
+	if (block == RGLIST_DUMMY_BLOCK)
+		return TRUE;
+	return FALSE;
+}
+
 /* ------------------------------------------------------------------------ */
 /* block_is_jindex                                                          */
 /* ------------------------------------------------------------------------ */
@@ -1021,6 +1120,7 @@ int block_has_extended_info(void)
 {
 	if (has_indirect_blocks() ||
 	    block_is_rindex() ||
+	    block_is_rglist() ||
 	    block_is_jindex() ||
 	    block_is_inum_file() ||
 	    block_is_statfs_file() ||
@@ -1038,12 +1138,18 @@ int display_extended(void)
 	struct gfs2_buffer_head *tmp_bh;
 
 	/* Display any indirect pointers that we have. */
-	if (display_indirect() == 0)
-		return -1;
-	else if (block_is_rindex()) {
+	if (block_is_rindex()) {
 		tmp_bh = bread(&sbd, block);
 		tmp_inode = inode_get(&sbd, tmp_bh);
-		print_rindex(tmp_inode);
+		parse_rindex(tmp_inode, TRUE);
+		brelse(tmp_bh, not_updated);
+	}
+	else if (display_indirect(indirect, indirect_blocks, 0, 0) == 0)
+		return -1;
+	else if (block_is_rglist()) {
+		tmp_bh = bread(&sbd, masterblock("rindex"));
+		tmp_inode = inode_get(&sbd, tmp_bh);
+		parse_rindex(tmp_inode, FALSE);
 		brelse(tmp_bh, not_updated);
 	}
 	else if (block_is_jindex()) {
@@ -1082,7 +1188,7 @@ void read_superblock(void)
 
 	sbd1 = (struct gfs_sb *)&sbd.sd_sb;
 	ioctl(fd, BLKFLSBUF, 0);
-	do_lseek(fd, 0x10 * bufsize);
+	do_lseek(fd, 0x10 * 4096);
 	do_read(fd, buf, bufsize); /* read in the desired block */
 	memset(&sbd, 0, sizeof(struct gfs2_sbd));
 	sbd.device_fd = fd;
@@ -1112,6 +1218,8 @@ void read_superblock(void)
 	}
 	else
 		gfs1 = FALSE;
+	bufsize = sbd.sd_sb.sb_bsize;
+	block = 0x10 * (4096 / bufsize);
 }
 
 /* ------------------------------------------------------------------------ */
@@ -1132,16 +1240,22 @@ void read_master_dir(void)
 /* ------------------------------------------------------------------------ */
 int display(int identify_only)
 {
+	uint64_t blk;
+
+	if (block == RGLIST_DUMMY_BLOCK)
+		blk = masterblock("rindex");
+	else
+		blk = block;
 	if (termlines) {
 		display_title_lines();
 		move(2,0);
 	}
-	if (block_in_mem != block) { /* If we changed blocks from the last read */
-		dev_offset = block * bufsize;
+	if (block_in_mem != blk) { /* If we changed blocks from the last read */
+		dev_offset = blk * bufsize;
 		ioctl(fd, BLKFLSBUF, 0);
 		do_lseek(fd, dev_offset);
 		do_read(fd, buf, bufsize); /* read in the desired block */
-		block_in_mem = block; /* remember which block is in memory */
+		block_in_mem = blk; /* remember which block is in memory */
 	}
 	line = 1;
 	gfs2_struct_type = display_block_type(buf);
@@ -1149,55 +1263,56 @@ int display(int identify_only)
 		return 0;
 	indirect_blocks = 0;
 	lines_per_row[dmode] = 1;
-	if (gfs2_struct_type == GFS2_METATYPE_SB || block == 0x10) {
+	if (gfs2_struct_type == GFS2_METATYPE_SB || blk == 0x10 * (4096 / bufsize)) {
 		gfs2_sb_in(&sbd.sd_sb, buf); /* parse it out into the sb structure */
-		memset(&indirect, 0, sizeof(indirect));
-		indirect[0].block = sbd.sd_sb.sb_master_dir.no_addr;
-		indirect[0].is_dir = TRUE;
-		indirect[0].dirents = 2;
+		memset(indirect, 0, sizeof(indirect));
+		indirect->ii[0].block = sbd.sd_sb.sb_master_dir.no_addr;
+		indirect->ii[0].is_dir = TRUE;
+		indirect->ii[0].dirents = 2;
 
-		memcpy(&indirect[0].dirent[0].filename, "root", 4);
-		indirect[0].dirent[0].dirent.de_inum.no_formal_ino =
+		memcpy(&indirect->ii[0].dirent[0].filename, "root", 4);
+		indirect->ii[0].dirent[0].dirent.de_inum.no_formal_ino =
 			sbd.sd_sb.sb_root_dir.no_formal_ino;
-		indirect[0].dirent[0].dirent.de_inum.no_addr =
+		indirect->ii[0].dirent[0].dirent.de_inum.no_addr =
 			sbd.sd_sb.sb_root_dir.no_addr;
-		indirect[0].dirent[0].block = sbd.sd_sb.sb_root_dir.no_addr;
-		indirect[0].dirent[0].dirent.de_type = DT_DIR;
+		indirect->ii[0].dirent[0].block = sbd.sd_sb.sb_root_dir.no_addr;
+		indirect->ii[0].dirent[0].dirent.de_type = DT_DIR;
 
-		memcpy(&indirect[0].dirent[1].filename, "master", 7);
-		indirect[0].dirent[1].dirent.de_inum.no_formal_ino = 
+		memcpy(&indirect->ii[0].dirent[1].filename, "master", 7);
+		indirect->ii[0].dirent[1].dirent.de_inum.no_formal_ino = 
 			sbd.sd_sb.sb_master_dir.no_formal_ino;
-		indirect[0].dirent[1].dirent.de_inum.no_addr =
+		indirect->ii[0].dirent[1].dirent.de_inum.no_addr =
 			sbd.sd_sb.sb_master_dir.no_addr;
-		indirect[0].dirent[1].block = sbd.sd_sb.sb_master_dir.no_addr;
-		indirect[0].dirent[1].dirent.de_type = DT_DIR;
+		indirect->ii[0].dirent[1].block = sbd.sd_sb.sb_master_dir.no_addr;
+		indirect->ii[0].dirent[1].dirent.de_type = DT_DIR;
 	}
 	else if (gfs2_struct_type == GFS2_METATYPE_DI) {
 		gfs2_dinode_in(&di, buf); /* parse disk inode into structure */
 		do_dinode_extended(&di, buf); /* get extended data, if any */
 	}
 	else if (gfs2_struct_type == GFS2_METATYPE_IN) { /* indirect block list */
-		do_indirect_extended(buf);
+		do_indirect_extended(buf, indirect);
 	}
 	else if (gfs2_struct_type == GFS2_METATYPE_LF) { /* directory leaf */
 		int x;
 		struct gfs2_dirent de;
 
 		indirect_blocks = 1;
-		memset(&indirect, 0, sizeof(indirect));
+		memset(indirect, 0, sizeof(indirect));
 		/* Directory Entries: */
 		for (x = sizeof(struct gfs2_leaf); x < bufsize;
 		     x += de.de_rec_len) {
 			gfs2_dirent_in(&de, buf + x);
 			if (de.de_inum.no_addr) {
-				indirect[indirect_blocks].block = de.de_inum.no_addr;
-				indirect[indirect_blocks].dirent[x].block = de.de_inum.no_addr;
-				memcpy(&indirect[indirect_blocks].dirent[x].dirent, &de,
+				indirect->ii[indirect_blocks].block = de.de_inum.no_addr;
+				indirect->ii[indirect_blocks].dirent[x].block = de.de_inum.no_addr;
+				memcpy(&indirect->ii[indirect_blocks].dirent[x].dirent, &de,
 					   sizeof(struct gfs2_dirent));
-				memcpy(&indirect[indirect_blocks].dirent[x].filename,
+				memcpy(&indirect->ii[indirect_blocks].dirent[x].filename,
 					   buf + x + sizeof(struct gfs2_dirent), de.de_name_len);
-				indirect[indirect_blocks].is_dir = TRUE;
-				indirect[indirect_blocks].dirents++;
+				indirect->ii[indirect_blocks].dirent[x].filename[de.de_name_len] = '\0';
+				indirect->ii[indirect_blocks].is_dir = TRUE;
+				indirect->ii[indirect_blocks].dirents++;
 			}
 			if (de.de_rec_len <= sizeof(struct gfs2_dirent))
 				break;
@@ -1303,15 +1418,33 @@ uint64_t goto_block(void)
 				else if (!strcmp(string, "quota"))
 					temp_blk = gfs1_quota_di.no_addr;
 			}
+			else {
+				if (!strcmp(string, "rgs"))
+					temp_blk = RGLIST_DUMMY_BLOCK;
+				else
+					temp_blk = masterblock(string);
+			}
+		}
+		else if (string[0] == '+') {
+			if (string[1] == '0' && string[2] == 'x')
+				sscanf(string, "%"SCNx64, &temp_blk);
 			else
-				temp_blk = masterblock(string);
+				sscanf(string, "%" PRIu64, &temp_blk);
+			temp_blk += block;
+		}
+		else if (string[0] == '-') {
+			if (string[1] == '0' && string[2] == 'x')
+				sscanf(string, "%"SCNx64, &temp_blk);
+			else
+				sscanf(string, "%" PRIu64, &temp_blk);
+			temp_blk -= block;
 		}
 		else if (string[0] == '0' && string[1] == 'x')
 			sscanf(string, "%"SCNx64, &temp_blk); /* retrieve in hex */
 		else
 			sscanf(string, "%" PRIu64, &temp_blk); /* retrieve decimal */
 
-		if (temp_blk < max_block) {
+		if (temp_blk == RGLIST_DUMMY_BLOCK || temp_blk < max_block) {
 			offset = 0;
 			block = temp_blk;
 			push_block(block);
@@ -1464,9 +1597,11 @@ void jump(void)
 		unsigned int col2;
 		uint64_t *b;
 		
-		col2 = edit_col[dmode] & 0x08;/* thus 0-7->0, 8-15->8 */
-		b = (uint64_t *)&buf[edit_row[dmode]*16 + offset + col2];
-		temp_blk=be64_to_cpu(*b);
+		if (edit_row[dmode] >= 0) {
+			col2 = edit_col[dmode] & 0x08;/* thus 0-7->0, 8-15->8 */
+			b = (uint64_t *)&buf[edit_row[dmode]*16 + offset + col2];
+			temp_blk=be64_to_cpu(*b);
+		}
 	}
 	else
 		sscanf(estring, "%"SCNx64, &temp_blk);/* retrieve in hex */
@@ -1550,7 +1685,7 @@ void interactive_mode(void)
 				edit_row[dmode] = 0;
 			}
 			else {
-				block = 0x10;
+				block = 0x10 * (4096 / bufsize);
 				push_block(block);
 				offset = 0;
 			}
@@ -1736,6 +1871,79 @@ void interactive_mode(void)
     endwin();
 }/* interactive_mode */
 
+/* ------------------------------------------------------------------------ */
+/* dump_journal - dump a journal file's contents.                           */
+/* ------------------------------------------------------------------------ */
+void dump_journal(const char *journal)
+{
+	struct gfs2_buffer_head *jindex_bh, *j_bh;
+	uint64_t jindex_block, jblock, j_size, jb;
+	int error, start_line, journal_num;
+	struct gfs2_dinode jdi;
+	char jbuf[bufsize];
+	struct gfs2_inode *j_inode;
+
+	start_line = line;
+	lines_per_row[dmode] = 1;
+	error = 0;
+	journal_num = atoi(journal + 7);
+	print_gfs2("Dumping journal #%d.", journal_num);
+	eol(0);
+	/* Figure out the block of the jindex file */
+	jindex_block = masterblock("jindex");
+	/* read in the block */
+	jindex_bh = bread(&sbd, jindex_block);
+	/* get the dinode data from it. */
+	gfs2_dinode_in(&di, jindex_bh->b_data); /* parse disk inode into structure */
+
+	do_dinode_extended(&di, jindex_bh->b_data); /* which parses the directory. */
+	brelse(jindex_bh, not_updated);
+
+	jblock = indirect->ii[0].dirent[journal_num + 2].block;
+	j_bh = bread(&sbd, jblock);
+	j_inode = inode_get(&sbd, j_bh);
+	gfs2_dinode_in(&jdi, j_bh->b_data); /* parse disk inode into structure */
+	j_size = jdi.di_size;
+
+	for (jb = 0; jb < j_size; jb += bufsize) {
+		error = gfs2_readi(j_inode, (void *)&jbuf, jb, bufsize);
+		if (!error) /* end of file */
+			break;
+		if (get_block_type(jbuf) == GFS2_METATYPE_LD) {
+			uint64_t *b;
+			int i = 0;
+
+			print_gfs2("Block #%4llx: ", jb / bufsize);
+			b = (uint64_t *)(jbuf + sizeof(struct gfs2_log_descriptor));
+			while (*b && (char *)b < (jbuf + bufsize)) {
+				if (!termlines ||
+				    (print_entry_ndx >= start_row[dmode] &&
+				     ((print_entry_ndx - start_row[dmode])+1) *
+				     lines_per_row[dmode] <= termlines - start_line - 2)) {
+					if (i && i % 4 == 0) {
+						eol(0);
+						print_gfs2("             ");
+					}
+					i++;
+					print_gfs2("0x%08llx   ", be64_to_cpu(*b));
+				}
+				b++;
+			}
+			eol(0);
+		} else if (get_block_type(jbuf) == GFS2_METATYPE_LH) {
+			struct gfs2_log_header lh;
+
+			gfs2_log_header_in(&lh, jbuf);
+			print_gfs2("Block #%4llx: Log header: Seq = 0x%x, tail = 0x%x, blk = 0x%x",
+				   jb / bufsize, lh.lh_sequence, lh.lh_tail,
+				   lh.lh_blkno);
+			eol(0);
+		}
+	}
+	brelse(j_bh, not_updated);
+	blockhist = -1; /* So we don't print anything else */
+}
+
 /* ------------------------------------------------------------------------ */
 /* usage - print command line usage                                         */
 /* ------------------------------------------------------------------------ */
@@ -1744,6 +1952,9 @@ void usage(void)
 	fprintf(stderr,"\nFormat is: gfs2_edit [-c 1] [-V] [-x] [-h] [identify] [-p structures|blocks] /dev/device\n\n");
 	fprintf(stderr,"If only the device is specified, it enters into hexedit mode.\n");
 	fprintf(stderr,"identify - prints out only the block type, not the details.\n");
+	fprintf(stderr,"savemeta - save off your metadata for analysis and debugging.  The intelligent way (assume bitmap is correct).\n");
+	fprintf(stderr,"savemetaslow - save off your metadata for analysis and debugging.  The SLOW way (block by block).\n");
+	fprintf(stderr,"restoremeta - restore metadata for debugging (DANGEROUS).\n");
 	fprintf(stderr,"-V   prints version number.\n");
 	fprintf(stderr,"-c 1 selects alternate color scheme 1\n");
 	fprintf(stderr,"-p   prints GFS2 structures or blocks to stdout.\n");
@@ -1756,6 +1967,7 @@ void usage(void)
 	fprintf(stderr,"     inum - prints the inum file.\n");
 	fprintf(stderr,"     statfs - prints the statfs file.\n");
 	fprintf(stderr,"     rindex - prints the rindex file.\n");
+	fprintf(stderr,"     rgs - prints all the resource groups (rgs).\n");
 	fprintf(stderr,"     quota - prints the quota file.\n");
 	fprintf(stderr,"-x   print in hexmode.\n");
 	fprintf(stderr,"-h   prints this help.\n\n");
@@ -1788,7 +2000,7 @@ void process_parameters(int argc, char *argv[], int pass)
 		if (!pass) {
 			if (!strcasecmp(argv[i], "-V")) {
 				printf("%s version %s (built %s %s)\n", prog_name,
-					   GFS2_RELEASE_NAME, __DATE__, __TIME__);
+					   RELEASE_VERSION, __DATE__, __TIME__);
 				printf("%s\n", REDHAT_COPYRIGHT);
 				exit(0);
 			}
@@ -1807,6 +2019,12 @@ void process_parameters(int argc, char *argv[], int pass)
 				termlines = 0; /* initial value--we'll figure it out later */
 				dmode = GFS2_MODE;
 			}
+			else if (!strcasecmp(argv[i], "savemeta"))
+				savemeta(argv[i+1], argv[i+2], FALSE);
+			else if (!strcasecmp(argv[i], "savemetaslow"))
+				savemeta(argv[i+1], argv[i+2], TRUE);
+			else if (!strcasecmp(argv[i], "restoremeta"))
+				restoremeta(argv[i+1], argv[i+2]);
 			else if (strchr(argv[i],'/'))
 				strcpy(device, argv[i]);
 		}
@@ -1823,7 +2041,7 @@ void process_parameters(int argc, char *argv[], int pass)
 						   max_block, max_block);
 				else if (!strcmp(argv[i], "sb") ||
 						 !strcmp(argv[i], "superblock"))
-					push_block(0x10); /* superblock */
+					push_block(0x10 * (4096 / bufsize)); /* superblock */
 				else if (!strcmp(argv[i], "root") ||
 						 !strcmp(argv[i], "rootdir"))
 					push_block(sbd.sd_sb.sb_root_dir.no_addr);
@@ -1852,6 +2070,10 @@ void process_parameters(int argc, char *argv[], int pass)
 					else
 						push_block(masterblock("rindex"));
 				}
+				else if (!strcmp(argv[i], "rgs")) {
+					if (!gfs1)
+						push_block(RGLIST_DUMMY_BLOCK);
+				}
 				else if (!strcmp(argv[i], "quota")) {
 					if (gfs1)
 						push_block(gfs1_quota_di.no_addr);
@@ -1866,6 +2088,9 @@ void process_parameters(int argc, char *argv[], int pass)
 					sscanf(argv[i], "%"SCNd64, &temp_blk);
 					push_block(temp_blk);
 				}
+				else if (!strncmp(argv[i], "journal", 7) && isdigit(argv[i][7])) {
+					dump_journal(argv[i]);
+				}
 				else {
 					fprintf(stderr,"I don't know what '%s' means.\n", argv[i]);
 					usage();
@@ -1892,6 +2117,10 @@ int main(int argc, char *argv[])
 
 	prog_name = argv[0];
 
+	indirect = malloc(sizeof(struct iinfo));
+	if (!indirect)
+		die("Out of memory.");
+	memset(indirect, 0, sizeof(struct iinfo));
 	memset(start_row, 0, sizeof(start_row));
 	memset(lines_per_row, 0, sizeof(lines_per_row));
 	memset(end_row, 0, sizeof(end_row));
@@ -1920,7 +2149,7 @@ int main(int argc, char *argv[])
 
 	fd = open(device, O_RDWR);
 	if (fd < 0)
-		die("can't open %s: %s\n", argv[1], strerror(errno));
+		die("can't open %s: %s\n", device, strerror(errno));
 	max_block = lseek(fd, 0, SEEK_END) / bufsize;
 
 	read_superblock();
@@ -1935,11 +2164,13 @@ int main(int argc, char *argv[])
 	else { /* print all the structures requested */
 		for (i = 0; i <= blockhist; i++) {
 			block = blockstack[i + 1].block;
+			if (!block)
+				break;
 			display(identify);
 			if (!identify) {
 				display_extended();
 				printf("-------------------------------------" \
-					   "-----------------");
+				       "-----------------");
 				eol(0);
 			}
 			block = pop_block();
@@ -1948,5 +2179,7 @@ int main(int argc, char *argv[])
 	close(fd);
 	if (buf)
 		free(buf);
+	if (indirect)
+		free(indirect);
  	exit(EXIT_SUCCESS);
 }
