@@ -1,7 +1,7 @@
 /*****************************************************************************
 *******************************************************************************
 **
-**  Copyright (C) 2005 Red Hat, Inc.  All rights reserved.
+**  Copyright (C) 2005-2007 Red Hat, Inc.  All rights reserved.
 **
 **  This copyrighted material is made available to anyone wishing to use,
 **  modify, copy, or redistribute it subject to the terms and conditions
@@ -24,6 +24,44 @@
 #include "metawalk.h"
 #include "hash.h"
 
+int dirent_repair(struct gfs2_inode *ip, struct gfs2_buffer_head *bh,
+		  struct gfs2_dirent *de, struct gfs2_dirent *dent,
+		  int type, int first)
+{
+	char *bh_end, *p;
+	int calc_de_name_len = 0;
+	
+	/* If this is a sentinel, just fix the length and move on */
+	if (first && !de->de_inum.no_formal_ino) { /* Is it a sentinel? */
+		if (type == DIR_LINEAR)
+			de->de_rec_len = bh->b_size -
+				sizeof(struct gfs2_dinode);
+		else
+			de->de_rec_len = bh->b_size - sizeof(struct gfs2_leaf);
+	}
+	else {
+		bh_end = bh->b_data + bh->b_size;
+		/* first, figure out a probable name length */
+		p = (char *)dent + sizeof(struct gfs2_dirent);
+		while (*p &&         /* while there's a non-zero char and */
+		       p < bh_end) { /* not past end of buffer */
+			calc_de_name_len++;
+			p++;
+		}
+		if (!calc_de_name_len)
+			return 1;
+		/* There can often be noise at the end, so only          */
+		/* Trust the shorter of the two in case we have too much */
+		/* Or rather, only trust ours if it's shorter.           */
+		if (!de->de_name_len || de->de_name_len > NAME_MAX ||
+		    calc_de_name_len < de->de_name_len) /* if dent is hosed */
+			de->de_name_len = calc_de_name_len; /* use ours */
+		de->de_rec_len = GFS2_DIRENT_SIZE(de->de_name_len);
+	}
+	gfs2_dirent_out(de, (char *)dent);
+	return 0;
+}
+
 int check_entries(struct gfs2_inode *ip, struct gfs2_buffer_head *bh,
 				  int index, int type, int *update, uint16_t *count,
 				  struct metawalk_fxns *pass)
@@ -61,6 +99,28 @@ int check_entries(struct gfs2_inode *ip, struct gfs2_buffer_head *bh,
 		gfs2_dirent_in(&de, (char *)dent);
 		filename = (char *)dent + sizeof(struct gfs2_dirent);
 
+		if (de.de_rec_len < sizeof(struct gfs2_dirent) +
+		    de.de_name_len || !de.de_name_len) {
+			log_err("Directory block %" PRIu64 "(0x%"
+				PRIx64 "), entry %d of directory %"
+				PRIu64 "(0x%" PRIx64 ") is corrupt.\n",
+				bh->b_blocknr, bh->b_blocknr, (*count) + 1,
+				ip->i_di.di_num.no_addr,
+				ip->i_di.di_num.no_addr);
+			if (query(&opts, "Attempt to repair it? (y/n) ")) {
+				if (dirent_repair(ip, bh, &de, dent, type,
+						  first))
+					break;
+				else
+					*update = 1;
+			}
+			else {
+				log_err("Corrupt directory entry ignored, "
+					"stopped after checking %d entries.\n",
+					*count);
+				break;
+			}
+		}
 		if (!de.de_inum.no_formal_ino){
 			if(first){
 				log_debug("First dirent is a sentinel (place holder).\n");
@@ -80,7 +140,8 @@ int check_entries(struct gfs2_inode *ip, struct gfs2_buffer_head *bh,
 				de.de_inum.no_addr = de.de_inum.no_formal_ino;
 				de.de_inum.no_formal_ino = 0;
 				gfs2_dirent_out(&de, (char *)dent);
-				*update = 1; /* Mark dirent buffer as modified */
+				*update = (opts.no ? not_updated : updated);
+				/* Mark dirent buffer as modified */
 				first = 0;
 			}
 			else {
@@ -96,15 +157,6 @@ int check_entries(struct gfs2_inode *ip, struct gfs2_buffer_head *bh,
 			}
 		}
 
-		if (de.de_rec_len < sizeof(struct gfs2_dirent)) {
-			log_err("Entry %" PRIu64 "(0x%"
-				PRIx64 ") of directory %" PRIu64 "(0x%"
-				PRIx64 ") is corrupt, skipping.\n",
-				bh->b_blocknr, bh->b_blocknr,
-				ip->i_di.di_num.no_addr,
-				ip->i_di.di_num.no_addr);
-                        break;
-                }
 		if ((char *)dent + de.de_rec_len >= bh_end){
 			log_debug("Last entry processed.\n");
 			break;
@@ -120,12 +172,37 @@ int check_entries(struct gfs2_inode *ip, struct gfs2_buffer_head *bh,
 	return 0;
 }
 
+/* Process a bad leaf pointer and ask to repair the first time.      */
+/* The repair process involves extending the previous leaf's entries */
+/* so that they replace the bad ones.  We have to hack up the old    */
+/* leaf a bit, but it's better than deleting the whole directory,    */
+/* which is what used to happen before.                              */
+void warn_and_patch(struct gfs2_inode *ip, uint64_t *leaf_no, 
+		    uint64_t *bad_leaf, uint64_t old_leaf, int index,
+		    const char *msg)
+{
+	if (*bad_leaf != *leaf_no) {
+		log_err("Directory Inode %" PRIu64 "(0x%"
+			PRIx64 ") points to leaf %" PRIu64 "(0x%"
+			PRIx64 ") %s.\n", ip->i_di.di_num.no_addr,
+			ip->i_di.di_num.no_addr, *leaf_no, *leaf_no, msg);
+	}
+	if (*leaf_no == *bad_leaf ||
+	    query(&opts, "Attempt to patch around it? (y/n) ")) {
+		gfs2_put_leaf_nr(ip, index, old_leaf);
+	}
+	else
+		log_err("Bad leaf left in place.\n");
+	*bad_leaf = *leaf_no;
+	*leaf_no = old_leaf;
+}
+
 /* Checks exhash directory entries */
 int check_leaf(struct gfs2_inode *ip, int *update, struct metawalk_fxns *pass)
 {
 	int error;
-	struct gfs2_leaf leaf;
-	uint64_t leaf_no, old_leaf;
+	struct gfs2_leaf leaf, oldleaf;
+	uint64_t leaf_no, old_leaf, bad_leaf = -1;
 	struct gfs2_buffer_head *lbh;
 	int index;
 	struct gfs2_sbd *sbp = ip->i_sbd;
@@ -133,60 +210,101 @@ int check_leaf(struct gfs2_inode *ip, int *update, struct metawalk_fxns *pass)
 	int ref_count = 0, exp_count = 0;
 
 	old_leaf = 0;
+	memset(&oldleaf, 0, sizeof(oldleaf));
 	for(index = 0; index < (1 << ip->i_di.di_depth); index++) {
 		gfs2_get_leaf_nr(ip, index, &leaf_no);
 
 		/* GFS has multiple indirect pointers to the same leaf
 		 * until those extra pointers are needed, so skip the
 		 * dups */
-		if(old_leaf == leaf_no) {
+		if (leaf_no == bad_leaf) {
+			gfs2_put_leaf_nr(ip, index, old_leaf); /* fill w/old
+								  leaf info */
+			ref_count++;
+			continue;
+		}
+		else if(old_leaf == leaf_no) {
 			ref_count++;
 			continue;
 		} else {
 			if(ref_count != exp_count){
-				log_err("Dir #%" PRIu64 " (0x%" PRIx64 ") has an incorrect "
-						"number of pointers to leaf #%" PRIu64 " (0x%" PRIx64
-						")\n\tFound: %u,  Expected: %u\n",
-						ip->i_di.di_num.no_addr, ip->i_di.di_num.no_addr,
-						old_leaf, old_leaf, ref_count, exp_count);
-				return 1;
+				log_err("Dir #%" PRIu64 " (0x%"
+					PRIx64 ") has an incorrect "
+					"number of pointers to leaf #%"
+					PRIu64 " (0x%" PRIx64
+					")\n\tFound: %u,  Expected: %u\n",
+					ip->i_di.di_num.no_addr,
+					ip->i_di.di_num.no_addr,
+					old_leaf, old_leaf, ref_count,
+					exp_count);
+				if (query(&opts, "Attempt to fix it? (y/n) "))
+				{
+					int factor = 0, divisor = ref_count;
+
+					lbh = bread(sbp, old_leaf);
+					while (divisor > 1) {
+						factor++;
+						divisor /= 2;
+					}
+					oldleaf.lf_depth = ip->i_di.di_depth -
+						factor;
+					gfs2_leaf_out(&oldleaf, lbh->b_data);
+					brelse(lbh, updated);
+				}
+				else
+					return 1;
 			}
 			ref_count = 1;
 		}
 
 		count = 0;
 		do {
-			/* FIXME: Do other checks (see old
-			 * pass3:dir_exhash_scan() */
-			lbh = NULL;
-			if(pass->check_leaf) {
-				error = pass->check_leaf(ip, leaf_no, &lbh, pass->private);
-				if(error < 0) {
-					stack;
-					return -1;
-				}
-				if(error > 0) {
-					lbh = NULL;
-					return 1;
-				}
+			/* Make sure the block number is in range. */
+			if(gfs2_check_range(ip->i_sbd, leaf_no)){
+				log_err("Leaf block #%" PRIu64 " (0x%"
+					PRIx64 ") is out of range for "
+					"directory #%" PRIu64 " (0x%"
+					PRIx64 ").\n", leaf_no, leaf_no,
+					ip->i_di.di_num.no_addr,
+					ip->i_di.di_num.no_addr);
+				warn_and_patch(ip, &leaf_no, &bad_leaf,
+					       old_leaf, index,
+					       "that is out of range");
+				memcpy(&leaf, &oldleaf, sizeof(oldleaf));
+				break;
 			}
 
 			*update = not_updated;
+			/* Try to read in the leaf block. */
 			lbh = bread(sbp, leaf_no);
+			/* Make sure it's really a valid leaf block. */
+			if (gfs2_check_meta(lbh, GFS2_METATYPE_LF)) {
+				warn_and_patch(ip, &leaf_no, &bad_leaf,
+					       old_leaf, index,
+					       "that is not really a leaf");
+				memcpy(&leaf, &oldleaf, sizeof(oldleaf));
+				brelse(lbh, (opts.no ? not_updated : updated));
+				break;
+			}
 			gfs2_leaf_in(&leaf, lbh->b_data);
+			if(pass->check_leaf) {
+				error = pass->check_leaf(ip, leaf_no, lbh,
+							 pass->private);
+			}
 
 			/*
-			 * Early versions of GFS2 had an endianess bug in the kernel
-			 * that set lf_dirent_format to cpu_to_be16(GFS2_FORMAT_DE).
-			 * This was fixed to use cpu_to_be32(), but we should check
-			 * for incorrect values and replace them with the correct value. */
+			 * Early versions of GFS2 had an endianess bug in the
+			 * kernel that set lf_dirent_format to
+			 * cpu_to_be16(GFS2_FORMAT_DE).  This was fixed to use
+			 * cpu_to_be32(), but we should check for incorrect 
+			 * values and replace them with the correct value. */
 
 			if (leaf.lf_dirent_format == (GFS2_FORMAT_DE << 16)) {
 				log_debug("incorrect lf_dirent_format at leaf #%" PRIu64 "\n", leaf_no);
 				leaf.lf_dirent_format = GFS2_FORMAT_DE;
 				gfs2_leaf_out(&leaf, lbh->b_data);
 				log_debug("Fixing lf_dirent_format.\n");
-				*update = updated;
+				*update = (opts.no ? not_updated : updated);
 			}
 
 			/* Make sure it's really a leaf. */
@@ -253,6 +371,7 @@ int check_leaf(struct gfs2_inode *ip, int *update, struct metawalk_fxns *pass)
 			}
 		} while(1);
 		old_leaf = leaf_no;
+		memcpy(&oldleaf, &leaf, sizeof(oldleaf));
 	}
 	return 0;
 }
@@ -383,9 +502,9 @@ static int check_indirect_eattr(struct gfs2_inode *ip, uint64_t indirect,
 			error = check_leaf_eattr(ip, block, indirect, pass);
 			ea_leaf_ptr++;
 		}
+		brelse(indirect_buf, not_updated);
 	}
 
-	brelse(indirect_buf, not_updated);
 	return error;
 }
 
@@ -425,79 +544,81 @@ int check_inode_eattr(struct gfs2_inode *ip, struct metawalk_fxns *pass)
  * @mlp:
  */
 static int build_and_check_metalist(struct gfs2_inode *ip,
-									struct metawalk_fxns *pass)
+				    osi_list_t *mlp,
+				    struct metawalk_fxns *pass)
 {
 	uint32_t height = ip->i_di.di_height;
-	struct gfs2_buffer_head *nbh, *metabh;
-	int head_size;
+	struct gfs2_buffer_head *bh, *nbh, *metabh;
+	osi_list_t *prev_list, *cur_list, *tmp;
+	int i, head_size;
 	uint64_t *ptr, block;
 	int err;
 
 	metabh = bread(ip->i_sbd, ip->i_di.di_num.no_addr);
 
+	osi_list_add(&metabh->b_altlist, &mlp[0]);
+
 	/* if(<there are no indirect blocks to check>) */
-	if (height < 1) {
+	if (height < 2) {
 		brelse(metabh, not_updated);
 		return 0;
 	}
-	head_size = sizeof(struct gfs2_dinode);
-	for (ptr = (uint64_t *)(metabh->b_data + head_size);
-		 (char *)ptr < (metabh->b_data + metabh->b_size); ptr++) {
-		nbh = NULL;
-		
-		if (!*ptr)
-			continue;
+	for (i = 1; i < height; i++){
+		prev_list = &mlp[i - 1];
+		cur_list = &mlp[i];
 
-		block = be64_to_cpu(*ptr);
-		if (height > 1) {
-			uint64_t *subptr, subblock, prevsubblock;
-			int head_size2;
-			struct gfs2_buffer_head *nbh2;
-			
-			nbh2 = NULL;
-			prevsubblock = 0;
-			head_size2 = sizeof(struct gfs2_meta_header);
-			err = pass->check_metalist(ip, block, &nbh, pass->private);
-			if (!nbh)
-				nbh = nbh2 = bread(ip->i_sbd, block);
-			for (subptr = (uint64_t *)(nbh->b_data + head_size2);
-				 (char *)subptr < (nbh->b_data + nbh->b_size);
-				 subptr++) {
+		for (tmp = prev_list->next; tmp != prev_list; tmp = tmp->next){
+			bh = osi_list_entry(tmp, struct gfs2_buffer_head,
+					    b_altlist);
+
+			head_size = (i > 1 ?
+				     sizeof(struct gfs2_meta_header) :
+				     sizeof(struct gfs2_dinode));
+
+			for (ptr = (uint64_t *)(bh->b_data + head_size);
+			     (char *)ptr < (bh->b_data + bh->b_size);
+			     ptr++) {
+				nbh = NULL;
+		
+				if (!*ptr)
+					continue;
+
+				block = be64_to_cpu(*ptr);
+				err = pass->check_metalist(ip, block, &nbh,
+							   pass->private);
 				if(err < 0) {
 					stack;
 					goto fail;
 				}
 				if(err > 0) {
-					log_debug("Skipping block %" PRIu64 " (0x%" PRIx64 ")\n",
-							  block, block);
+					log_debug("Skipping block %" PRIu64
+						  " (0x%" PRIx64 ")\n",
+						  block, block);
 					continue;
 				}
-				subblock = be64_to_cpu(*subptr);
-				if (subblock) {
-					if (subblock != prevsubblock) {
-						prevsubblock = subblock;
-						if(pass->check_data &&
-						   (pass->check_data(ip, subblock,
-											 pass->private) < 0)) {
-							stack;
-							return -1;
-						}
-					} /* if different than the previous block */
-				} /* if non-zero indirect block pointer */
+				if(!nbh) {
+					nbh = bread(ip->i_sbd, block);
+					osi_list_add(&nbh->b_altlist, cur_list);
+				}
+				else
+					osi_list_add(&nbh->b_altlist, cur_list);
 			} /* for all data on the indirect block */
-			if (nbh2)
-				brelse(nbh2, not_updated);
-		} /* if height */
-		else if(pass->check_data &&
-				(pass->check_data(ip, block, pass->private) < 0)) {
-			stack;
-			brelse(metabh, not_updated);
-			return -1;
-		}
-	} /* for all level 1 pointers */
-fail:
+		} /* for blocks at that height */
+	} /* for height */
 	brelse(metabh, not_updated);
 	return 0;
+fail:
+	for (i = 0; i < GFS2_MAX_META_HEIGHT; i++) {
+		osi_list_t *list;
+		list = &mlp[i];
+		while (!osi_list_empty(list)) {
+			nbh = osi_list_entry(list->next,
+					     struct gfs2_buffer_head, b_altlist);
+			osi_list_del(&nbh->b_altlist);
+		}
+	}
+	brelse(metabh, not_updated);
+	return -1;
 }
 
 /**
@@ -508,18 +629,71 @@ fail:
  */
 int check_metatree(struct gfs2_inode *ip, struct metawalk_fxns *pass)
 {
+	osi_list_t metalist[GFS2_MAX_META_HEIGHT];
+	osi_list_t *list, *tmp;
+	struct gfs2_buffer_head *bh;
+	uint64_t block, *ptr;
 	uint32_t height = ip->i_di.di_height;
+	int  i, head_size;
 	int update = 0;
 	int error = 0;
 
-	if (height) {
-		/* create metalist for each level */
-		if(build_and_check_metalist(ip, pass)){
-			stack;
-			return -1;
+	if (!height)
+		goto end;
+
+	for (i = 0; i < GFS2_MAX_META_HEIGHT; i++)
+		osi_list_init(&metalist[i]);
+
+	/* create metalist for each level */
+	if (build_and_check_metalist(ip, &metalist[0], pass)){
+		stack;
+		return -1;
+	}
+
+	/* We don't need to record directory blocks - they will be
+	 * recorded later...i think... */
+        if (S_ISDIR(ip->i_di.di_mode))
+		log_debug("Directory with height > 0 at %"PRIu64"\n",
+			  ip->i_di.di_num.no_addr);
+
+	/* check data blocks */
+	list = &metalist[height - 1];
+
+	for (tmp = list->next; tmp != list; tmp = tmp->next) {
+		bh = osi_list_entry(tmp, struct gfs2_buffer_head, b_altlist);
+
+		head_size = (height != 1 ? sizeof(struct gfs2_meta_header) :
+			     sizeof(struct gfs2_dinode));
+		ptr = (uint64_t *)(bh->b_data + head_size);
+
+		for ( ; (char *)ptr < (bh->b_data + bh->b_size); ptr++)	{
+			if (!*ptr)
+				continue;
+
+			block =  be64_to_cpu(*ptr);
+
+			if(pass->check_data &&
+			   (pass->check_data(ip, block, pass->private) < 0)) {
+				stack;
+				return -1;
+			}
 		}
 	}
-	if (S_ISDIR(ip->i_di.di_mode)) {
+
+	/* free metalists */
+	for (i = 0; i < GFS2_MAX_META_HEIGHT; i++)
+	{
+		list = &metalist[i];
+		while (!osi_list_empty(list))
+		{
+			bh = osi_list_entry(list->next,
+					    struct gfs2_buffer_head, b_altlist);
+			osi_list_del(&bh->b_altlist);
+		}
+	}
+
+end:
+        if (S_ISDIR(ip->i_di.di_mode)) {
 		/* check validity of leaf blocks and leaf chains */
 		if (ip->i_di.di_flags & GFS2_DIF_EXHASH) {
 			error = check_leaf(ip, &update, pass);
@@ -577,7 +751,7 @@ int check_dir(struct gfs2_sbd *sbp, uint64_t block, struct metawalk_fxns *pass)
 		}
 	}
 
-	inode_put(ip, not_updated); /* does a brelse */
+	inode_put(ip, opts.no ? not_updated : update); /* does a brelse */
 	return error;
 }
 
@@ -595,15 +769,12 @@ static int remove_dentry(struct gfs2_inode *ip, struct gfs2_dirent *dent,
 	memset(&dentry, 0, sizeof(struct gfs2_dirent));
 	gfs2_dirent_in(&dentry, (char *)dent);
 	de = &dentry;
+	*update = (opts.no ? not_updated : updated);
 
-	if(de->de_inum.no_addr == *dentryblock) {
-		*update = 1;
+	if(de->de_inum.no_addr == *dentryblock)
 		dirent2_del(ip, bh, prev_de, dent);
-	}
-	else {
+	else
 		(*count)++;
-		*update = 1;
-	}
 
 	return 0;
 
