@@ -2,7 +2,7 @@
 *******************************************************************************
 **
 **  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
-**  Copyright (C) 2004-2005 Red Hat, Inc.  All rights reserved.
+**  Copyright (C) 2004-2007 Red Hat, Inc.  All rights reserved.
 **
 **  This copyrighted material is made available to anyone wishing to use,
 **  modify, copy, or redistribute it subject to the terms and conditions
@@ -44,51 +44,14 @@ struct block_count {
 };
 
 static int leaf(struct gfs2_inode *ip, uint64_t block,
-				struct gfs2_buffer_head **bh, void *private)
+				struct gfs2_buffer_head *bh, void *private)
 {
-	struct gfs2_sbd *sdp = ip->i_sbd;
 	struct block_count *bc = (struct block_count *) private;
-	struct gfs2_meta_header *mh;
-	uint64_t *val;
-	enum update_flags f;
 
-	f = not_updated;
-	if(gfs2_check_range(sdp, block)){
-		log_warn("Leaf block #%" PRIu64 " (0x%" PRIx64 ") is out of range for "
-				 "directory #%" PRIu64 " (0x%" PRIx64 ").\n",
-				 block, block, ip->i_di.di_num.no_addr,
-				 ip->i_di.di_num.no_addr);
-		gfs2_block_set(bl, ip->i_di.di_num.no_addr, gfs2_bad_block);
-		return 1;
-	}
-	*bh = bread(sdp, block);
-	mh = (struct gfs2_meta_header *)(*bh)->b_data;
-	val = (uint64_t *)mh;
-
-	if(gfs2_check_meta(*bh, GFS2_METATYPE_LF)){
-		log_err("Bad meta header for leaf block #%" PRIu64 "(0x%" PRIx64 
-				") in directory #%" PRIu64 ". - is %u, should be %u\n",
-				(*bh)->b_data, (*bh)->b_data, ip->i_di.di_num.no_addr,
-				ip->i_di.di_num.no_addr,
-				(struct gfs2_meta_header *)mh->mh_type,
-				GFS2_METATYPE_LF);
-		if(query(&opts, "Clear directory inode at %" PRIu64 " (0x%"
-				 PRIx64 ")? (y/n) ", ip->i_di.di_num.no_addr,
-				 ip->i_di.di_num.no_addr)) {
-			gfs2_block_set(bl, ip->i_di.di_num.no_addr, gfs2_meta_inval);
-			log_err("Directory inode marked invalid\n");
-			f = updated;
-		} else
-			log_err("Invalid block %" PRIu64 " (0x%" PRIx64 ") ignored\n",
-				ip->i_di.di_num.no_addr, ip->i_di.di_num.no_addr);
-		brelse(*bh, f);
-		return 1;
-	}
 	log_debug("\tLeaf block at %15" PRIu64 " (0x%" PRIx64 ")\n",
 			  block, block);
 	gfs2_block_set(bl, block, gfs2_leaf_blk);
 	bc->indir_count++;
-	brelse(*bh, f);
 	return 0;
 }
 
@@ -184,10 +147,6 @@ static int check_eattr_indir(struct gfs2_inode *ip, uint64_t indirect,
 
 	/* This inode contains an eattr - it may be invalid, but the
 	 * eattr attributes points to a non-zero block */
-	log_debug("Setting %" PRIu64 " (0x%" PRIx64 ") to eattr block\n",
-			  indirect, indirect);
-	gfs2_block_set(bl, indirect, gfs2_eattr_block);
-
 	if(gfs2_check_range(sdp, indirect)) {
 		/*log_warn("EA indirect block #%"PRIu64" is out of range.\n",
 			indirect);
@@ -208,6 +167,10 @@ static int check_eattr_indir(struct gfs2_inode *ip, uint64_t indirect,
 		ret = 1;
 	}
 	else {
+		log_debug("Setting %" PRIu64 " (0x%"
+			  PRIx64 ") to eattr block\n", indirect, indirect);
+		gfs2_block_set(bl, indirect, gfs2_eattr_block);
+
 		*bh = bread(sdp, indirect);
 		block = be64_to_cpu(*(*bh)->b_data);
 		if(gfs2_check_meta(*bh, GFS2_METATYPE_IN)) {
@@ -223,7 +186,6 @@ static int check_eattr_indir(struct gfs2_inode *ip, uint64_t indirect,
 			gfs2_block_set(bl, block, gfs2_indir_blk);
 			bc->ea_count++;
 		}
-		brelse(*bh, not_updated);
 	}
 	return ret;
 }
@@ -287,11 +249,11 @@ static int check_extended_leaf_eattr(struct gfs2_inode *ip, uint64_t *data_ptr,
 }
 
 static int check_eattr_leaf(struct gfs2_inode *ip, uint64_t block,
-							uint64_t parent, struct gfs2_buffer_head **bh,
-							void *private)
+			    uint64_t parent, struct gfs2_buffer_head **bh,
+			    void *private)
 {
 	struct gfs2_sbd *sdp = ip->i_sbd;
-	struct gfs2_buffer_head *leaf_bh;
+	struct gfs2_buffer_head *leaf_bh = NULL;
 	int ret = 0;
 	struct gfs2_block_query q = {0};
 	struct block_count *bc = (struct block_count *) private;
@@ -324,10 +286,13 @@ static int check_eattr_leaf(struct gfs2_inode *ip, uint64_t block,
 	else {
 		leaf_bh = bread(sdp, block);
 		if(gfs2_check_meta(leaf_bh, GFS2_METATYPE_EA)) {
-			log_warn("EA leaf block has incorrect type.\n");
+			log_warn("EA leaf block %"PRIu64" (0x%" 
+				 PRIx64") for inode %"PRIu64" (0x%"
+				 PRIx64") has incorrect type.\n",
+				 block, block, ip->i_di.di_num.no_addr,
+				 ip->i_di.di_num.no_addr);
 			gfs2_block_set(bl, block, gfs2_meta_inval);
-			brelse(leaf_bh, not_updated);
-			ret = 1;
+			ret = -1;
 		}
 		else {
 			log_debug("Setting block %" PRIu64 " (0x%" PRIx64
@@ -430,7 +395,7 @@ int clear_data(struct gfs2_inode *ip, uint64_t block, void *private)
 }
 
 int clear_leaf(struct gfs2_inode *ip, uint64_t block,
-	       struct gfs2_buffer_head **bh, void *private)
+	       struct gfs2_buffer_head *bh, void *private)
 {
 	struct gfs2_block_query q = {0};
 	log_crit("Clearing leaf %" PRIu64 " (0x%" PRIx64 ")\n", block, block);
@@ -468,7 +433,7 @@ int add_to_dir_list(struct gfs2_sbd *sbp, uint64_t block)
 		return -1;
 	}
 
-	if(!(newdi = (struct dir_info *) malloc(sizeof(*newdi)))) {
+	if(!(newdi = (struct dir_info *) malloc(sizeof(struct dir_info)))) {
 		log_crit("Unable to allocate dir_info structure\n");
 		return -1;
 	}
@@ -801,8 +766,10 @@ int pass1(struct gfs2_sbd *sbp)
 			if (gfs2_next_rg_meta(rgd, &block, first))
 				break;
 			warm_fuzzy_stuff(block);
-			if (fsck_abort) /* if asked to abort */
+			if (fsck_abort) { /* if asked to abort */
+				gfs2_rgrp_relse(rgd, not_updated);
 				return 0;
+			}
 			if (skip_this_pass) {
 				printf("Skipping pass 1 is not a good idea.\n");
 				skip_this_pass = FALSE;
