@@ -2,7 +2,7 @@
 *******************************************************************************
 **
 **  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
-**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
+**  Copyright (C) 2004-2007 Red Hat, Inc.  All rights reserved.
 **
 **  This copyrighted material is made available to anyone wishing to use,
 **  modify, copy, or redistribute it subject to the terms and conditions
@@ -195,7 +195,7 @@ static void unstuff_dinode(struct gfs2_inode *ip)
 			buffer_copy_tail(bh, 0,
 					 ip->i_bh, sizeof(struct gfs2_dinode));
 
-			brelse(bh, not_updated);
+			brelse(bh, updated);
 		}
 	}
 
@@ -335,7 +335,8 @@ static void lookup_block(struct gfs2_inode *ip,
 }
 
 void block_map(struct gfs2_inode *ip, uint64_t lblock, int *new,
-			   uint64_t *dblock, uint32_t *extlen, int prealloc)
+	       uint64_t *dblock, uint32_t *extlen, int prealloc,
+	       enum update_flags if_changed)
 {
 	struct gfs2_sbd *sdp = ip->i_sbd;
 	struct gfs2_buffer_head *bh;
@@ -345,9 +346,7 @@ void block_map(struct gfs2_inode *ip, uint64_t lblock, int *new,
 	unsigned int height;
 	unsigned int end_of_metadata;
 	unsigned int x;
-	enum update_flags f;
 
-	f = not_updated;
 	*new = 0;
 	*dblock = 0;
 	if (extlen)
@@ -379,7 +378,7 @@ void block_map(struct gfs2_inode *ip, uint64_t lblock, int *new,
 
 	for (x = 0; x < end_of_metadata; x++) {
 		lookup_block(ip, bh, x, mp, create, new, dblock);
-		brelse(bh, not_updated);
+		brelse(bh, if_changed);
 		if (!*dblock)
 			goto out;
 
@@ -390,7 +389,6 @@ void block_map(struct gfs2_inode *ip, uint64_t lblock, int *new,
 			mh.mh_type = GFS2_METATYPE_IN;
 			mh.mh_format = GFS2_FORMAT_IN;
 			gfs2_meta_header_out(&mh, bh->b_data);
-			f = updated;
 		} else
 			bh = bread(sdp, *dblock);
 	}
@@ -420,7 +418,7 @@ void block_map(struct gfs2_inode *ip, uint64_t lblock, int *new,
 		}
 	}
 
-	brelse(bh, f);
+	brelse(bh, if_changed);
 
  out:
 	free(mp);
@@ -466,7 +464,7 @@ int gfs2_readi(struct gfs2_inode *ip, void *buf,
 		lblock = offset;
 		o = do_div(lblock, sdp->sd_jbsize);
 	} else {
-		lblock = offset >> sdp->bsize_shift;
+		lblock = offset >> sdp->sd_sb.sb_bsize_shift;
 		o = offset & (sdp->bsize - 1);
 	}
 
@@ -481,7 +479,8 @@ int gfs2_readi(struct gfs2_inode *ip, void *buf,
 			amount = sdp->bsize - o;
 
 		if (!extlen)
-			block_map(ip, lblock, &not_new, &dblock, &extlen, FALSE);
+			block_map(ip, lblock, &not_new, &dblock, &extlen,
+				  FALSE, not_updated);
 
 		if (dblock) {
 			bh = bread(sdp, dblock);
@@ -537,7 +536,7 @@ int gfs2_writei(struct gfs2_inode *ip, void *buf,
 		lblock = offset;
 		o = do_div(lblock, sdp->sd_jbsize);
 	} else {
-		lblock = offset >> sdp->bsize_shift;
+		lblock = offset >> sdp->sd_sb.sb_bsize_shift;
 		o = offset & (sdp->bsize - 1);
 	}
 
@@ -553,7 +552,8 @@ int gfs2_writei(struct gfs2_inode *ip, void *buf,
 
 		if (!extlen) {
 			new = TRUE;
-			block_map(ip, lblock, &new, &dblock, &extlen, FALSE);
+			block_map(ip, lblock, &new, &dblock, &extlen, FALSE,
+				  updated);
 		}
 
 		if (new) {
@@ -585,7 +585,7 @@ int gfs2_writei(struct gfs2_inode *ip, void *buf,
 }
 
 struct gfs2_buffer_head *get_file_buf(struct gfs2_inode *ip, uint64_t lbn,
-									  int prealloc)
+				      int prealloc)
 {
 	struct gfs2_sbd *sdp = ip->i_sbd;
 	uint64_t dbn;
@@ -594,13 +594,13 @@ struct gfs2_buffer_head *get_file_buf(struct gfs2_inode *ip, uint64_t lbn,
 	if (inode_is_stuffed(ip))
 		unstuff_dinode(ip);
 
-	block_map(ip, lbn, &new, &dbn, NULL, prealloc);
+	block_map(ip, lbn, &new, &dbn, NULL, prealloc, updated);
 	if (!dbn)
 		die("get_file_buf\n");
 
 	if (!prealloc && new &&
-	    ip->i_di.di_size < (lbn + 1) << sdp->bsize_shift)
-		ip->i_di.di_size = (lbn + 1) << sdp->bsize_shift;
+	    ip->i_di.di_size < (lbn + 1) << sdp->sd_sb.sb_bsize_shift)
+		ip->i_di.di_size = (lbn + 1) << sdp->sd_sb.sb_bsize_shift;
 
 	if (new)
 		return bget(sdp, dbn);
@@ -729,11 +729,23 @@ void gfs2_get_leaf_nr(struct gfs2_inode *dip, uint32_t index,
 		      index * sizeof(uint64_t),
 		      sizeof(uint64_t));
 	if (count != sizeof(uint64_t))
-		die("gfs2_get_leaf_nr\n");
+		die("gfs2_get_leaf_nr:  Bad internal read.\n");
 
 	*leaf_out = be64_to_cpu(leaf_no);
 }
 
+void gfs2_put_leaf_nr(struct gfs2_inode *dip, uint32_t inx, uint64_t leaf_out)
+{
+	uint64_t leaf_no;
+	int count;
+
+	leaf_no = cpu_to_be64(leaf_out);
+	count = gfs2_writei(dip, (char *)&leaf_no, inx * sizeof(uint64_t),
+			    sizeof(uint64_t));
+	if (count != sizeof(uint64_t))
+		die("gfs2_put_leaf_nr:  Bad internal write.\n");
+}
+
 static void
 dir_split_leaf(struct gfs2_inode *dip, uint32_t index, uint64_t leaf_no)
 {
@@ -953,7 +965,11 @@ dir_e_add(struct gfs2_inode *dip, char *filename, int len,
 
  restart:
 	hash = gfs2_disk_hash(filename, len);
-	index = hash >> (32 - dip->i_di.di_depth);
+	/* Have to kludge because (hash >> 32) gives hash for some reason. */
+	if (dip->i_di.di_depth)
+		index = hash >> (32 - dip->i_di.di_depth);
+	else
+		index = 0;
 
 	gfs2_get_leaf_nr(dip, index, &leaf_no);
 
@@ -1177,19 +1193,24 @@ struct gfs2_inode *createi(struct gfs2_inode *dip, char *filename,
 	uint64_t bn;
 	struct gfs2_inum inum;
 	struct gfs2_buffer_head *bh;
+	struct gfs2_inode *ip;
 
-	bn = dinode_alloc(sdp);
+	gfs2_lookupi(dip, filename, strlen(filename), &ip);
+	if (!ip) {
+		bn = dinode_alloc(sdp);
 
-	inum.no_formal_ino = sdp->md.next_inum++;
-	inum.no_addr = bn;
+		inum.no_formal_ino = sdp->md.next_inum++;
+		inum.no_addr = bn;
 
-	dir_add(dip, filename, strlen(filename), &inum, IF2DT(mode));
+		dir_add(dip, filename, strlen(filename), &inum, IF2DT(mode));
 
-	if(S_ISDIR(mode))
-		dip->i_di.di_nlink++;
+		if(S_ISDIR(mode))
+			dip->i_di.di_nlink++;
 
-	bh = init_dinode(sdp, &inum, mode, flags, &dip->i_di.di_num);
-	return inode_get(sdp, bh);
+		bh = init_dinode(sdp, &inum, mode, flags, &dip->i_di.di_num);
+		ip = inode_get(sdp, bh);
+	}
+	return ip;
 }
 
 /**
