@@ -2,7 +2,7 @@
 *******************************************************************************
 **
 **  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
-**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
+**  Copyright (C) 2004-2007 Red Hat, Inc.  All rights reserved.
 **
 **  This copyrighted material is made available to anyone wishing to use,
 **  modify, copy, or redistribute it subject to the terms and conditions
@@ -17,6 +17,7 @@
 #include <inttypes.h>
 #include <sys/types.h>
 #include <linux/types.h>
+#include <linux/limits.h>
 
 #include "linux_endian.h"
 #include <linux/gfs2_ondisk.h>
@@ -63,17 +64,12 @@ static __inline__ uint64_t do_div_i(uint64_t *num, unsigned int den)
 #define RESRANDOM do { srandom(RANDOM(1000000000)); } while (0)
 #define RANDOM(values) ((values) * (random() / (RAND_MAX + 1.0)))
 
-struct subdevice {
+struct device {
 	uint64_t start;
 	uint64_t length;
 	uint32_t rgf_flags;
 };
 
-struct device {
-	unsigned int nsubdev;
-	struct subdevice *subdev;
-};
-
 struct gfs2_bitmap
 {
 	uint32_t   bi_offset;  /* The offset in the buffer of the first byte */
@@ -84,9 +80,6 @@ typedef struct gfs2_bitmap gfs2_bitmap_t;
 
 struct rgrp_list {
 	osi_list_t list;
-
-	uint32_t subdevice;	/* The subdevice who holds this resource group */
-
 	uint64_t start;	   /* The offset of the beginning of this resource group */
 	uint64_t length;	/* The length of this resource group */
 	uint32_t rgf_flags;
@@ -100,13 +93,13 @@ struct rgrp_list {
 struct gfs2_buffer_head {
 	osi_list_t b_list;
 	osi_list_t b_hash;
+	osi_list_t b_altlist; /* alternate list */
 
 	unsigned int b_count;
 	uint64_t b_blocknr;
 	char *b_data;
 	unsigned int b_size;
 
-	int b_uninit;
 	int b_changed;
 };
 
@@ -169,16 +162,14 @@ struct gfs2_sbd {
 
 	int debug;
 	int quiet;
-	int test;
 	int expert;
 	int override;
 
-	char *device_name;
+	char device_name[PATH_MAX];
 	char *path_name;
 
 	/* Constants */
 
-	unsigned int bsize_shift;
 	uint32_t sd_fsb2bb;
 	uint32_t sd_fsb2bb_shift;
 	uint32_t sd_diptrs;
@@ -227,16 +218,20 @@ struct gfs2_sbd {
 
 	unsigned int spills;
 	unsigned int writes;
+	int metafs_fd;
+	int metafs_mounted; /* If metafs was already mounted */
+	char metafs_path[PATH_MAX]; /* where metafs is mounted */
 };
 
 extern char *prog_name;
 
 #define GFS2_DEFAULT_BSIZE          (4096)
-#define GFS2_DEFAULT_JSIZE          (32)
+#define GFS2_DEFAULT_JSIZE          (128)
 #define GFS2_DEFAULT_RGSIZE         (256)
 #define GFS2_DEFAULT_UTSIZE         (1)
 #define GFS2_DEFAULT_QCSIZE         (1)
 #define GFS2_MIN_GROW_SIZE          (10)
+#define GFS2_EXCESSIVE_RGS          (10000)
 
 #define DATA (1)
 #define META (2)
@@ -319,7 +314,7 @@ struct gfs2_block_list {
 };
 
 struct gfs2_block_list *gfs2_block_list_create(uint64_t size,
-											   uint64_t *addl_mem_needed);
+					       uint64_t *addl_mem_needed);
 int gfs2_block_mark(struct gfs2_block_list *il, uint64_t block,
 					enum gfs2_mark_block mark);
 int gfs2_block_set(struct gfs2_block_list *il, uint64_t block,
@@ -329,14 +324,14 @@ int gfs2_block_clear(struct gfs2_block_list *il, uint64_t block,
 int gfs2_block_check(struct gfs2_block_list *il, uint64_t block,
 					 struct gfs2_block_query *val);
 int gfs2_block_check_for_mark(struct gfs2_block_list *il, uint64_t block,
-							  enum gfs2_mark_block mark);
+			      enum gfs2_mark_block mark);
 void *gfs2_block_list_destroy(struct gfs2_block_list *il);
 int gfs2_find_next_block_type(struct gfs2_block_list *il,
-							  enum gfs2_mark_block m, uint64_t *b);
+			      enum gfs2_mark_block m, uint64_t *b);
 
 /* buf.c */
 struct gfs2_buffer_head *bget_generic(struct gfs2_sbd *sdp, uint64_t num,
-									  int find_existing, int read_disk);
+				      int find_existing, int read_disk);
 struct gfs2_buffer_head *bget(struct gfs2_sbd *sdp, uint64_t num);
 struct gfs2_buffer_head *bread(struct gfs2_sbd *sdp, uint64_t num);
 struct gfs2_buffer_head *bget_zero(struct gfs2_sbd *sdp, uint64_t num);
@@ -350,7 +345,6 @@ void write_buffer(struct gfs2_sbd *sdp, struct gfs2_buffer_head *bh);
 /* device_geometry.c */
 void device_geometry(struct gfs2_sbd *sdp);
 void fix_device_geometry(struct gfs2_sbd *sdp);
-void munge_device_geometry_for_grow(struct gfs2_sbd *sdp);
 
 /* fs_bits.c */
 #define BFITNOENT (0xFFFFFFFF)
@@ -359,12 +353,12 @@ void munge_device_geometry_for_grow(struct gfs2_sbd *sdp);
 uint32_t gfs2_bitcount(unsigned char *buffer, unsigned int buflen,
                      unsigned char state);
 uint32_t gfs2_bitfit(unsigned char *buffer, unsigned int buflen,
-					 uint32_t goal, unsigned char old_state);
+		     uint32_t goal, unsigned char old_state);
 
 /* functions with blk #'s that are rgrp relative */
 uint32_t gfs2_blkalloc_internal(struct rgrp_list *rgd, uint32_t goal,
-								unsigned char old_state,
-								unsigned char new_state, int do_it);
+				unsigned char old_state,
+				unsigned char new_state, int do_it);
 int gfs2_check_range(struct gfs2_sbd *sdp, uint64_t blkno);
 
 /* functions with blk #'s that are file system relative */
@@ -373,15 +367,19 @@ int gfs2_get_bitmap(struct gfs2_sbd *sdp, uint64_t blkno,
 int gfs2_set_bitmap(struct gfs2_sbd *sdp, uint64_t blkno, int state);
 
 /* fs_geometry.c */
-void compute_rgrp_layout(struct gfs2_sbd *sdp, int new_fs);
-void build_rgrps(struct gfs2_sbd *sdp);
+void rgblocks2bitblocks(unsigned int bsize, uint32_t *rgblocks,
+			uint32_t *bitblocks);
+uint64_t how_many_rgrps(struct gfs2_sbd *sdp, struct device *dev,
+			int rgsize_specified);
+void compute_rgrp_layout(struct gfs2_sbd *sdp, int rgsize_specified);
+void build_rgrps(struct gfs2_sbd *sdp, int write);
 
 /* fs_ops.c */
 #define IS_LEAF     (1)
 #define IS_DINODE   (2)
 
 struct gfs2_inode *inode_get(struct gfs2_sbd *sdp,
-							 struct gfs2_buffer_head *bh);
+			     struct gfs2_buffer_head *bh);
 void inode_put(struct gfs2_inode *ip, enum update_flags updated);
 uint64_t data_alloc(struct gfs2_inode *ip);
 uint64_t meta_alloc(struct gfs2_inode *ip);
@@ -390,26 +388,30 @@ int gfs2_readi(struct gfs2_inode *ip, void *buf,
 			   uint64_t offset, unsigned int size);
 int gfs2_writei(struct gfs2_inode *ip, void *buf,
 				uint64_t offset, unsigned int size);
-struct gfs2_buffer_head *get_file_buf(struct gfs2_inode *ip, uint64_t lbn);
+struct gfs2_buffer_head *get_file_buf(struct gfs2_inode *ip, uint64_t lbn,
+				      int prealloc);
 struct gfs2_buffer_head *init_dinode(struct gfs2_sbd *sdp,
-									 struct gfs2_inum *inum,
-									 unsigned int mode, uint32_t flags,
-									 struct gfs2_inum *parent);
+				     struct gfs2_inum *inum,
+				     unsigned int mode, uint32_t flags,
+				     struct gfs2_inum *parent);
 struct gfs2_inode *createi(struct gfs2_inode *dip, char *filename,
-						   unsigned int mode, uint32_t flags);
+			   unsigned int mode, uint32_t flags);
 void dirent2_del(struct gfs2_inode *dip, struct gfs2_buffer_head *bh,
-				 struct gfs2_dirent *prev, struct gfs2_dirent *cur);
+		 struct gfs2_dirent *prev, struct gfs2_dirent *cur);
 struct gfs2_inode *gfs2_load_inode(struct gfs2_sbd *sbp, uint64_t block);
 int gfs2_lookupi(struct gfs2_inode *dip, const char *filename, int len,
-				 struct gfs2_inode **ipp);
+		 struct gfs2_inode **ipp);
 void dir_add(struct gfs2_inode *dip, char *filename, int len,
 			 struct gfs2_inum *inum, unsigned int type);
 int gfs2_dirent_del(struct gfs2_inode *dip, struct gfs2_buffer_head *bh,
-					const char *filename, int filename_len);
+		    const char *filename, int filename_len);
 void block_map(struct gfs2_inode *ip, uint64_t lblock, int *new,
-			   uint64_t *dblock, uint32_t *extlen);
+	       uint64_t *dblock, uint32_t *extlen, int prealloc,
+	       enum update_flags if_changed);
 void gfs2_get_leaf_nr(struct gfs2_inode *dip, uint32_t index,
 					  uint64_t *leaf_out);
+void gfs2_put_leaf_nr(struct gfs2_inode *dip, uint32_t inx, uint64_t leaf_out);
+
 int gfs2_freedi(struct gfs2_sbd *sdp, uint64_t block);
 int gfs2_get_leaf(struct gfs2_inode *dip, uint64_t leaf_no,
 				  struct gfs2_buffer_head **bhp);
@@ -417,6 +419,10 @@ int gfs2_dirent_first(struct gfs2_inode *dip, struct gfs2_buffer_head *bh,
 					  struct gfs2_dirent **dent);
 int gfs2_dirent_next(struct gfs2_inode *dip, struct gfs2_buffer_head *bh,
 					 struct gfs2_dirent **dent);
+void build_height(struct gfs2_inode *ip, int height);
+unsigned int calc_tree_height(struct gfs2_inode *ip, uint64_t size);
+void write_journal(struct gfs2_sbd *sdp, struct gfs2_inode *ip, unsigned int j,
+				   unsigned int blocks);
 
 /**
  * device_size - figure out a device's size
@@ -436,6 +442,7 @@ struct gfs2_options {
 	char *device;
 	int yes:1;
 	int no:1;
+	int query:1;
 };
 
 #define MSG_DEBUG       7
@@ -490,17 +497,29 @@ void increase_verbosity(void);
 void decrease_verbosity(void);
 void print_fsck_log(int iif, int priority, char *file, int line,
 					const char *format, ...);
-int query(struct gfs2_options *opts, const char *format, ...);
+char gfs2_getch(void);
+
+char generic_interrupt(const char *caller, const char *where,
+		       const char *progress, const char *question,
+		       const char *answers);
+int gfs2_query(int *setonabort, struct gfs2_options *opts,
+	       const char *format, ...);
 
 /* misc.c */
 void compute_constants(struct gfs2_sbd *sdp);
+int find_gfs2_meta(struct gfs2_sbd *sdp);
+int dir_exists(const char *dir);
+void check_for_gfs2(struct gfs2_sbd *sdp);
+void mount_gfs2_meta(struct gfs2_sbd *sdp);
+void lock_for_admin(struct gfs2_sbd *sdp);
+void cleanup_metafs(struct gfs2_sbd *sdp);
 
 /* rgrp.c */
 int gfs2_compute_bitstructs(struct gfs2_sbd *sdp, struct rgrp_list *rgd);
 struct rgrp_list *gfs2_blk2rgrpd(struct gfs2_sbd *sdp, uint64_t blk);
 uint64_t gfs2_rgrp_read(struct gfs2_sbd *sdp, struct rgrp_list *rgd);
 void gfs2_rgrp_relse(struct rgrp_list *rgd, enum update_flags updated);
-void gfs2_rgrp_free(struct gfs2_sbd *sdp, enum update_flags updated);
+void gfs2_rgrp_free(osi_list_t *rglist, enum update_flags updated);
 
 /* structures.c */
 void build_master(struct gfs2_sbd *sdp);
@@ -521,9 +540,11 @@ int gfs2_next_rg_meta_free(struct rgrp_list *rgd, uint64_t *block, int first,
 int gfs2_next_rg_metatype(struct gfs2_sbd *sdp, struct rgrp_list *rgd,
 						  uint64_t *block, uint32_t type, int first);
 /* super.c */
+int check_sb(struct gfs2_sb *sb);
 int read_sb(struct gfs2_sbd *sdp);
 int ji_update(struct gfs2_sbd *sdp);
-int ri_update(struct gfs2_sbd *sdp, int *rgcount);
+int rindex_read(struct gfs2_sbd *sdp, int fd, int *count1);
+int ri_update(struct gfs2_sbd *sdp, int fd, int *rgcount);
 int write_sb(struct gfs2_sbd *sdp);
 
 /* ondisk.c */
@@ -532,7 +553,7 @@ uint32_t gfs2_disk_hash(const char *data, int len);
 extern void print_it(const char *label, const char *fmt,
 					 const char *fmt2, ...);
 #define pv(struct, member, fmt, fmt2) do {				   \
-		print_it("  "#member":", fmt, fmt2, struct->member); \
+		print_it("  "#member, fmt, fmt2, struct->member); \
         } while (FALSE);
 #define pv2(struct, member, fmt, fmt2) do {		 \
 		print_it("  ", fmt, fmt2, struct->member);	\
