@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2005 Red Hat, Inc.  All rights reserved.
+ * Copyright (C) 2005-2007 Red Hat, Inc.  All rights reserved.
  *
  * This copyrighted material is made available to anyone wishing to use,
  * modify, copy, or redistribute it subject to the terms and conditions
@@ -11,10 +11,11 @@
 char *prog_name;
 char *fsname;
 int verbose;
+static sigset_t old_sigset;
 
 static void print_version(void)
 {
-	printf("mount.gfs2 %s (built %s %s)\n", GFS2_RELEASE_NAME,
+	printf("mount.gfs2 %s (built %s %s)\n", RELEASE_VERSION,
 	       __DATE__, __TIME__);
 }
 
@@ -24,21 +25,25 @@ static void print_usage(void)
 	printf("This program is called by mount(8), it should not be used directly.\n");
 }
 
-static void block_signals(int how)
+static void block_sigint(void)
 {
-	sigset_t sigs;
-	sigfillset(&sigs);
-	sigdelset(&sigs, SIGTRAP);
-	sigdelset(&sigs, SIGSEGV);
-	sigdelset(&sigs, SIGINT);
-	sigprocmask(how, &sigs, (sigset_t *) 0);
+	sigset_t new;
+
+	sigemptyset(&new);
+	sigaddset(&new, SIGINT);
+	sigprocmask(SIG_BLOCK, &new, &old_sigset);
+}
+
+static void unblock_sigint(void)
+{
+	sigprocmask(SIG_SETMASK, &old_sigset, NULL);
 }
 
 static void read_options(int argc, char **argv, struct mount_options *mo)
 {
 	int cont = 1;
 	int optchar;
-	int l;
+	char *real;
 
 	/* FIXME: check for "quiet" option and don't print in that case */
 
@@ -77,18 +82,22 @@ static void read_options(int argc, char **argv, struct mount_options *mo)
 		}
 	}
 
-	if (optind < argc && argv[optind])
-		strncpy(mo->dev, argv[optind], PATH_MAX);
+	if (optind < argc && argv[optind]) {
+		real = realpath(argv[optind], NULL);
+		if (!real)
+			die("invalid device path \"%s\"\n", argv[optind]);
+		strncpy(mo->dev, real, PATH_MAX);
+		free(real);
+	}
 
 	++optind;
 
 	if (optind < argc && argv[optind]) {
-		strncpy(mo->dir, argv[optind], PATH_MAX);
-		l = strlen(mo->dir) - 1;
-		while (l > 0 && mo->dir[l] == '/') {
-			mo->dir[l] = '\0';
-			l--;
-		};
+		real = realpath(argv[optind], NULL);
+		if (!real)
+			die("invalid mount point path \"%s\"\n", argv[optind]);
+		strncpy(mo->dir, real, PATH_MAX);
+		free(real);
 	}
 
 	log_debug("mount %s %s", mo->dev, mo->dir);
@@ -146,6 +155,30 @@ static void umount_lockproto(char *proto, struct mount_options *mo,
 		lock_dlm_leave(mo, sb, mnterr);
 }
 
+#if 0
+static void check_sys_fs(char *fsname)
+{
+	DIR *d;
+	struct dirent *de;
+
+	d = opendir("/sys/fs/");
+	if (!d)
+		die("no /sys/fs/ directory found: %d\n", errno);
+
+	while ((de = readdir(d))) {
+		if (strnlen(fsname, 5) != strnlen(de->d_name, 5))
+			continue;
+		if (!strncmp(fsname, de->d_name, strnlen(fsname, 5))) {
+			closedir(d);
+			return;
+		}
+	}
+	closedir(d);
+	die("fs type \"%s\" not found in /sys/fs/, is the module loaded?\n",
+	    fsname);
+}
+#endif
+
 int main(int argc, char **argv)
 {
 	struct mount_options mo;
@@ -168,6 +201,12 @@ int main(int argc, char **argv)
 		exit(EXIT_SUCCESS);
 	}
 
+	/* This breaks on-demand fs module loading from the kernel; could we
+	   try to load the module first here and then check again and fail if
+	   nothing?  I'd really like to avoid joining the group and then
+	   backing out if the mount fails to load the module. */
+	/* check_sys_fs(fsname); */
+
 	read_options(argc, argv, &mo);
 	check_options(&mo);
 	get_sb(mo.dev, &sb);
@@ -175,12 +214,15 @@ int main(int argc, char **argv)
 
 	proto = select_lockproto(&mo, &sb);
 
+	/* there are three parts to the mount and we want all three or none
+	   to happen:  joining the mountgroup, doing the kernel mount, and
+	   adding the mtab entry */
+	block_sigint();
+
 	rv = mount_lockproto(proto, &mo, &sb);
 	if (rv < 0)
 		die("error mounting lockproto %s\n", proto);
 
-	block_signals(SIG_BLOCK);
-
 	rv = mount(mo.dev, mo.dir, fsname, mo.flags, mo.extra_plus);
 	if (rv) {
 		log_debug("mount(2) failed error %d errno %d", rv, errno);
@@ -189,19 +231,22 @@ int main(int argc, char **argv)
 		if (!(mo.flags & MS_REMOUNT))
 			umount_lockproto(proto, &mo, &sb, errno);
 
-		block_signals(SIG_UNBLOCK);
 		if (errno == EBUSY)
 			die("%s already mounted or %s busy\n", mo.dev, mo.dir);
-		die("error %d mounting %s on %s\n", errno, mo.dev, mo.dir);
+		die("error mounting %s on %s: %s\n", mo.dev, mo.dir,
+		    strerror(errno));
 	}
 	log_debug("mount(2) ok");
 	mount_result_lockproto(proto, &mo, &sb, 0);
 
-	block_signals(SIG_UNBLOCK);
-
-	if (!(mo.flags & MS_REMOUNT))
+	if (mo.flags & MS_REMOUNT) {
+                del_mtab_entry(&mo);
+                add_mtab_entry(&mo);
+        } else
 		add_mtab_entry(&mo);
 
+	unblock_sigint();
+
 	return rv ? 1 : 0;
 }
 
