@@ -12,10 +12,11 @@ char *prog_name;
 char *fsname;
 char *expert;
 int verbose;
+static sigset_t old_sigset;
 
 static void print_version(void)
 {
-	printf("umount.gfs2 %s (built %s %s)\n", GFS2_RELEASE_NAME,
+	printf("umount.gfs2 %s (built %s %s)\n", RELEASE_VERSION,
 	       __DATE__, __TIME__);
 }
 
@@ -28,6 +29,20 @@ static void print_usage(void)
 
 }
 
+static void block_sigint(void)
+{
+	sigset_t new;
+
+	sigemptyset(&new);
+	sigaddset(&new, SIGINT);
+	sigprocmask(SIG_BLOCK, &new, &old_sigset);
+}
+
+static void unblock_sigint(void)
+{
+	sigprocmask(SIG_SETMASK, &old_sigset, NULL);
+}
+
 static void read_options(int argc, char **argv, struct mount_options *mo)
 {
 	int cont = 1;
@@ -129,6 +144,8 @@ int main(int argc, char **argv)
 	get_sb(mo.dev, &sb);
 	parse_opts(&mo);
 
+	block_sigint();
+
 	rv = umount(mo.dir);
 	if (rv) {
 		if (errno == EBUSY)
@@ -141,6 +158,8 @@ int main(int argc, char **argv)
 
 	del_mtab_entry(&mo);
 
+	unblock_sigint();
+
 	return 0;
 }
 
