@@ -2,7 +2,7 @@
 *******************************************************************************
 **
 **  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
-**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
+**  Copyright (C) 2004-2007 Red Hat, Inc.  All rights reserved.
 **
 **  This copyrighted material is made available to anyone wishing to use,
 **  modify, copy, or redistribute it subject to the terms and conditions
@@ -180,34 +180,62 @@ scan_fs(dev_t device, char *dirname,
  *
  */
 static void
-read_quota_file(commandline_t *comline, osi_list_t *uid, osi_list_t *gid)
+read_quota_file(struct gfs2_sbd *sdp, commandline_t *comline,
+		osi_list_t *uid, osi_list_t *gid)
 {
 	int fd;
-	struct gfs2_sb sb;
 	char buf[sizeof(struct gfs2_quota)];
 	struct gfs2_quota q;
 	uint64_t offset = 0;
-	uint32_t id;
-	int error;
+	uint32_t id, prev, maxid;
+	int error, pass, id_type;
 	char quota_file[BUF_SIZE];
 	
-	check_for_gfs2(comline->filesystem);
-	read_superblock(&sb);
-	if (!find_gfs2_meta(comline->filesystem))
-		mount_gfs2_meta();
-	lock_for_admin();
+	strcpy(sdp->path_name, comline->filesystem);
+	check_for_gfs2(sdp);
+	read_superblock(&sdp->sd_sb, sdp);
+	if (!find_gfs2_meta(sdp))
+		mount_gfs2_meta(sdp);
+	lock_for_admin(sdp);
 	
-	strcpy(quota_file, metafs_path);
+	strcpy(quota_file, sdp->metafs_path);
 	strcat(quota_file, "/quota");
 
 	fd = open(quota_file, O_RDONLY);
 	if (fd < 0) {
-		close(metafs_fd);
-		cleanup();
+		close(sdp->metafs_fd);
+		cleanup_metafs(sdp);
 		die("can't open file %s: %s\n", comline->filesystem,
 		    strerror(errno));
 	}
 
+	if (!is_valid_quota_list(fd)) {
+		print_quota_list_warning();
+		goto do_old_school;
+	}
+	get_last_quota_id(fd, &maxid);
+	
+	for (pass=0; pass<2; pass++) {
+		id = 0;
+		id_type = pass ? GQ_ID_GROUP : GQ_ID_USER;
+		
+		do {
+			read_quota_internal(fd, id, id_type, &q);
+			prev = id;
+			q.qu_value <<= sdp->sd_sb.sb_bsize_shift - 9;
+			
+			if (q.qu_value) {
+				if (pass)
+					add_value(gid, id, q.qu_value);
+				else
+					add_value(uid, id, q.qu_value);
+			}
+			id = q.qu_ll_next;
+		} while(id && id > prev && id <= maxid);
+	}
+	goto out;
+	
+do_old_school:
 	do {
 		
 		memset(buf, 0, sizeof(struct gfs2_quota));
@@ -216,15 +244,15 @@ read_quota_file(commandline_t *comline, osi_list_t *uid, osi_list_t *gid)
 		error = read(fd, buf, sizeof(struct gfs2_quota));
 		if (error < 0) {
 			close(fd);
-			close(metafs_fd);
-			cleanup();
+			close(sdp->metafs_fd);
+			cleanup_metafs(sdp);
 			die("can't read quota file (%d): %s\n",
 			    error, strerror(errno));
 		}
 		gfs2_quota_in(&q, buf);
 
 		id = (offset / sizeof(struct gfs2_quota)) >> 1;
-		q.qu_value <<= sb.sb_bsize_shift - 9;
+		q.qu_value <<= sdp->sd_sb.sb_bsize_shift - 9;
 
 		if (q.qu_value) {
 			if (id * sizeof(struct gfs2_quota) * 2 == offset)
@@ -236,9 +264,10 @@ read_quota_file(commandline_t *comline, osi_list_t *uid, osi_list_t *gid)
 		offset += sizeof(struct gfs2_quota);
 	} while (error == sizeof(struct gfs2_quota));
 
+out:
 	close(fd);
-	close(metafs_fd);
-	cleanup();
+	close(sdp->metafs_fd);
+	cleanup_metafs(sdp);
 }
 
 /**
@@ -383,7 +412,7 @@ verify_pathname(commandline_t *comline)
  */
 
 void
-do_check(commandline_t *comline)
+do_check(struct gfs2_sbd *sdp, commandline_t *comline)
 {
 	dev_t device;
 	osi_list_t fs_uid, fs_gid, qf_uid, qf_gid;
@@ -399,7 +428,7 @@ do_check(commandline_t *comline)
 	device = verify_pathname(comline);
 
 	scan_fs(device, comline->filesystem, &fs_uid, &fs_gid, &hl);
-	read_quota_file(comline, &qf_uid, &qf_gid);
+	read_quota_file(sdp, comline, &qf_uid, &qf_gid);
 
 	print_list("fs user ", &fs_uid);
 	print_list("fs group", &fs_gid);
@@ -423,10 +452,10 @@ do_check(commandline_t *comline)
  */
 
 static void
-set_list(commandline_t *comline, int user, osi_list_t *list, int64_t multiplier)
+set_list(struct gfs2_sbd *sdp, commandline_t *comline, int user, 
+	 osi_list_t *list, int64_t multiplier)
 {
 	int fd, fd1;
-	struct gfs2_sb sb;
 	osi_list_t *tmp;
 	values_t *v;
 	uint64_t offset;
@@ -434,20 +463,22 @@ set_list(commandline_t *comline, int user, osi_list_t *list, int64_t multiplier)
 	int error;
 	char quota_file[BUF_SIZE];
 	char sys_q_refresh[BUF_SIZE];
+	char id_str[16];
+
+	strcpy(sdp->path_name, comline->filesystem);
+	check_for_gfs2(sdp);
+	read_superblock(&sdp->sd_sb, sdp);
+	if (!find_gfs2_meta(sdp))
+		mount_gfs2_meta(sdp);
+	lock_for_admin(sdp);
 	
-	check_for_gfs2(comline->filesystem);
-	read_superblock(&sb);
-	if (!find_gfs2_meta(comline->filesystem))
-		mount_gfs2_meta();
-	lock_for_admin();
-	
-	strcpy(quota_file, metafs_path);
+	strcpy(quota_file, sdp->metafs_path);
 	strcat(quota_file, "/quota");
 
 	fd = open(quota_file, O_WRONLY);
 	if (fd < 0) {
-		close(metafs_fd);
-		cleanup();
+		close(sdp->metafs_fd);
+		cleanup_metafs(sdp);
 		die("can't open file %s: %s\n", comline->filesystem,
 		    strerror(errno));
 	}
@@ -460,7 +491,7 @@ set_list(commandline_t *comline, int user, osi_list_t *list, int64_t multiplier)
 		offset += (unsigned long)(&((struct gfs2_quota *)NULL)->qu_value);
 
 		value = v->v_blocks * multiplier;
-		value >>= sb.sb_bsize_shift - 9;
+		value >>= sdp->sd_sb.sb_bsize_shift - 9;
 		value = cpu_to_be64(value);
 
 		lseek(fd, offset, SEEK_SET);
@@ -471,9 +502,11 @@ set_list(commandline_t *comline, int user, osi_list_t *list, int64_t multiplier)
 			goto out;
 		}
 
-		/* Write "1" to sysfs quota refresh file to refresh gfs quotas */
-		sprintf(sys_q_refresh, "%s%s%s", "/sys/fs/gfs2/", sb.sb_locktable, 
-			(user) ? "/quota_refresh_user" : "/quota_refresh_group");
+		/* Write the id to sysfs quota refresh file to refresh gfs quotas */
+		sprintf(sys_q_refresh, "%s%s%s", "/sys/fs/gfs2/",
+			sdp->sd_sb.sb_locktable, 
+			(user) ? "/quota_refresh_user" :
+			"/quota_refresh_group");
 		
 		fd1 = open(sys_q_refresh, O_WRONLY);
 		if (fd1 < 0) {
@@ -481,7 +514,10 @@ set_list(commandline_t *comline, int user, osi_list_t *list, int64_t multiplier)
 				sys_q_refresh, strerror(errno));
 			goto out;
 		}
-		if (write(fd1,(void*)"1", 1) != 1) {
+
+		sprintf(id_str, "%d", comline->id);
+		
+		if (write(fd1,(void*)id_str, strlen(id_str)) != strlen(id_str)) {
 			close(fd1);
 			fprintf(stderr, "failed to write to %s: %s\n", 
 				sys_q_refresh, strerror(errno));
@@ -492,8 +528,8 @@ set_list(commandline_t *comline, int user, osi_list_t *list, int64_t multiplier)
 
 out:
 	close(fd);
-	close(metafs_fd);
-	cleanup();
+	close(sdp->metafs_fd);
+	cleanup_metafs(sdp);
 }
 
 /**
@@ -503,7 +539,7 @@ out:
  */
 
 void
-do_quota_init(commandline_t *comline)
+do_quota_init(struct gfs2_sbd *sdp, commandline_t *comline)
 {
 	dev_t device;
 	osi_list_t fs_uid, fs_gid, qf_uid, qf_gid;
@@ -519,7 +555,7 @@ do_quota_init(commandline_t *comline)
 	device = verify_pathname(comline);
 
 	scan_fs(device, comline->filesystem, &fs_uid, &fs_gid, &hl);
-	read_quota_file(comline, &qf_uid, &qf_gid);
+	read_quota_file(sdp, comline, &qf_uid, &qf_gid);
 
 	type_zalloc(v, values_t, 1);
 	v->v_id = 0;
@@ -536,12 +572,12 @@ do_quota_init(commandline_t *comline)
 	print_list("qf user ", &qf_uid);
 	print_list("qf group", &qf_gid);
 
-	set_list(comline, TRUE, &qf_uid, 0);
-	set_list(comline, FALSE, &qf_gid, 0);
-	set_list(comline, TRUE, &fs_uid, 1);
-	set_list(comline, FALSE, &fs_gid, 1);
+	set_list(sdp, comline, TRUE, &qf_uid, 0);
+	set_list(sdp, comline, FALSE, &qf_gid, 0);
+	set_list(sdp, comline, TRUE, &fs_uid, 1);
+	set_list(sdp, comline, FALSE, &fs_gid, 1);
 	
-	do_sync(comline);
+	do_sync(sdp, comline);
 
-	do_check(comline);
+	do_check(sdp, comline);
 }
