@@ -2,7 +2,7 @@
 *******************************************************************************
 **
 **  Copyright (C) Sistina Software, Inc.  1997-2003  All rights reserved.
-**  Copyright (C) 2004 Red Hat, Inc.  All rights reserved.
+**  Copyright (C) 2004-2007 Red Hat, Inc.  All rights reserved.
 **
 **  This copyrighted material is made available to anyone wishing to use,
 **  modify, copy, or redistribute it subject to the terms and conditions
@@ -77,6 +77,7 @@ print_usage()
 	printf("  warn             set a quota warning value for an ID\n");
 	printf("  check            check the quota file\n");
 	printf("  init             initialize the quota file\n");
+	printf("  reset            reset the quota file\n");
 	printf("\n");
 	printf("Options:\n");
 	printf("  -b               sizes are in FS blocks\n");
@@ -92,13 +93,6 @@ print_usage()
 	printf("  -V               Print program version information, then exit\n");
 }
 
-/**
- * check_for_gfs2 - Check to see if a descriptor is a file on a GFS2 filesystem
- * @fd: the file descriptor
- * @path: the path used to open the descriptor
- *
- */
-
 /**
  * decode_arguments - parse command line arguments
  * @argc: well, it's argc...
@@ -164,7 +158,7 @@ decode_arguments(int argc, char *argv[], commandline_t *comline)
 			break;
 
 		case 'V':
-			printf("gfs2_quota %s (built %s %s)\n", GFS2_RELEASE_NAME,
+			printf("gfs2_quota %s (built %s %s)\n", RELEASE_VERSION,
 			       __DATE__, __TIME__);
 			printf("%s\n", REDHAT_COPYRIGHT);
 			exit(EXIT_SUCCESS);
@@ -221,6 +215,10 @@ decode_arguments(int argc, char *argv[], commandline_t *comline)
 			if (comline->operation)
 				die("can't specify two operations\n");
 			comline->operation = GQ_OP_INIT;
+		} else if (strcmp(argv[optind], "reset") == 0) {
+			if (comline->operation)
+				die("can't specify two operations\n");
+			comline->operation = GQ_OP_RESET;
 		} else
 			die("unknown option %s\n", argv[optind]);
 
@@ -286,167 +284,16 @@ print_quota(commandline_t *comline,
 	}
 }
 
-void
-lock_for_admin()
-{
-        int error;
-
-        for (;;) {
-
-                metafs_fd = open(metafs_path, O_RDONLY | O_NOFOLLOW);
-                if (metafs_fd < 0)
-                        die("can't open %s: %s\n",
-                            metafs_path, strerror(errno));
-
-                error = flock(metafs_fd, LOCK_EX);
-                if (error)
-                        die("can't flock %s: %s\n", metafs_path,
-                            strerror(errno));
-
-                break;
-        }
-}
-
-int 
-find_gfs2_meta(const char *mnt)
-{
-	FILE *fp = fopen("/proc/mounts", "r");
-	char name[] = "gfs2meta";
-	char buffer[BUF_SIZE];
-	char fstype[80], mfsoptions[BUF_SIZE];
-	char meta_device[BUF_SIZE];
-	int fsdump, fspass;
-
-	metafs_mounted = 0;
-
-	if (fp == NULL) {
-		perror("open: /proc/mounts");
-		exit(EXIT_FAILURE);
-	}
-	while ((fgets(buffer, 4095, fp)) != NULL) {
-		buffer[4095] = 0;
-		if (strstr(buffer, name) == 0)
-			continue;
-
-		if (sscanf(buffer, "%s %s %s %s %d %d", meta_device, 
-			   metafs_path, fstype,mfsoptions, &fsdump, 
-			   &fspass) != 6)
-			continue;
-		
-		if (strcmp(meta_device, device_name) != 0
-		    && strcmp(meta_device, mnt) != 0)
-			continue;
-		
-		metafs_mounted = 1;
-		
-		fclose(fp);
-		return TRUE;
-	}
-	fclose(fp);
-	return FALSE;
-}
-
-static int
-dir_exists(const char *dir)
-{
-	int fd, ret;
-	struct stat statbuf;
-	fd = open(dir, O_RDONLY);
-	if (fd<0) { 
-		if (errno == ENOENT)
-			return 0;
-		die("Couldn't open %s : %s\n", dir, strerror(errno));
-	}
-	ret = fstat(fd, &statbuf);
-	if (ret)
-		die("stat failed on %s : %s\n", dir, strerror(errno));
-	if (S_ISDIR(statbuf.st_mode)) {
-		close(fd);
-		return 1;
-	}
-	close(fd);
-	die("%s exists, but is not a directory. Cannot mount metafs here\n", dir);
-}
-
 void 
-mount_gfs2_meta()	 
-{
-	int ret;
-	/* mount the meta fs */
-	if (!dir_exists(meta_mount)) {
-		ret = mkdir(meta_mount, 0700);
-		if (ret)
-			die("Couldn't create %s : %s\n", meta_mount,
-			    strerror(errno));
-	}
-		
-	ret = mount(device_name, meta_mount, "gfs2meta", 0, NULL);
-	if (ret)
-		die("Couldn't mount %s : %s\n", meta_mount,
-		    strerror(errno));
-	strcpy(metafs_path, meta_mount);
-}
-
-void
-cleanup()
-{
-	int ret;
-	if (!metafs_mounted) { /* was mounted by us */
-		ret = umount(meta_mount);
-		if (ret)
-			fprintf(stderr, "Couldn't unmount %s : %s\n", meta_mount, 
-			    strerror(errno));
-	}
-}
-
-void
-check_for_gfs2(const char *mnt)
-{
-	FILE *fp = fopen("/proc/mounts", "r");
-	char buffer[4096];
-	char fstype[80];
-	int fsdump, fspass, ret;
-
-	if (fp == NULL) {
-		perror("open: /proc/mounts");
-		exit(EXIT_FAILURE);
-	}
-	while ((fgets(buffer, 4095, fp)) != NULL) {
-		buffer[4095] = 0;
-
-		if (strstr(buffer, "0") == 0)
-			continue;
-
-		if ((ret = sscanf(buffer, "%s %s %s %s %d %d", device_name, fspath, 
-				  fstype, fsoptions, &fsdump, &fspass)) != 6) 
-			continue;
-
-		if (strcmp(fstype, "gfs2") != 0)
-			continue;
-
-		if (strcmp(fspath, mnt) != 0)
-			continue;
-
-		fclose(fp);
-		if (strncmp(device_name, "/dev/loop", 9) == 0)
-			die("Cannot add journal(s) to a loopback GFS mount\n");
-		
-		return;
-	}
-	fclose(fp);
-	die("gfs2 Filesystem %s not found\n", mnt);
-}
-
-void 
-read_superblock(struct gfs2_sb *sb)
+read_superblock(struct gfs2_sb *sb, struct gfs2_sbd *sdp)
 {
 	int fd;
 	char buf[PATH_MAX];
 	
-	fd = open(device_name, O_RDONLY);
+	fd = open(sdp->device_name, O_RDONLY);
 	if (fd < 0) {
 		die("Could not open the block device %s: %s\n",
-			device_name, strerror(errno));
+			sdp->device_name, strerror(errno));
 	}
 	do_lseek(fd, 0x10 * 4096);
 	do_read(fd, buf, PATH_MAX);
@@ -455,6 +302,239 @@ read_superblock(struct gfs2_sb *sb)
 	close(fd);
 }
 
+inline void 
+read_quota_internal(int fd, uint32_t id, int id_type, struct gfs2_quota *q)
+{
+	/* seek to the appropriate offset in the quota file and read the 
+	   quota info */
+	uint64_t offset;
+	char buf[256];
+	int error;
+	if (id_type == GQ_ID_USER)
+		offset = (2 * (uint64_t)id) * sizeof(struct gfs2_quota);
+	else
+		offset = (2 * (uint64_t)id + 1) * sizeof(struct gfs2_quota);
+	lseek(fd, offset, SEEK_SET);
+	error = read(fd, buf, sizeof(struct gfs2_quota));
+	if (error < 0)
+		die("failed to read from quota file: %s\n", strerror(errno));
+	if (error != sizeof(struct gfs2_quota))
+		die("Couldn't read %d bytes from quota file at offset %llu\n",
+		    sizeof(struct gfs2_quota), offset);
+	gfs2_quota_in(q, buf);
+}
+
+inline void 
+write_quota_internal(int fd, uint32_t id, int id_type, struct gfs2_quota *q)
+{
+	/* seek to the appropriate offset in the quota file and read the
+	   quota info */
+	uint64_t offset;
+	char buf[256];
+	int error;
+	if (id_type == GQ_ID_USER)
+		offset = (2 * (uint64_t)id) * sizeof(struct gfs2_quota);
+	else
+		offset = (2 * (uint64_t)id + 1) * sizeof(struct gfs2_quota);
+	lseek(fd, offset, SEEK_SET);
+	gfs2_quota_out(q, buf);
+	error = write(fd, buf, sizeof(struct gfs2_quota));
+	if (error != sizeof(struct gfs2_quota))
+		die("failed to write to quota file: %s\n", strerror(errno));
+}
+
+/**
+ * get_last_quota_id - Get the last quota in the quota file
+ * @fd: an open file descriptor of the quota file
+ * @id_type: GQ_ID_USER or GQ_ID_GROUP
+ * @max_id: return the maximum id obtained
+ */
+void 
+get_last_quota_id(int fd, uint32_t *max_id)
+{
+	/* stat(2) the quota file to find how big it is. This will give us a
+	 * a rough idea of what the last valid quota uid/gid is. It is possible 
+	 * that the last quota in the file belongs to a group with gid:x and 
+	 * the corresponding user quota with uid:x doesn't exist. In such cases
+	 * we still return x as max_id. This max_id is ONLY A HINT. If used 
+	 * as a terminating condition of a loop, another condition should also
+	 * be specified.
+	 */
+	struct stat st;
+	uint32_t qsize = sizeof(struct gfs2_quota);
+	uint64_t size;
+	if (fstat(fd, &st))
+		die("failed to stat the quota file: %s\n", strerror(errno));
+	size = st.st_size;
+	if (!size)
+		die("error: quota file is truncated to zero!\n");
+	if (size % qsize) {
+		printf("warning: quota file size not a multiple of "
+		       "struct gfs2_quota\n");
+		size = qsize * (size / qsize);
+	}
+	*max_id = (size - 1) / (2 * qsize);
+}
+
+/**
+ * is_valid_quota_list - Check if we have a valid quota list
+ * @fd: an open file descriptor of the quota file
+ * Returns 0 or 1.
+ */
+int 
+is_valid_quota_list(int fd)
+{
+	/* This is a slow test to determine if the quotas are in a 
+	 * linked list. We should come up with something better
+	 * Quota linked list format is identified by the following.
+	 * step1: Get the maximum groupid and userid having valid
+	 *        quotas in the quota file.
+	 * step2: Obtain the size of the quota file. The size of the 
+	 *        quota file (position of the last valid quota) 
+	 *        determines the last user/group id.
+	 * step3: If we can obtain the last valid quota through the 
+	 *        lists, then our lists are good. Else, the lists are 
+	 *        either corrupt or an older quota file format is in use
+	 */
+	int id_type = GQ_ID_GROUP;
+	uint32_t id = 0, prev, ulast = 0, glast = 0, max;
+	struct gfs2_quota q;
+
+	get_last_quota_id(fd, &max);
+again:
+	do {
+		read_quota_internal(fd, id, id_type, &q);
+		prev = id;
+		id = q.qu_ll_next;
+		if (id > max)
+			return 0;
+	} while (id && id > prev);
+
+	if (id && id <= prev)
+		return 0;
+
+	if (id_type == GQ_ID_GROUP)
+		glast = prev;
+	else
+		ulast = prev;
+
+	if (id_type == GQ_ID_GROUP) {
+		id_type = GQ_ID_USER;
+		id = 0;
+		goto again;
+	}
+
+	if (glast != max && ulast != max)
+		return 0;
+	
+	return 1;
+}
+
+void 
+print_quota_list_warning()
+{
+	printf("\nWarning: This filesystem doesn't seem to have the new quota "
+	       "list format or the quota list is corrupt. list, check and init "
+	       "operation performance will suffer due to this. It is recommended "
+	       "that you run the 'gfs2_quota reset' operation to reset the quota "
+	       "file. All current quota information will be lost and you will "
+	       "have to reassign all quota limits and warnings\n\n"); 
+}
+
+/**
+ * adjust_quota_list - Adjust the quota linked list
+ * @fd: The quota file descriptor
+ * @comline: the struct containing the parsed command line arguments
+ */
+static void
+adjust_quota_list(int fd, commandline_t *comline)
+{
+	uint32_t prev = 0, next = 0, id = comline->id;
+	struct gfs2_quota tmpq, q;
+	int id_type = comline->id_type;
+	
+	if (id == 0) /* root quota, don't do anything */
+		goto out;
+	/* We just wrote the quota for id in do_set(). Get it */
+	next = 0;
+	do {
+		read_quota_internal(fd, next, id_type, &q);
+		prev = next;
+		next = q.qu_ll_next;
+		if (prev == id) /* no duplicates, bail */
+			goto out;
+		if (prev < id && id < next) /* gotcha! */
+			break;
+	} while(next && next > prev);
+	read_quota_internal(fd, id, id_type, &tmpq);
+	tmpq.qu_ll_next = next;
+	q.qu_ll_next = id;
+	write_quota_internal(fd, id, id_type, &tmpq);
+	write_quota_internal(fd, prev, id_type, &q);
+
+out:
+	return;
+}
+
+/**
+ * do_reset - Reset all the quota data for a filesystem
+ * @comline: the struct containing the parsed command line arguments
+ */
+
+static void
+do_reset(struct gfs2_sbd *sdp, commandline_t *comline)
+{
+	int fd;
+	char quota_file[BUF_SIZE], c;
+	struct gfs2_quota q;
+
+	if (!*comline->filesystem)
+		die("need a filesystem to work on\n");
+
+	printf("This operation will permanently erase all quota information. "
+	       "You will have to re-assign all quota limit/warn values. "
+	       "Proceed [y/N]? ");
+	c = getchar();
+	if (c != 'y' && c != 'Y')
+		return;
+
+	strcpy(sdp->path_name, comline->filesystem);
+	check_for_gfs2(sdp);
+	read_superblock(&sdp->sd_sb, sdp);
+	if (!find_gfs2_meta(sdp))
+		mount_gfs2_meta(sdp);
+	lock_for_admin(sdp);
+	
+	strcpy(quota_file, sdp->metafs_path);
+	strcat(quota_file, "/quota");
+
+	fd = open(quota_file, O_RDWR);
+	if (fd < 0) {
+		close(sdp->metafs_fd);
+		cleanup_metafs(sdp);
+		die("can't open file %s: %s\n", quota_file,
+		    strerror(errno));
+	}
+
+	read_quota_internal(fd, 0, GQ_ID_USER, &q);
+	q.qu_ll_next = 0;
+	write_quota_internal(fd, 0, GQ_ID_USER, &q);
+
+	read_quota_internal(fd, 0, GQ_ID_GROUP, &q);
+	q.qu_ll_next = 0;
+	write_quota_internal(fd, 0, GQ_ID_GROUP, &q);
+
+	/* truncate the quota file such that only the first
+	 * two quotas(uid=0 and gid=0) remain.
+	 */
+	if (ftruncate(fd, (sizeof(struct gfs2_quota)) * 2))
+		die("couldn't truncate quota file %s\n", strerror(errno));
+	
+	close(fd);
+	close(sdp->metafs_fd);
+	cleanup_metafs(sdp);
+}
+
 /**
  * do_list - List all the quota data for a filesystem
  * @comline: the struct containing the parsed command line arguments
@@ -462,37 +542,62 @@ read_superblock(struct gfs2_sb *sb)
  */
 
 static void 
-do_list(commandline_t *comline)
+do_list(struct gfs2_sbd *sdp, commandline_t *comline)
 {
 	int fd;
-	struct gfs2_sb sb;
 	struct gfs2_quota q;
 	char buf[sizeof(struct gfs2_quota)];
 	uint64_t offset;
-	uint32_t id;
+	uint32_t id, prev, maxid;
 	int pass = 0;
 	int error = 0;
 	char quota_file[BUF_SIZE];
+	int id_type = comline->id_type;
 	
 	if (!*comline->filesystem)
 		die("need a filesystem to work on\n");
 
-	check_for_gfs2(comline->filesystem);
-	read_superblock(&sb);
-	if (!find_gfs2_meta(comline->filesystem))
-		mount_gfs2_meta();
-	lock_for_admin();
+	strcpy(sdp->path_name, comline->filesystem);
+	check_for_gfs2(sdp);
+	read_superblock(&sdp->sd_sb, sdp);
+	if (!find_gfs2_meta(sdp))
+		mount_gfs2_meta(sdp);
+	lock_for_admin(sdp);
 	
-	strcpy(quota_file, metafs_path);
+	strcpy(quota_file, sdp->metafs_path);
 	strcat(quota_file, "/quota");
 
 	fd = open(quota_file, O_RDONLY);
 	if (fd < 0) {
-		close(metafs_fd);
-		cleanup();
+		close(sdp->metafs_fd);
+		cleanup_metafs(sdp);
 		die("can't open file %s: %s\n", quota_file,
 		    strerror(errno));
 	}
+	
+	if (!is_valid_quota_list(fd)) {
+		print_quota_list_warning();
+		goto do_old_school;
+	}
+	get_last_quota_id(fd, &maxid);
+	
+	for (pass=0; pass<2; pass++) {
+		id = 0;
+		id_type = pass ? GQ_ID_GROUP : GQ_ID_USER;
+		
+		do {
+			read_quota_internal(fd, id, id_type, &q);
+			prev = id;
+			if (q.qu_limit || q.qu_warn || q.qu_value)
+				print_quota(comline, 
+					    id_type == GQ_ID_USER ? TRUE : FALSE, 
+					    id, &q, &sdp->sd_sb);
+			id = q.qu_ll_next;
+		} while(id && id > prev && id <= maxid);
+	}
+	goto out;
+
+do_old_school:
 	for (pass=0; pass<2; pass++) {
 		if (!pass)
 			offset = 0;
@@ -512,15 +617,15 @@ do_list(commandline_t *comline)
 
 			if (q.qu_limit || q.qu_warn || q.qu_value)
 				print_quota(comline, (pass) ? FALSE : TRUE, id,
-					    &q, &sb);
+					    &q, &sdp->sd_sb);
 
 			offset += 2 * sizeof(struct gfs2_quota);
 		} while (error == sizeof(struct gfs2_quota));
 	}
-
+out:
 	close(fd);
-	close(metafs_fd);
-	cleanup();
+	close(sdp->metafs_fd);
+	cleanup_metafs(sdp);
 }
 
 /**
@@ -531,29 +636,30 @@ do_list(commandline_t *comline)
  */
 
 static void 
-do_get_one(commandline_t *comline, char *filesystem)
+do_get_one(struct gfs2_sbd *sdp, commandline_t *comline, char *filesystem)
 {
 	int fd;
 	char buf[256];
 	struct gfs2_quota q;
-	struct gfs2_sb sb;
 	uint64_t offset;
 	int error;
+	uint32_t maxid;
 	char quota_file[BUF_SIZE];
 
-	check_for_gfs2(filesystem);
-	read_superblock(&sb);
-	if (!find_gfs2_meta(filesystem))
-		mount_gfs2_meta();
-	lock_for_admin();
+	strcpy(sdp->path_name, filesystem);
+	check_for_gfs2(sdp);
+	read_superblock(&sdp->sd_sb, sdp);
+	if (!find_gfs2_meta(sdp))
+		mount_gfs2_meta(sdp);
+	lock_for_admin(sdp);
 	
-	strcpy(quota_file, metafs_path);
+	strcpy(quota_file, sdp->metafs_path);
 	strcat(quota_file, "/quota");
 
 	fd = open(quota_file, O_RDONLY);
 	if (fd < 0) {
-		close(metafs_fd);
-		cleanup();
+		close(sdp->metafs_fd);
+		cleanup_metafs(sdp);
 		die("can't open file %s: %s\n", quota_file,
 		    strerror(errno));
 	}
@@ -565,12 +671,16 @@ do_get_one(commandline_t *comline, char *filesystem)
 
 	memset(&q, 0, sizeof(struct gfs2_quota));
 	
+	get_last_quota_id(fd, &maxid);
+	if (comline->id > maxid)
+		goto print_empty_quota;
+
 	lseek(fd, offset, SEEK_SET);
 	error = read(fd, buf, sizeof(struct gfs2_quota));
 	if (error < 0) {
 		close(fd);
-		close(metafs_fd);
-		cleanup();
+		close(sdp->metafs_fd);
+		cleanup_metafs(sdp);
 		die("can't get quota info (%d): %s\n",
 		    error, strerror(errno));
 	}
@@ -578,13 +688,14 @@ do_get_one(commandline_t *comline, char *filesystem)
 	gfs2_quota_in(&q, buf);
 
 
+print_empty_quota:
 	print_quota(comline,
 		    (comline->id_type == GQ_ID_USER), comline->id,
-		    &q, &sb);
+		    &q, &sdp->sd_sb);
 
 	close(fd);
-	close(metafs_fd);
-	cleanup();	
+	close(sdp->metafs_fd);
+	cleanup_metafs(sdp);	
 }
 
 /**
@@ -594,12 +705,12 @@ do_get_one(commandline_t *comline, char *filesystem)
  */
 
 static void
-do_get(commandline_t *comline)
+do_get(struct gfs2_sbd *sdp, commandline_t *comline)
 {
 	int first = TRUE;
 
 	if (*comline->filesystem)
-		do_get_one(comline, comline->filesystem);
+		do_get_one(sdp, comline, comline->filesystem);
 	else {
 		char buf[256], device[256], path[256], type[256];
 		FILE *file;
@@ -620,7 +731,7 @@ do_get(commandline_t *comline)
 				printf("\n");
 
 			printf("%s\n", path);
-			do_get_one(comline, path);
+			do_get_one(sdp, comline, path);
 		}
 
 		fclose(file);
@@ -633,16 +744,16 @@ do_get(commandline_t *comline)
  *
  */
 static void 
-do_sync_one(char *filesystem)
+do_sync_one(struct gfs2_sbd *sdp, char *filesystem)
 {
 	int fd;
-	struct gfs2_sb sb;
 	char sys_quota_sync[PATH_MAX];
 
-	check_for_gfs2(filesystem);
-	read_superblock(&sb);
+	strcpy(sdp->path_name, filesystem);
+	check_for_gfs2(sdp);
+	read_superblock(&sdp->sd_sb, sdp);
 	sprintf(sys_quota_sync, "%s%s%s", 
-		"/sys/fs/gfs2/", sb.sb_locktable, "/quota_sync");
+		"/sys/fs/gfs2/", sdp->sd_sb.sb_locktable, "/quota_sync");
 	
 	fd = open(sys_quota_sync, O_WRONLY);
 	if (fd < 0)
@@ -662,12 +773,12 @@ do_sync_one(char *filesystem)
  */
 
 void
-do_sync(commandline_t *comline)
+do_sync(struct gfs2_sbd *sdp, commandline_t *comline)
 {
 	sync();
 
 	if (*comline->filesystem)
-		do_sync_one(comline->filesystem);
+		do_sync_one(sdp, comline->filesystem);
 	else {
 		char buf[256], device[256], path[256], type[256];
 		FILE *file;
@@ -682,7 +793,7 @@ do_sync(commandline_t *comline)
 			if (strcmp(type, "gfs2") != 0)
 				continue;
 
-			do_sync_one(path);
+			do_sync_one(sdp, path);
 		}
 
 		fclose(file);
@@ -696,38 +807,45 @@ do_sync(commandline_t *comline)
  */
 
 static void
-do_set(commandline_t *comline)
+do_set(struct gfs2_sbd *sdp, commandline_t *comline)
 {
 	int fd, fd1;
 	uint64_t offset;
-	struct gfs2_sb sb;
 	uint64_t new_value;
-	int error;
+	int error, adj_flag = 0;;
 	char quota_file[BUF_SIZE];
 	char sys_q_refresh[BUF_SIZE];
+	char id_str[16];
+	struct stat stat_buf;
 	
 	if (!*comline->filesystem)
 		die("need a filesystem to work on\n");
 	if (!comline->new_value_set)
 		die("need a new value\n");
 
-	check_for_gfs2(comline->filesystem);
-	read_superblock(&sb);
-	if (!find_gfs2_meta(comline->filesystem))
-		mount_gfs2_meta();
-	lock_for_admin();
+	strcpy(sdp->path_name, comline->filesystem);
+	check_for_gfs2(sdp);
+	read_superblock(&sdp->sd_sb, sdp);
+	if (!find_gfs2_meta(sdp))
+		mount_gfs2_meta(sdp);
+	lock_for_admin(sdp);
 	
-	strcpy(quota_file, metafs_path);
+	strcpy(quota_file, sdp->metafs_path);
 	strcat(quota_file, "/quota");
 
-	fd = open(quota_file, O_WRONLY);
+	fd = open(quota_file, O_RDWR);
 	if (fd < 0) {
-		close(metafs_fd);
-		cleanup();
+		close(sdp->metafs_fd);
+		cleanup_metafs(sdp);
 		die("can't open file %s: %s\n", quota_file,
 		    strerror(errno));
 	}
 	
+	if (is_valid_quota_list(fd))
+		adj_flag = 1;
+	else
+		print_quota_list_warning();
+
 	switch (comline->id_type) {
 	case GQ_ID_USER:
 		offset = (2 * (uint64_t)comline->id) * sizeof(struct gfs2_quota);
@@ -742,30 +860,18 @@ do_set(commandline_t *comline)
 		goto out;
 	}
 
-	switch (comline->operation) {
-	case GQ_OP_LIMIT:
-		offset += (unsigned long)(&((struct gfs2_quota *) NULL)->qu_limit);
-		break;
-
-	case GQ_OP_WARN:
-		offset += (unsigned long)(&((struct gfs2_quota *) NULL)->qu_warn);
-		break;
-
-	default:
-		fprintf(stderr, "invalid operation\n");
-		goto out;
-	};
-
 	switch (comline->units) {
 	case GQ_UNITS_MEGABYTE:
-		new_value = comline->new_value << (20 - sb.sb_bsize_shift);
+		new_value =
+			comline->new_value << (20 - sdp->sd_sb.sb_bsize_shift);
 		break;
 
 	case GQ_UNITS_KILOBYTE:
-		if (sb.sb_bsize == 512)
+		if (sdp->sd_sb.sb_bsize == 512)
 			new_value = comline->new_value * 2;
 		else
-			new_value = comline->new_value >> (sb.sb_bsize_shift - 10);
+			new_value = comline->new_value >>
+				(sdp->sd_sb.sb_bsize_shift - 10);
 		break;
 
 	case GQ_UNITS_FSBLOCK:
@@ -773,7 +879,8 @@ do_set(commandline_t *comline)
 		break;
 
 	case GQ_UNITS_BASICBLOCK:
-		new_value = comline->new_value >> (sb.sb_bsize_shift - 9);
+		new_value = comline->new_value >>
+			(sdp->sd_sb.sb_bsize_shift - 9);
 		break;
 
 	default:
@@ -782,17 +889,78 @@ do_set(commandline_t *comline)
 	}
 
 	new_value = cpu_to_be64(new_value);
-
-	lseek(fd, offset, SEEK_SET);
-	error = write(fd, (char*)&new_value, sizeof(uint64_t));
-	if (error != sizeof(uint64_t)) {
-		fprintf(stderr, "can't write quota file (%d): %s\n",
-		    error, strerror(errno));
+	/*
+	 * Hack to force writing the entire gfs2_quota structure to 
+	 * the quota file instead of just the limit or warn values.
+	 * This is because of a bug in gfs2 which doesn't extend 
+	 * the quotafile appropriately to write the usage value of a 
+	 * given id. For instance, if you write a limit value (8 bytes) 
+	 * for userid x at offset 2*x, gfs2 will not extend the file and write 
+	 * 8 bytes at offset (2*x + 16) when it has to update the usage 
+	 * value for id x. Therefore, we extend the quota file to 
+	 * a struct gfs2_quota boundary. i.e. The size of the quota file
+	 * will always be a multiple of sizeof(struct gfs2_quota)
+	 */
+	if (fstat(fd, &stat_buf)) {
+		fprintf(stderr, "stat failed: %s\n", strerror(errno));
 		goto out;
 	}
+	if (stat_buf.st_size < (offset + sizeof(struct gfs2_quota))) {
+		struct gfs2_quota tmp;
+		memset((void*)(&tmp), 0, sizeof(struct gfs2_quota));
+		switch (comline->operation) {
+		case GQ_OP_LIMIT:
+			tmp.qu_limit = new_value; break;
+		case GQ_OP_WARN:
+			tmp.qu_warn = new_value; break;
+		}
+		
+		lseek(fd, offset, SEEK_SET);
+		error = write(fd, (void*)(&tmp), sizeof(struct gfs2_quota));
+		if (error != sizeof(struct gfs2_quota)) {
+			fprintf(stderr, "can't write quota file (%d): %s\n", 
+				error, strerror(errno));
+			goto out;
+		}
+		/* Also, if the id type is USER, append another empty 
+		 * struct gfs2_quota for the GROUP with the same id
+		 */
+		if (comline->id_type == GQ_ID_USER) {
+			memset((void*)(&tmp), 0, sizeof(struct gfs2_quota));
+			error = write(fd, (void*)(&tmp), sizeof(struct gfs2_quota));
+			if (error != sizeof(struct gfs2_quota)) {
+				fprintf(stderr, "can't write quota file (%d): %s\n", 
+					error, strerror(errno));
+				goto out;
+			}
+		}
+	} else {
+		switch (comline->operation) {
+		case GQ_OP_LIMIT:
+			offset += (unsigned long)(&((struct gfs2_quota *) NULL)->qu_limit);
+			break;
+			
+		case GQ_OP_WARN:
+			offset += (unsigned long)(&((struct gfs2_quota *) NULL)->qu_warn);
+			break;
+			
+		default:
+			fprintf(stderr, "invalid operation\n");
+			goto out;
+		};
 
-	/* Write "1" to sysfs quota refresh file to refresh gfs quotas */
-	sprintf(sys_q_refresh, "%s%s%s", "/sys/fs/gfs2/", sb.sb_locktable, 
+		lseek(fd, offset, SEEK_SET);
+		error = write(fd, (char*)&new_value, sizeof(uint64_t));
+		if (error != sizeof(uint64_t)) {
+			fprintf(stderr, "can't write quota file (%d): %s\n",
+				error, strerror(errno));
+			goto out;
+		}
+	}
+
+	/* Write the id to sysfs quota refresh file to refresh gfs quotas */
+	sprintf(sys_q_refresh, "%s%s%s", "/sys/fs/gfs2/",
+		sdp->sd_sb.sb_locktable, 
 		comline->id_type == GQ_ID_USER ? "/quota_refresh_user" : 
 		"/quota_refresh_group");
 	
@@ -802,18 +970,22 @@ do_set(commandline_t *comline)
 			strerror(errno));
 		goto out;
 	}
+
+	sprintf(id_str, "%d", comline->id);
 	
-	if (write(fd1,(void*)"1", 1) != 1) {
+	if (write(fd1,(void*)id_str, strlen(id_str)) != strlen(id_str)) {
 		close(fd1);
 		fprintf(stderr, "failed to write to %s: %s\n", 
 			sys_q_refresh, strerror(errno));
 		goto out;
 	}
 	close(fd1);
+	if (adj_flag)
+		adjust_quota_list(fd, comline);
 out:
 	close(fd);
-	close(metafs_fd);
-	cleanup();
+	close(sdp->metafs_fd);
+	cleanup_metafs(sdp);
 }
 
 /**
@@ -827,51 +999,61 @@ out:
 int
 main(int argc, char *argv[])
 {
+    struct gfs2_sbd sbd, *sdp = &sbd;
 	commandline_t comline;
 
 	prog_name = argv[0];
-	metafs_mounted = 0;
+	sdp->metafs_mounted = 0;
 
+	memset(sdp, 0, sizeof(struct gfs2_sbd));
 	memset(&comline, 0, sizeof(commandline_t));
 
 	decode_arguments(argc, argv, &comline);
+	sdp->path_name = (char*) malloc(512);
+	if (!sdp->path_name)
+		die("Can't malloc! %s\n", strerror(errno));
 
 	switch (comline.operation) {
 	case GQ_OP_LIST:
-		do_list(&comline);
+		do_list(sdp, &comline);
 		break;
 
 	case GQ_OP_GET:
-		do_get(&comline);
+		do_get(sdp, &comline);
 		break;
 
 	case GQ_OP_LIMIT:
 	case GQ_OP_WARN:
-		do_set(&comline);
+		do_set(sdp, &comline);
 		break;
 
 	case GQ_OP_SYNC:
-		do_sync(&comline);
+		do_sync(sdp, &comline);
 		break;
 
 	case GQ_OP_CHECK:
-		do_sync(&comline);
-		do_check(&comline);
+		do_sync(sdp, &comline);
+		do_check(sdp, &comline);
 		break;
 
 	case GQ_OP_INIT:
-		do_sync(&comline);
-		do_quota_init(&comline);
+		do_sync(sdp, &comline);
+		do_quota_init(sdp, &comline);
 		break;
 
+	case GQ_OP_RESET:
+		do_reset(sdp, &comline);
+		break;
 	default:
 		if (!comline.id_type) {
 			comline.id_type = GQ_ID_USER;
 			comline.id = geteuid();
 		}
-		do_get(&comline);
+		do_get(sdp, &comline);
 		break;
 	}
+	
+	free(sdp->path_name);
 
 	exit(EXIT_SUCCESS);
 }
