@@ -10,10 +10,14 @@
 *******************************************************************************
 ******************************************************************************/
 
+#include <signal.h>
+#include <time.h>
+
 #include "gd_internal.h"
 
 #define OPTION_STRING			"DhVv"
 #define LOCKFILE_NAME			"/var/run/groupd.pid"
+#define LOG_FILE				"/var/log/groupd.log"
 
 extern struct list_head recovery_sets;
 
@@ -69,6 +73,8 @@ static int do_read(int fd, void *buf, size_t count)
 			return -1;
 		if (rv == -1 && errno == EINTR)
 			continue;
+		if (rv == -1)
+			return -1;
 		off += rv;
 	}
 	return 0;
@@ -132,6 +138,8 @@ int check_uncontrolled_groups(void)
 	char *argv[4];
 	int status, rv = 0;
 
+	/* FIXME: ignore gfs/gfs2 nolock fs's */
+
 	rv += kernel_instance_count("/sys/kernel/dlm");
 	rv += kernel_instance_count("/sys/fs/gfs");
 	rv += kernel_instance_count("/sys/fs/gfs2");
@@ -230,6 +238,7 @@ void app_start(app_t *a)
 		type = NODE_FAILED;
 	else {
 		/* report error */
+		type = -1;
 	}
 
 	/* start <name> <event_nr> <type> <count> <memb0> <memb1>... */
@@ -314,6 +323,7 @@ enum {
 	DO_GET_GROUPS,
 	DO_GET_GROUP,
 	DO_DUMP,
+	DO_LOG,
 };
 
 int get_action(char *buf)
@@ -361,6 +371,9 @@ int get_action(char *buf)
 	if (!strncmp(act, "dump", 16))
 		return DO_DUMP;
 
+	if (!strncmp(act, "log", 16))
+		return DO_LOG;
+
 	return -1;
 }
 
@@ -439,6 +452,7 @@ static void do_setup(int ci, int argc, char **argv)
 static void copy_group_data(group_t *g, group_data_t *data)
 {
 	node_t *node;
+	event_t *ev;
 	int i = 0;
 
 	strncpy(data->client_name, client[g->app->client].type, 32);
@@ -462,16 +476,23 @@ static void copy_group_data(group_t *g, group_data_t *data)
 			else
 				data->event_local_status = -1;
 		}
-
-		if (g->app->sent_event_id)
-			log_group(g, "sent_event_id %llx",
-				  g->app->sent_event_id);
 	}
 
 	data->member_count = g->app->node_count;
 	list_for_each_entry(node, &g->app->nodes, list) {
 		data->members[i] = node->nodeid;
 		i++;
+
+		if (node->nodeid == our_nodeid)
+			data->member = 1;
+	}
+
+	/* we're in the member list but are still joining */
+	if (data->member) {
+		ev = g->app->current_event;
+		if (ev && is_our_join(ev) &&
+		    (ev->state <= EST_JOIN_ALL_STARTED))
+			data->member = 0;
 	}
 }
 
@@ -479,30 +500,26 @@ static int do_get_groups(int ci, int argc, char **argv)
 {
 	group_t *g;
 	group_data_t *data;
-	int i, rv, len, count = 0, max = atoi(argv[1]);
+	int rv, count = 0, max = atoi(argv[1]);
 
-	list_for_each_entry(g, &gd_groups, list)
-		count++;
-	if (count > max)
-		count = max;
-	/* if no groups, send back one empty data struct */
-	if (!count)
-		count = 1;
+	data = malloc(sizeof(group_data_t));
+	count = 0;
 
-	len = count * sizeof(group_data_t);
-	data = malloc(len);
-	memset(data, 0, len);
-
-	i = 0;
 	list_for_each_entry(g, &gd_groups, list) {
-		copy_group_data(g, &data[i]);
-		i++;
+		memset(data, 0, sizeof(group_data_t));
+		copy_group_data(g, data);
+		rv = do_write(client[ci].fd, data, sizeof(group_data_t));
+		if (rv < 0) {
+			log_print("do_get_groups write error");
+			break;
+		}
+		count++;
+		if (count >= max)
+			break;
 	}
-
-	rv = do_write(client[ci].fd, data, len);
-	if (rv < 0)
-		log_print("do_get_groups write error");
-
+	/* Now write an empty one indicating there aren't anymore: */
+	memset(data, 0, sizeof(group_data_t));
+	rv = do_write(client[ci].fd, data, sizeof(group_data_t));
 	free(data);
 	return 0;
 }
@@ -534,21 +551,27 @@ static int do_get_group(int ci, int argc, char **argv)
 	return 0;
 }
 
-static int do_dump(int ci, int argc, char **argv)
+static int do_dump(int fd)
 {
-	int rv, len;
+	int rv, len = DUMP_SIZE;
 
 	if (dump_wrap) {
 		len = DUMP_SIZE - dump_point;
-		rv = write(client[ci].fd, dump_buf + dump_point, len);
-		if (rv != len)
-			log_print("write error %d errno %d", rv, errno);
+		rv = do_write(fd, dump_buf + dump_point, len);
+		if (rv < 0)
+			log_print("dump write error %d errno %d", rv, errno);
+		len = dump_point;
 	}
-	len = dump_point;
 
-	rv = write(client[ci].fd, dump_buf, len);
-	if (rv != len)
-		log_print("write error %d errno %d", rv, errno);
+	rv = do_write(fd, dump_buf, len);
+	if (rv < 0)
+		log_print("dump write error %d errno %d", rv, errno);
+	return 0;
+}
+
+static int do_log(int fd, const char *comment)
+{
+	log_print("%s", comment);
 	return 0;
 }
 
@@ -643,11 +666,15 @@ static void process_connection(int ci)
 		break;
 
 	case DO_DUMP:
-		do_dump(ci, argc, argv);
+		do_dump(client[ci].fd);
+		break;
+
+	case DO_LOG:
+		do_log(client[ci].fd, &buf[4]);
 		break;
 
 	default:
-		log_print("unknown action %d client %d bytes %d", act, ci, rv);
+		log_print("unknown action %d client %d", act, ci);
 		log_print("invalid message: \"%s\"", buf);
 	}
 }
@@ -878,6 +905,18 @@ static void decode_arguments(int argc, char **argv)
 	}
 }
 
+void set_oom_adj(int val)
+{
+	FILE *fp;
+
+	fp = fopen("/proc/self/oom_adj", "w");
+	if (!fp)
+		return;
+
+	fprintf(fp, "%i", val);
+	fclose(fp);
+}
+
 void set_scheduler(void)
 {
 	struct sched_param sched_param;
@@ -885,7 +924,7 @@ void set_scheduler(void)
 
 	rv = sched_get_priority_max(SCHED_RR);
 	if (rv != -1) {
-		sched_param.sched_priority = 2;
+		sched_param.sched_priority = rv;
 		rv = sched_setscheduler(0, SCHED_RR, &sched_param);
 		if (rv == -1)
 			log_print("could not set SCHED_RR priority %d err %d",
@@ -896,6 +935,29 @@ void set_scheduler(void)
 	}
 }
 
+void bail_with_log(int sig)
+{
+	int fd;
+	time_t now;
+
+	unlink(LOG_FILE);
+	fd = creat(LOG_FILE, S_IRUSR | S_IWUSR);
+	if (fd > 0) {
+		char now_ascii[32];
+
+		do_dump(fd);
+		memset(now_ascii, 0, sizeof(now_ascii));
+		time(&now);
+		sprintf(now_ascii, "%ld", now);
+		write(fd, now_ascii, strlen(now_ascii));
+		write(fd, " groupd segfault log follows:\n", 30);
+		close(fd);
+	} else
+		perror(LOG_FILE);
+	if (sig == SIGSEGV)
+		exit(0);
+}
+
 int main(int argc, char *argv[])
 {
 	prog_name = argv[0];
@@ -908,10 +970,14 @@ int main(int argc, char *argv[])
 
 	decode_arguments(argc, argv);
 
+	signal(SIGSEGV, bail_with_log);
+	signal(SIGUSR1, bail_with_log);
+
 	if (!groupd_debug_opt)
 		daemonize();
 
 	set_scheduler();
+	set_oom_adj(-16);
 
 	pollfd = malloc(NALLOC * sizeof(struct pollfd));
 	if (!pollfd)
