@@ -11,6 +11,7 @@
 ******************************************************************************/
 
 #include "lock_dlm.h"
+#include "ccs.h"
 
 #define SYSFS_DIR	"/sys/fs"
 #define JID_INIT	-9
@@ -1536,18 +1537,18 @@ int add_another_mountpoint(struct mountgroup *mg, char *dir, char *dev, int ci)
 	   fs with the same name (which is an error) */
 
 	if (strcmp(mg->dev, dev)) {
-		log_group(mg, "different fs dev %s with same name", mg->dev);
-		return -EINVAL;
+		log_error("different fs dev %s with same name", mg->dev);
+		return -EADDRINUSE;
 	}
 
 	if (find_mountpoint(mg, dir)) {
-		log_group(mg, "mount point %s already used", dir);
+		log_error("mount point %s already used", dir);
 		return -EBUSY;
 	}
 
 	/* we only really need to check one of these */
 	if (mg->mount_client || mg->mount_client_fd || !mg->kernel_mount_done) {
-		log_group(mg, "other mount in progress client %d fd %d done %d",
+		log_error("other mount in progress client %d fd %d done %d",
 			  mg->mount_client, mg->mount_client_fd,
 			  mg->kernel_mount_done);
 		return -EBUSY;
@@ -1575,14 +1576,14 @@ int do_mount(int ci, char *dir, char *type, char *proto, char *table,
 	struct mountgroup *mg = NULL;
 	char table2[MAXLINE];
 	char *cluster = NULL, *name = NULL;
-	int rv;
+	int rv, new_mg = 0;
 
 	log_debug("mount: %s %s %s %s %s %s",
 		  dir, type, proto, table, options, dev);
 
 	if (strcmp(proto, "lock_dlm")) {
 		log_error("mount: lockproto %s not supported", proto);
-		rv = -EINVAL;
+		rv = -EPROTONOSUPPORT;
 		goto out;
 	}
 
@@ -1590,7 +1591,7 @@ int do_mount(int ci, char *dir, char *type, char *proto, char *table,
 	    strstr(options, "first=") ||
 	    strstr(options, "id=")) {
 		log_error("mount: jid, first and id are reserved options");
-		rv = -EINVAL;
+		rv = -EOPNOTSUPP;
 		goto out;
 	}
 
@@ -1601,7 +1602,7 @@ int do_mount(int ci, char *dir, char *type, char *proto, char *table,
 
 	name = strstr(table2, ":");
 	if (!name) {
-		rv = -EINVAL;
+		rv = -EBADFD;
 		goto out;
 	}
 
@@ -1616,7 +1617,13 @@ int do_mount(int ci, char *dir, char *type, char *proto, char *table,
 
 	mg = find_mg(name);
 	if (mg) {
-		rv = add_another_mountpoint(mg, dir, dev, ci);
+		if (mg->reject_mounts) {
+			/* fs is being unmounted */
+			rv = -ESTALE;
+			log_error("mount: reject mount due to unmount");
+		} else {
+			rv = add_another_mountpoint(mg, dir, dev, ci);
+		}
 		goto out;
 	}
 
@@ -1625,6 +1632,7 @@ int do_mount(int ci, char *dir, char *type, char *proto, char *table,
 		rv = -ENOMEM;
 		goto out;
 	}
+	new_mg = 1;
 
 	mg->mount_client = ci;
 	strncpy(mg->type, type, sizeof(mg->type));
@@ -1634,7 +1642,7 @@ int do_mount(int ci, char *dir, char *type, char *proto, char *table,
 
 	if (strlen(cluster) != strlen(clustername) ||
 	    strlen(cluster) == 0 || strcmp(cluster, clustername)) {
-		rv = -1;
+		rv = -EBADR;
 		log_error("mount: fs requires cluster=\"%s\" current=\"%s\"",
 			  cluster, clustername);
 		goto out;
@@ -1646,7 +1654,7 @@ int do_mount(int ci, char *dir, char *type, char *proto, char *table,
 		mg->spectator = 1;
 	} else {
 		if (!we_are_in_fence_domain()) {
-			rv = -EINVAL;
+			rv = -ENOANO;
 			log_error("mount: not in default fence domain");
 			goto out;
 		}
@@ -1656,7 +1664,7 @@ int do_mount(int ci, char *dir, char *type, char *proto, char *table,
 		mg->rw = 1;
 	else if (strstr(options, "ro")) {
 		if (mg->spectator) {
-			rv = -EINVAL;
+			rv = -EROFS;
 			log_error("mount: readonly invalid with spectator");
 			goto out;
 		}
@@ -1664,8 +1672,8 @@ int do_mount(int ci, char *dir, char *type, char *proto, char *table,
 	}
 
 	if (strlen(options) > MAX_OPTIONS_LEN-1) {
-		rv = -EINVAL;
-		log_error("mount: options too long %d", strlen(options));
+		rv = -EMLINK;
+		log_error("mount: options too long %zu", strlen(options));
 		goto out;
 	}
 
@@ -1673,8 +1681,12 @@ int do_mount(int ci, char *dir, char *type, char *proto, char *table,
 	group_join(gh, name);
 	rv = 0;
  out:
-	*mg_ret = mg;
-	log_group(mg, "do_mount: rv %d", rv);
+	if (mg) {
+		*mg_ret = mg;
+		log_group(mg, "do_mount: rv %d", rv);
+	}
+	if (rv && new_mg)
+		free(mg);
 	return rv;
 }
 
@@ -2024,19 +2036,23 @@ int do_unmount(int ci, char *dir, int mnterr)
 	}
 
  out:
+	mg->reject_mounts = 1;
 	group_leave(gh, mg->name);
 	return 0;
 }
 
+#define LOCKSPACE_NODIR "/cluster/dlm/lockspace[@name=\"%s\"]/@nodir"
+
 void notify_mount_client(struct mountgroup *mg)
 {
-	char buf[MAXLINE];
-	int rv, error = 0;
+	char buf[MAXLINE], path[PATH_MAX], *str, tmp[MAXLINE];
+	int cd, rv, error = 0;
 	struct mg_member *memb;
-	
+
 	memb = find_memb_nodeid(mg, our_nodeid);
 
 	memset(buf, 0, MAXLINE);
+	memset(tmp, 0, MAXLINE);
 
 	if (mg->error_msg[0]) {
 		strncpy(buf, mg->error_msg, MAXLINE);
@@ -2053,6 +2069,28 @@ void notify_mount_client(struct mountgroup *mg)
 		else
 			snprintf(buf, MAXLINE, "hostdata=jid=%d:id=%u:first=%d",
 		 		 mg->our_jid, mg->id, mg->first_mounter);
+
+		if ((cd = ccs_connect()) < 0) {
+			log_error("notify_mount_client: ccs_connect failed");
+		}
+
+		memset(path, 0, PATH_MAX);
+		sprintf(path, LOCKSPACE_NODIR, mg->name);
+
+		rv = ccs_get(cd, path, &str);
+		if (rv || !str) {
+			log_debug("notify_mount_client: nodir not found for "
+				  "lockspace %s", mg->name);
+		} else {
+			snprintf(tmp, MAXLINE, ":nodir=%d", atoi(str));
+			strcat(buf, tmp);
+			free(str);
+		}
+
+		if (cd) {
+			log_debug("notify_mount_client: ccs_disconnect");
+			ccs_disconnect(cd);
+		}
 	}
 
 	log_debug("notify_mount_client: %s", buf);
