@@ -40,7 +40,11 @@
 #define L_SYS (1<<1)
 #define L_USER (1<<0)
 
-int configure_logging(int ccsfd, int debug);
+#ifdef WRAP_THREADS
+void dump_thread_states(FILE *);
+#endif
+int configure_rgmanager(int ccsfd, int debug);
+void set_transition_throttling(int);
 
 void node_event(int, int, int, int);
 void node_event_q(int, int, int, int);
@@ -48,7 +52,6 @@ int daemon_init(char *);
 int init_resource_groups(int);
 void kill_resource_groups(void);
 void set_my_id(int);
-int eval_groups(int, int, int);
 void flag_shutdown(int sig);
 void hard_exit(void);
 int send_rg_states(msgctx_t *, int);
@@ -64,7 +67,7 @@ static char *rgmanager_lsname = "rgmanager"; /* XXX default */
 
 int next_node_id(cluster_member_list_t *membership, int me);
 int rg_event_q(char *svcName, uint32_t state, int owner);
-
+void malloc_dump_table(FILE *, size_t, size_t);
 
 void
 segfault(int sig)
@@ -101,8 +104,7 @@ send_node_states(msgctx_t *ctx)
 					ml->cml_members[x].cn_nodeid, 0);
 		}
 	}
-	msg_send_simple(ctx, RG_SUCCESS,
-			ml->cml_members[x].cn_nodeid, 0);
+	msg_send_simple(ctx, RG_SUCCESS, 0, 0);
 	msg_receive(ctx, &hdr, sizeof(hdr), 10);
 	free_member_list(ml);
 }
@@ -200,6 +202,7 @@ membership_update(void)
 
 	cman_finish(h);
 	member_list_update(new_ml);
+	member_set_state(0, 0);		/* Mark qdisk as dead */
 
 	/*
 	 * Handle nodes lost.  Do our local node event first.
@@ -259,6 +262,7 @@ membership_update(void)
 
 	free_member_list(node_delta);
 	free_member_list(new_ml);
+	free_member_list(old_membership);
 
 	rg_unlockall(L_SYS);
 
@@ -404,7 +408,9 @@ dispatch_msg(msgctx_t *ctx, int nodeid, int need_close)
 	/* Peek-a-boo */
 	sz = msg_receive(ctx, msg_hdr, sizeof(msgbuf), 1);
 	if (sz < sizeof (generic_msg_hdr)) {
-		clulog(LOG_ERR, "#37: Error receiving message header (%d)\n", sz);
+		clulog(LOG_ERR,
+		       "#37: Error receiving header from %d sz=%d CTX %p\n",
+		       nodeid, sz, ctx);
 		goto out;
 	}
 
@@ -492,7 +498,9 @@ dispatch_msg(msgctx_t *ctx, int nodeid, int need_close)
 		/* Queue request */
 		rt_enqueue_request(msg_sm->sm_data.d_svcName,
 		  		   msg_sm->sm_data.d_action,
-		  		   ctx, 0, msg_sm->sm_data.d_svcOwner, 0, 0);
+		  		   ctx, 0, msg_sm->sm_data.d_svcOwner,
+		  		   msg_sm->sm_hdr.gh_arg1,
+		  		   msg_sm->sm_hdr.gh_arg2);
 		return 0;
 
 	case RG_EVENT:
@@ -590,6 +598,7 @@ handle_cluster_event(msgctx_t *ctx)
 		break;
 
 	case M_DATA:
+		nodeid = msg_get_nodeid(ctx);
 		dispatch_msg(ctx, nodeid, 0);
 		break;
 		
@@ -614,22 +623,43 @@ handle_cluster_event(msgctx_t *ctx)
 		clulog(LOG_WARNING, "#67: Shutting down uncleanly\n");
 		rg_set_inquorate();
 		rg_doall(RG_INIT, 1, "Emergency stop of %s");
+		rg_set_uninitialized();
 #if defined(LIBCMAN_VERSION) && LIBCMAN_VERSION >= 2
 		/* cman_replyto_shutdown() */
 #endif
-		exit(0);
+		running = 0;
+		break;
 	}
 
 	return ret;
 }
 
 
-void dump_threads(void);
+void dump_threads(FILE *fp);
+void dump_config_version(FILE *fp);
+void dump_vf_states(FILE *fp);
+void dump_cluster_ctx(FILE *fp);
+
+void
+dump_internal_state(char *loc)
+{
+	FILE *fp;
+	fp=fopen(loc, "w+");
+ 	dump_config_version(fp);
+ 	dump_threads(fp);
+ 	dump_vf_states(fp);
+#ifdef WRAP_THREADS
+	dump_thread_states(fp);
+#endif
+	dump_cluster_ctx(fp);
+	//malloc_dump_table(fp, 1, 16384); /* Only works if alloc.c us used */
+ 	fclose(fp);
+}
 
 int
 event_loop(msgctx_t *localctx, msgctx_t *clusterctx)
 {
-	int n, max, ret;
+	int n = 0, max, ret;
 	fd_set rfds;
 	msgctx_t *newctx;
 	struct timeval tv;
@@ -640,10 +670,8 @@ event_loop(msgctx_t *localctx, msgctx_t *clusterctx)
 
 	if (signalled) {
 		signalled = 0;
-		/*
-		malloc_stats();
-		dump_threads();
-		 */
+ 
+		dump_internal_state("/tmp/rgmanager-dump");
 	}
 
 	while (running && (tv.tv_sec || tv.tv_usec)) {
@@ -697,9 +725,12 @@ event_loop(msgctx_t *localctx, msgctx_t *clusterctx)
 		}
 	}
 
+	if (!running)
+		return 0;
+
 	if (need_reconfigure || check_config_update()) {
 		need_reconfigure = 0;
-		configure_logging(-1, 0);
+		configure_rgmanager(-1, 0);
 		init_resource_groups(1);
 		return 0;
 	}
@@ -739,7 +770,6 @@ void
 cleanup(msgctx_t *clusterctx)
 {
 	kill_resource_groups();
-	member_list_update(NULL);
 	send_exit_msg(clusterctx);
 }
 
@@ -752,14 +782,14 @@ statedump(int sig)
 }
 
 
-void malloc_dump_table(size_t, size_t);
+void malloc_dump_table(FILE *, size_t, size_t);
 
 
 /*
  * Configure logging based on data in cluster.conf
  */
 int
-configure_logging(int ccsfd, int dbg)
+configure_rgmanager(int ccsfd, int dbg)
 {
 	char *v;
 	char internal = 0;
@@ -782,6 +812,12 @@ configure_logging(int ccsfd, int dbg)
 		free(v);
 	}
 
+	if (ccs_get(ccsfd, "/cluster/rm/@transition_throttling", &v) == 0) {
+		if (!dbg)
+			set_transition_throttling(atoi(v));
+		free(v);
+	}
+
 	if (internal)
 		ccs_disconnect(ccsfd);
 
@@ -838,10 +874,13 @@ shutdown_thread(void *arg)
 	rg_doall(RG_STOP_EXITING, 1, NULL);
 	running = 0;
 
-	return 0;
+	pthread_exit(NULL);
 }
 
 
+#ifdef WRAP_THREADS
+void dump_thread_states(FILE *);
+#endif
 int
 main(int argc, char **argv)
 {
@@ -863,7 +902,9 @@ main(int argc, char **argv)
 			break;
 		case 'f':
 			foreground = 1;
+			break;
 		default:
+			return 1;
 			break;
 		}
 	}
@@ -882,6 +923,18 @@ main(int argc, char **argv)
 			clulog(LOG_NOTICE, "Failed to start watchdog\n");
 	}
 
+	setup_signal(SIGINT, flag_shutdown);
+	setup_signal(SIGTERM, flag_shutdown);
+	setup_signal(SIGUSR1, statedump);
+	unblock_signal(SIGCHLD);
+	setup_signal(SIGPIPE, SIG_IGN);
+
+	if (debug) {
+		setup_signal(SIGSEGV, segfault);
+	} else {
+		unblock_signal(SIGSEGV);
+	}
+
 	clu_initialize(&clu);
 	if (cman_init_subsys(clu) < 0) {
 		perror("cman_init_subsys");
@@ -909,7 +962,7 @@ main(int argc, char **argv)
 	   We know we're quorate.  At this point, we need to
 	   read the resource group trees from ccsd.
 	 */
-	configure_logging(-1, debug);
+	configure_rgmanager(-1, debug);
 	clulog(LOG_NOTICE, "Resource Group Manager Starting\n");
 
 	if (init_resource_groups(0) != 0) {
@@ -917,18 +970,6 @@ main(int argc, char **argv)
 		return -1;
 	}
 
-	setup_signal(SIGINT, flag_shutdown);
-	setup_signal(SIGTERM, flag_shutdown);
-	setup_signal(SIGUSR1, statedump);
-	unblock_signal(SIGCHLD);
-	setup_signal(SIGPIPE, SIG_IGN);
-
-	if (debug) {
-		setup_signal(SIGSEGV, segfault);
-	} else {
-		unblock_signal(SIGSEGV);
-	}
-
 	if (msg_listen(MSG_SOCKET, RGMGR_SOCK, me.cn_nodeid, &local_ctx) < 0) {
 		clulog(LOG_CRIT,
 		       "#10: Couldn't set up cluster message system: %s\n",
@@ -976,13 +1017,17 @@ main(int argc, char **argv)
 		event_loop(local_ctx, cluster_ctx);
 
 		if (shutdown_pending == 1) {
+			/* Kill local socket; local requests need to
+			   be ignored here */
+			msg_close(local_ctx);
 			++shutdown_pending;
 			clulog(LOG_NOTICE, "Shutting down\n");
 			pthread_create(&th, NULL, shutdown_thread, NULL);
 		}
 	}
 
-	cleanup(cluster_ctx);
+	if (rg_initialized())
+		cleanup(cluster_ctx);
 	clulog(LOG_NOTICE, "Shutdown complete, exiting\n");
 	clu_lock_finished(rgmanager_lsname);
 	cman_finish(clu);
