@@ -79,17 +79,27 @@ res_attr_value(resource_t *res, char *attrname)
 
    @param node		Resource tree node to look examine
    @param attrname	Attribute to retrieve.
+   @param ptype		Resource type to look for (if inheritance)
    @return 		value of attribute or NULL if not found
  */
-char *
-attr_value(resource_node_t *node, char *attrname)
+static char *
+_attr_value(resource_node_t *node, char *attrname, char *ptype)
 {
-	resource_t *res = node->rn_resource;
+	resource_t *res;
 	resource_attr_t *ra;
-	char *c, *p_type;
+	char *c, p_type[32];
 	ssize_t len;
 	int x;
 
+	if (!node)
+		return NULL;
+
+	res = node->rn_resource;
+
+	/* Go up the tree if it's not the right parent type */
+	if (ptype && strcmp(res->r_rule->rr_type, ptype))
+		return _attr_value(node->rn_parent, attrname, ptype);
+
 	for (x = 0; res->r_attrs && res->r_attrs[x].ra_name; x++) {
 		if (strcmp(attrname, res->r_attrs[x].ra_name))
 			continue;
@@ -108,28 +118,29 @@ attr_value(resource_node_t *node, char *attrname)
 		if (!c) {
 			/* Someone doesn't care or uses older
 			   semantics on inheritance */
-			return attr_value(node->rn_parent, ra->ra_value);
+			return _attr_value(node->rn_parent, ra->ra_value,
+					   NULL);
 		}
-
-		p_type = node->rn_parent->rn_resource->r_rule->rr_type;
+		
 		len = (c - ra->ra_value);
-
-		/* Different sizes? */
-		if (strlen(p_type) != len)
-			return NULL;
-
-		/* Check for parent type match */
-		if (strncmp(p_type, ra->ra_value, len) != 0)
-			return NULL;
-
+		memset(p_type, 0, sizeof(p_type));
+		memcpy(p_type, ra->ra_value, len);
+		
 		/* Skip the "%" and recurse */
-		return attr_value(node->rn_parent, ++c);
+		return _attr_value(node->rn_parent, ++c, p_type);
 	}
 
 	return NULL;
 }
 
 
+char *
+attr_value(resource_node_t *node, char *attrname)
+{
+	return _attr_value(node, attrname, NULL);
+}
+
+
 /**
   Run to the top of the tree.  Used to determine certain attributes of the
   resource group in-line, during resource tree operations.
@@ -165,18 +176,29 @@ primary_attr_value(resource_t *res)
 /**
    Compare two resources.
 
+  @param left	Left resource
+  @param right	Right resource	
+  @return	-1 on different resource, 0 if the same, 1 if different,
+		2 if different, but only safe resources are different
+
  */
 int
 rescmp(resource_t *left, resource_t *right)
 {
-	int x, y = 0, found;
+	int x, y = 0, found = 0, ret = 0;
+
 
 	/* Completely different resource class... */
 	if (strcmp(left->r_rule->rr_type, right->r_rule->rr_type)) {
-		//printf("Er, wildly different resource type! ");
 		return -1;
 	}
 
+	/*
+	printf("Comparing %s:%s to %s:%s\n",
+	       left->r_rule->rr_type, left->r_attrs[0].ra_value,
+	       right->r_rule->rr_type, right->r_attrs[0].ra_value)
+	 */
+
 	for (x = 0; left->r_attrs && left->r_attrs[x].ra_name; x++) {
 
 		found = 0;
@@ -192,35 +214,52 @@ rescmp(resource_t *left, resource_t *right)
 			    left->r_attrs[x].ra_flags) {
 				/* Flags are different.  Change in
 				   resource agents? */
-				//printf("flags differ ");
+				/*
+				printf("* flags differ %08x vs %08x\n",
+				       left->r_attrs[x].ra_flags,
+				       right->r_attrs[y].ra_flags);
+				 */
 				return 1;
 			}
 
 			if (strcmp(right->r_attrs[y].ra_value,
 				   left->r_attrs[x].ra_value)) {
 				/* Different attribute value. */
-				//printf("different value for attr '%s'  ",
-				       //right->r_attrs[y].ra_name);
-				return 1;
+				/*
+				printf("* different value for attr '%s':"
+				       " '%s' vs '%s'",
+				       right->r_attrs[y].ra_name,
+				       left->r_attrs[x].ra_value,
+				       right->r_attrs[y].ra_value);
+				 */
+				if (left->r_attrs[x].ra_flags & RA_RECONFIG) {
+					/* printf(" [SAFE]\n"); */
+					ret = 2;
+			 	} else {
+					/* printf("\n"); */
+					return 1;
+				}
 			}
 		}
 
 		/* Attribute missing -> different attribute value. */
 		if (!found) {
-			//printf("Attribute %s deleted  ",
-			       //left->r_attrs[x].ra_name);
+			/*
+			printf("* Attribute '%s' deleted\n",
+			       left->r_attrs[x].ra_name);
+			 */
 			return 1;
 		}
 	}
 
 	/* Different attribute count */
 	if (x != y) {
-		//printf("Attribute count differ (attributes added!) ");
+		/* printf("* Attribute count differ (attributes added!) "); */
 		return 1;
 	}
 
 	/* All the same */
-	return 0;
+	return ret;
 }
 
 
@@ -277,7 +316,7 @@ find_root_by_ref(resource_t **reslist, char *ref)
 	resource_t *curr;
 	char ref_buf[128];
 	char *type;
-	char *name;
+	char *name = ref;
 	int x;
 
 	snprintf(ref_buf, sizeof(ref_buf), "%s", ref);
@@ -413,17 +452,52 @@ xpath_get_one(xmlDocPtr doc, xmlXPathContextPtr ctx, char *query)
 {
 	char *val = NULL, *ret = NULL;
 	xmlXPathObjectPtr obj;
+	xmlNodePtr node;
+	size_t size = 0;
+	int nnv = 0;
 
 	obj = xmlXPathEvalExpression((unsigned char *)query, ctx);
 	if (!obj)
 		return NULL;
+	if (!obj->nodesetval)
+		goto out;
+	if (obj->nodesetval->nodeNr <= 0)
+		goto out;
 
-	if (obj->nodesetval && obj->nodesetval->nodeNr >= 1) {
-		val = (char *)obj->nodesetval->nodeTab[0]->children->content;
-		if (strlen(val) >= 1)
-			ret = strdup(val);
+	node = obj->nodesetval->nodeTab[0];
+	if(!node)
+		goto out;
+
+	if (((node->type == XML_ATTRIBUTE_NODE) && strstr(query, "@*")) ||
+	    ((node->type == XML_ELEMENT_NODE) && strstr(query, "child::*"))){
+		if (node->children && node->children->content)
+	  		size = strlen((char *)node->children->content)+
+				      strlen((char *)node->name)+2;
+		else 
+			size = strlen((char *)node->name)+2;
+		nnv = 1;
+	} else {
+		if (node->children && node->children->content) {
+			size = strlen((char *)node->children->content)+1;
+		} else {
+			goto out;
+		}
 	}
 
+	val = (char *)malloc(size);
+	if(!val)
+		goto out;
+	memset(val, 0, size);
+	if (nnv) {
+		sprintf(val, "%s=%s", node->name, node->children ?
+			(char *)node->children->content:"");
+	} else {
+		sprintf(val, "%s", node->children ? node->children->content :
+			node->name);
+	}
+
+	ret = val;
+out:
 	xmlXPathFreeObject(obj);
 
 	return ret;
@@ -498,6 +572,8 @@ print_resource(resource_t *res)
 		printf(" [NEEDSTOP]");
 	if (res->r_flags & RF_COMMON)
 		printf(" [COMMON]");
+	if (res->r_flags & RF_RECONFIG)
+		printf(" [RECONFIG]");
 	printf("\n");
 
 	if (res->r_rule->rr_maxrefs)
@@ -533,6 +609,8 @@ print_resource(resource_t *res)
 			printf(" unique");
 		if (res->r_attrs[x].ra_flags & RA_REQUIRED)
 			printf(" required");
+		if (res->r_attrs[x].ra_flags & RA_RECONFIG)
+			printf(" reconfig");
 		if (res->r_attrs[x].ra_flags & RA_INHERIT)
 			printf(" inherit(\"%s\")", res->r_attrs[x].ra_value);
 		printf(" ]\n");
